/******************** (C) COPYRIGHT 2010 Intersil ********************
* File Name          : video_en_driver.c
* Author             : zhang zhao
* Revision           : 1.2.1.1289
* Date               : 2012-04-05
* Description        : the tc_driver's staff for video encode
* Change log         : none
********************************************************************************
* THE PRESENT SOFTWARE IS DISTRIBUTED IN THE HOPE THAT IT WILL PROVIDE CUSTOMERS 
* WITH CODING INFORMATION TO SPEED UP THEIR PRODUCT DEVELOPMENT, BUT WITHOUT
* ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF MERCHANTABILITY OR FITNESS
* FOR A PARTICULAR PURPOSE. AS A RESULT, TECHWELL SHALL NOT BE HELD LIABLE FOR ANY
* DIRECT, INDIRECT OR CONSEQUENTIAL DAMAGES WITH RESPECT TO ANY CLAIMS ARISING
* FROM THE CONTENT OF SUCH SOFTWARE AND/OR THE USE MADE BY CUSTOMERS OF THE
* CODING INFORMATION CONTAINED HEREIN IN CONNECTION WITH THEIR PRODUCTS.
*******************************************************************************/


#include <isil5864/tc_common.h>

/*******************************************************************************
* Function Name  : isil_video_en_register_driver_init
* Description    : create a driver instance for h264 video encode(main stream) according to arg
* Input          : struct isil_chip_device *tcd,struct isil_service_device * tsd,void * para,struct cmd_arg *arg
* Output         : NULL
* Return         : 0:success; !0:error
*******************************************************************************/
int isil_video_en_register_driver_init(struct isil_chip_device *tcd,struct isil_service_device * tsd,void * para,struct cmd_arg *arg)
{
    int ret, phy_slot_id;
    unsigned long bus_id,chip_id,lg_id;
    struct isil_logic_phy_table *logic_map_phy;
    isil_h264_logic_encode_chan_t  *ved;
    isil_chip_t  *chip = (isil_chip_t  *)para;

    logic_map_phy = &chip->chip_driver->logic_map_phy;
    bus_id = get_busid_from_epobj(&tcd->epobj);
    chip_id = get_chipid_from_epobj(&tcd->epobj);
    lg_id = arg->channel_idx;

    ved = chip->h264_master_encode_logic_chan;
    ved += lg_id;

    if(logic_map_phy->op->find_phy_by_logic(logic_map_phy, LOGIC_MAP_TABLE_BIND_H264E, lg_id, &phy_slot_id) < 0){
        dbg("can't find phy_id by logic_id %lu\n", lg_id);
        return -ENODEV;
    }
    ret = init_isil_h264_encode_chan(ved, chip->bus_id, chip->chip_id, phy_slot_id, ISIL_MASTER_BITSTREAM, chip);
    if(ret){
        dbg("h264 master driv init failed!\n");
        return ret;
    }
    tsd->ped = &ved->opened_logic_chan_ed;
    return ret;
}


/*******************************************************************************
* Function Name  : isil_video_en_register_driver_remove
* Description    : remove a driver instance for h264 video encode(main stream) according to arg
* Input          : struct isil_chip_device *tcd,struct isil_service_device * tsd,void * para,struct cmd_arg *arg
* Output         : NULL
* Return         : 0:success; !0:error
*******************************************************************************/
void isil_video_en_register_driver_remove(struct isil_chip_device *tcd,struct isil_service_device * tsd,void * para,struct cmd_arg *arg)
{
    isil_h264_logic_encode_chan_t  *vd;

    tc_trace;
    if(!tsd->ped){
        dbg("tsd has no ed!\n");
        return;
    }
    vd = container_of(tsd->ped,isil_h264_logic_encode_chan_t,opened_logic_chan_ed);
    remove_isil_h264_encode_chan(vd);
}


struct isil_dev_id  isil_video_en_dev_id =
{
    .epobj = {
		.vendor_id = ISIL,
#if	defined(PCI_PLATFORM)
        .bus_id =  ISIL_ED_PCI_BUS,
#else
        .bus_id =  ISIL_ED_HOST_BUS,
#endif
        .chip_id = TW5864,
        .func_id = STREAM_TYPE_MAJOR | CODEC_VIDEO_H264,
        .type = ISIL_ENCODER,                
    },
    .version = 1,
};


struct tsd_driver_operations  isil_video_en_ops =
{
    .init = isil_video_en_register_driver_init,
    .remove = isil_video_en_register_driver_remove,
};

struct isil_service_driver   isil_video_en_driv =
{
    .driver = {
        .name = "tw5864_video_driv",
    },
    .device_id = {
        .tid = &isil_video_en_dev_id,
        .ops = &isil_video_en_ops,
        //.driver_data =  //parameters
    },
};


/*******************************************************************************
* Function Name  : isil_video_en_driver_init
* Description    : register the video encode(main stream) releated dev_table_id to tc_core
* Input          : NULL
* Output         : NULL
* Return         : 0:success; !0:error
*******************************************************************************/
int isil_video_en_driver_init(void)
{
    tc_trace;
    //isil_video_en_driv.device_id.para = para;
    return register_tc_driver(&isil_video_en_driv);
}


/*******************************************************************************
* Function Name  : isil_video_en_driver_remove
* Description    : unregister the video encode(main stream) releated dev_table_id from tc_core
* Input          : NULL
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void isil_video_en_driver_remove()
{
    unregister_tc_driver(&isil_video_en_driv);
}


/*******************************************************************************
* Function Name  : isil_videosub_en_register_driver_init
* Description    : create a driver instance for h264 video encode(sub stream) according to arg
* Input          : struct isil_chip_device *tcd,struct isil_service_device * tsd,void * para,struct cmd_arg *arg
* Output         : NULL
* Return         : 0:success; !0:error
*******************************************************************************/
int isil_videosub_en_register_driver_init(struct isil_chip_device *tcd,struct isil_service_device * tsd,void * para,struct cmd_arg *arg)
{
    int ret, phy_slot_id;
    unsigned long bus_id,chip_id,lg_id;
    struct isil_logic_phy_table *logic_map_phy;
    isil_chip_t  *chip = (isil_chip_t  *)para;
    isil_h264_logic_encode_chan_t  *ved;

    logic_map_phy = &chip->chip_driver->logic_map_phy;
    bus_id = get_busid_from_epobj(&tcd->epobj);
    chip_id = get_chipid_from_epobj(&tcd->epobj);
    lg_id = arg->channel_idx;

    ved = chip->h264_sub_encode_logic_chan;
    ved += lg_id;

    if(logic_map_phy->op->find_phy_by_logic(logic_map_phy, LOGIC_MAP_TABLE_BIND_H264E, lg_id, &phy_slot_id) < 0){
        dbg("can't find phy_id by logic_id %lu\n", lg_id);
        return -ENODEV;
    }
    ret = init_isil_h264_encode_chan(ved, chip->bus_id, chip->chip_id, phy_slot_id, ISIL_SUB_BITSTREAM, chip);
    if(ret){
        dbg("sub h264 driv init failed!\n");
        return ret;
    }
    tsd->ped = &ved->opened_logic_chan_ed;
    return ret;
}


/*******************************************************************************
* Function Name  : isil_videosub_en_register_driver_remove
* Description    : remove a driver instance for h264 video encode(main stream) according to arg
* Input          : struct isil_chip_device *tcd,struct isil_service_device * tsd,void * para,struct cmd_arg *arg
* Output         : NULL
* Return         : 0:success; !0:error
*******************************************************************************/
void isil_videosub_en_register_driver_remove(struct isil_chip_device *tcd,struct isil_service_device * tsd,void * para,struct cmd_arg *arg)
{
    isil_h264_logic_encode_chan_t  *vd;

    tc_trace;
    if(!tsd->ped){
        dbg("tsd has no ed!\n");
        return;
    }
    vd = container_of(tsd->ped,isil_h264_logic_encode_chan_t,opened_logic_chan_ed);
    remove_isil_h264_encode_chan(vd);
}



struct isil_dev_id  isil_videosub_en_dev_id =
{
    .epobj = {
		.vendor_id = ISIL,
#if	defined(PCI_PLATFORM)
        .bus_id =  ISIL_ED_PCI_BUS,
#else
        .bus_id =  ISIL_ED_HOST_BUS,
#endif
        .chip_id = TW5864,
        .func_id = (STREAM_TYPE_MINOR << 16) | CODEC_VIDEO_H264,
        .type = ISIL_ENCODER,                
    },
    .version = 1,
};


struct tsd_driver_operations  isil_videosub_en_ops =
{
    .init = isil_videosub_en_register_driver_init,
    .remove = isil_videosub_en_register_driver_remove,
};

struct isil_service_driver   isil_videosub_en_driv =
{
    .driver = {
        .name = "tw5864_videosub_driv",
    },
    .device_id = {
        .tid = &isil_videosub_en_dev_id,
        .ops = &isil_videosub_en_ops,
        //.driver_data =  //parameters
    },
};


/*******************************************************************************
* Function Name  : isil_videosub_en_driver_init
* Description    : register video encode(sub stream) releated dev_table_id to tc_core
* Input          : NULL
* Output         : NULL
* Return         : 0:success; !0:error
*******************************************************************************/
int isil_videosub_en_driver_init(void)
{
    tc_trace;
    //isil_videosub_en_driv.device_id.para = para;
    return register_tc_driver(&isil_videosub_en_driv);
}


/*******************************************************************************
* Function Name  : isil_videosub_en_driver_remove
* Description    : unregister video encode(sub stream) releated dev_table_id from tc_core
* Input          : NULL
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void isil_videosub_en_driver_remove()
{
    unregister_tc_driver(&isil_videosub_en_driv);
}

