/******************** (C) COPYRIGHT 2010 Intersil ********************
* File Name         : isil_chip_vo.h
* Author            : rxchen@intersil.com
* Revision          : 
* Date              : 
* Description       : push buffer manager
********************************************************************************
* THE PRESENT SOFTWARE IS DISTRIBUTED IN THE HOPE THAT IT WILL PROVIDE CUSTOMERS 
* WITH CODING INFORMATION TO SPEED UP THEIR PRODUCT DEVELOPMENT, BUT WITHOUT
* ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF MERCHANTABILITY OR FITNESS
* FOR A PARTICULAR PURPOSE. AS A RESULT, TECHWELL SHALL NOT BE HELD LIABLE FOR ANY
* DIRECT, INDIRECT OR CONSEQUENTIAL DAMAGES WITH RESPECT TO ANY CLAIMS ARISING
* FROM THE CONTENT OF SUCH SOFTWARE AND/OR THE USE MADE BY CUSTOMERS OF THE
* CODING INFORMATION CONTAINED HEREIN IN CONNECTION WITH THEIR PRODUCTS.
*******************************************************************************/

#ifndef _ISIL_PUSH_BUF_MGT_H__
#define _ISIL_PUSH_BUF_MGT_H__

#ifdef __cplusplus
extern "C"
{
#endif

#define USE_DMA_ALLOC_COHERENT
	
#define H264_BUF_CNT    (4)
#define MV_BUF_CNT      (H264_BUF_CNT)
#define AUDIO_BUF_CNT   (8)
#define PREVIEW_BUF_CNT (4) 
#define JPEG_BUF_CNT    (1)

#define PREVIEW_BUF_SIZE    (0x100000)
//kb: change it to 1M
#define H264_BUF_SIZE       (0x100000)
//#define H264_BUF_SIZE       (0x80000)
#define AUDIO_BUF_SIZE      (0x1000)
#define MV_BUF_SIZE         (0x40000)
//kb: change it to 1M
#define JPEG_BUF_SIZE       (0x100000)
//#define JPEG_BUF_SIZE       (0x80000)

#define PREVIEW_BUF_ORDER	(8) //256*4096
#define H264_BUF_ORDER		(7)
#define AUDIO_BUF_ORDER		(0) 
#define MV_BUF_ORDER		(6)
#define JPEG_BUF_ORDER		(8) //256*4096

#define PIECE_PUSH_SIZE		(PREVIEW_BUF_SIZE)
#define D1_PUSH_SIZE		(PIECE_PUSH_SIZE)
#define CIF_PUSH_SIZE		(D1_PUSH_SIZE >> 2)
#define QCIF_PUSH_SIZE		(CIF_PUSH_SIZE >> 2)

#define PIECE_CNT_PER_BUF	(1)
#define D1_CNT_PER_PIECE	(1) 
#define CIF_CNT_PER_PIECE	(4)
#define QCIF_CNT_PER_PIECE	(16)

enum _push_buffer_type {
    BUF_TYPE_DEFAULT = 0,
    BUF_TYPE_H264,
    BUF_TYPE_AUDIO,
    BUF_TYPE_PREVIEW,
    BUF_TYPE_MV,
    BUF_TYPE_JPEG,
};

#define BUF_STATE_FREE 0
#define BUF_STATE_BUSY 1

#define BUF_NO_OCCUPIED 0x0 
#define BUF_SD_OCCUPIED 0x1 
#define BUF_HD_OCCUPIED 0x2

#define MASK_LOW16 0xffff 

typedef struct _push_buffer {
#ifndef USE_DMA_ALLOC_COHERENT
	unsigned long buf_addr;
#else
	void *buf_addr;
#endif
	dma_addr_t	dma_addr;
	int cache_sync_len;
	int state;
} push_buffer_t;

typedef struct _push_prev_buffer {
#ifndef USE_DMA_ALLOC_COHERENT
	unsigned long buf_addr;
#else
	void *buf_addr;
#endif	
	dma_addr_t	dma_addr;
	u32 bit_map[PIECE_CNT_PER_BUF];
	int fmt[PIECE_CNT_PER_BUF];
	int cache_sync_len;
	int occupied;
} push_prev_buffer_t;

typedef struct _push_buf_pool {
    push_buffer_t h264_bufs[H264_BUF_CNT];
    u32 h264_free_cnt;
    spinlock_t h264_queue_lock;

    push_buffer_t audio_bufs[AUDIO_BUF_CNT];
    u32 audio_free_cnt;
    spinlock_t audio_queue_lock;

    push_prev_buffer_t preview_bufs[PREVIEW_BUF_CNT];
    u32 preview_free_cnt;
    spinlock_t preview_queue_lock;

    push_buffer_t mv_bufs[MV_BUF_CNT];
    u32 mv_free_cnt;
    spinlock_t mv_queue_lock;	

    push_buffer_t jpeg_bufs[JPEG_BUF_CNT];
    u32 jpeg_free_cnt;
    spinlock_t jpeg_queue_lock;

    isil_chip_t	*chip;
} push_buf_pool_t;
/*
   typedef struct _push_buf_pool {
   void *buf;
   u32 use_count;
   } push_buf_pool_t;
   */
extern int push_buf_pool_init(isil_chip_t *chip);
extern void push_buf_pool_release(isil_chip_t *chip);

extern int  get_push_buf_cache_sync_len(int type, unsigned long addr, isil_chip_t *chip);
extern void update_push_buf_cache_sync_len(int type, int sync_len, unsigned long addr, isil_chip_t *chip);

extern unsigned long	get_push_buf(int type, int fmt, dma_addr_t *dma_addr, isil_chip_t *chip);
extern void release_push_buf(int type, int fmt, unsigned long addr, isil_chip_t *chip);
#ifdef __cplusplus
}
#endif

#endif //_ISIL_PUSH_BUF_MGT_H__
