/***************************************** (C) COPYRIGHT 2010 Intersil *****************************************
 * File Name         : isil_kthread.c
 * Author            : tang.sc@intersil.com
 * Revision          : 1.2.1.1275
 * Date              : 2012-03-29
 * Description       : driver have a kthread for sync chip level's operation. 
 * Change log        : 
 *                   
 ****************************************************************************************************************
 * THE PRESENT SOFTWARE IS DISTRIBUTED IN THE HOPE THAT IT WILL PROVIDE CUSTOMERS 
 * WITH CODING INFORMATION TO SPEED UP THEIR PRODUCT DEVELOPMENT, BUT WITHOUT
 * ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF MERCHANTABILITY OR FITNESS
 * FOR A PARTICULAR PURPOSE. AS A RESULT, TECHWELL SHALL NOT BE HELD LIABLE FOR ANY
 * DIRECT, INDIRECT OR CONSEQUENTIAL DAMAGES WITH RESPECT TO ANY CLAIMS ARISING
 * FROM THE CONTENT OF SUCH SOFTWARE AND/OR THE USE MADE BY CUSTOMERS OF THE
 * CODING INFORMATION CONTAINED HEREIN IN CONNECTION WITH THEIR PRODUCTS.
 ****************************************************************************************************************/
#include	<isil5864/isil_common.h>

/****************************************************************************************************************
 * Function Name :   h264_module_phy_video_slot_reset_preproessing
 * Description   :   this function process h264 module phy video slot preproessing.
 * Input         :   param1 is phy video slot
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void h264_module_phy_video_slot_reset_preproessing(isil_h264_phy_video_slot_t *phy_video_slot)
{
    isil_register_table_t   *opened_logic_chan_table, *logic_chan_table;
    isil_register_node_t    *register_node;
    struct list_head        *list;
    ed_tcb_t                *opened_logic_chan_ed;
    isil_h264_logic_encode_chan_t   *h264_logic_encode_chan;
    unsigned long           flags1, flags2;
    int entry_number, entry_number_pre_close, total_entry_number, have_precessed_number;

    opened_logic_chan_table = &phy_video_slot->opened_logic_chan_table;
    logic_chan_table = &phy_video_slot->logic_chan_table;
again:
    spin_lock_irqsave(&opened_logic_chan_table->lock, flags1);
    spin_lock_irqsave(&logic_chan_table->lock, flags2);
    total_entry_number = entry_number = opened_logic_chan_table->op->get_curr_register_node_number(opened_logic_chan_table);
    have_precessed_number = entry_number_pre_close = 0;
    if(entry_number){
        list_for_each(list, &opened_logic_chan_table->register_header){
            if(entry_number == 0){
                break;
            }
            entry_number--;
            register_node = to_get_register_node(list);
            opened_logic_chan_ed = to_get_endpoint_tcb_with_register_node(register_node);
            h264_logic_encode_chan = to_get_isil_h264_encode_chan_with_opened_logic_chan_ed(opened_logic_chan_ed);
            if(opened_logic_chan_ed->op->get_state(opened_logic_chan_ed) == ISIL_ED_UNREGISTER){
                entry_number_pre_close++;
                continue;
            } else {
                start_isil_h264_encode_chan_robust_process(h264_logic_encode_chan);
            }
            spin_unlock_irqrestore(&logic_chan_table->lock, flags2);
            spin_unlock_irqrestore(&opened_logic_chan_table->lock, flags1);

            reset_isil_h264_encode_chan_preprocessing(opened_logic_chan_ed);
            have_precessed_number++;
            if((entry_number_pre_close+have_precessed_number) < total_entry_number){
                goto again;
            } else {
                break;
            }
        }
    }
    spin_unlock_irqrestore(&logic_chan_table->lock, flags2);
    spin_unlock_irqrestore(&opened_logic_chan_table->lock, flags1);
}

/****************************************************************************************************************
 * Function Name :   h264_module_reset_preprocessing
 * Description   :   h264 module reset preproessing.
 * Input         :   param1 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void h264_module_reset_preprocessing(isil_chip_t *chip)
{
    isil_vd_cross_bus_t *chip_vd_cross_bus;
    isil_h264_phy_video_slot_t  *phy_video_slot;
    int     i;

    chip_vd_cross_bus = &chip->chip_vd_cross_bus;
    for(i=0; i<ISIL_PHY_VD_CHAN_NUMBER; i++){
        phy_video_slot = &chip_vd_cross_bus->phy_video_slot[i];
        h264_module_phy_video_slot_reset_preproessing(phy_video_slot);
    }
}

/****************************************************************************************************************
 * Function Name :   h264_module_phy_video_slot_reset_proessing
 * Description   :   h264 module phy video slot reset proessing.
 * Input         :   param1 is phy video slot
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void h264_module_phy_video_slot_reset_proessing(isil_h264_phy_video_slot_t *phy_video_slot)
{
    isil_register_table_t   *opened_logic_chan_table, *logic_chan_table;
    isil_register_node_t    *register_node;
    struct list_head        *list;
    ed_tcb_t                *opened_logic_chan_ed, *logic_chan_ed;
    isil_h264_logic_encode_chan_t   *encode_chan;
    unsigned long           flags1, flags2;

    opened_logic_chan_table = &phy_video_slot->opened_logic_chan_table;
    logic_chan_table = &phy_video_slot->logic_chan_table;
    spin_lock_irqsave(&opened_logic_chan_table->lock, flags1);
    spin_lock_irqsave(&logic_chan_table->lock, flags2);
    if(logic_chan_table->op->get_curr_register_node_number(logic_chan_table)){
        list_for_each(list, &logic_chan_table->register_header){
            register_node = to_get_register_node(list);
            spin_unlock_irqrestore(&logic_chan_table->lock, flags2);
            spin_unlock_irqrestore(&opened_logic_chan_table->lock, flags1);

            logic_chan_ed = to_get_endpoint_tcb_with_register_node(register_node);
            encode_chan = to_get_isil_h264_encode_chan_with_logic_chan_ed(logic_chan_ed);
            if(isil_h264_encode_chan_is_in_robust_processing(encode_chan) == 1){
                opened_logic_chan_ed = &encode_chan->opened_logic_chan_ed;
                reset_isil_h264_encode_chan_processing(opened_logic_chan_ed);
            }

            spin_lock_irqsave(&opened_logic_chan_table->lock, flags1);
            spin_lock_irqsave(&logic_chan_table->lock, flags2);
        }
    }
    spin_unlock_irqrestore(&logic_chan_table->lock, flags2);
    spin_unlock_irqrestore(&opened_logic_chan_table->lock, flags1);
}

/****************************************************************************************************************
 * Function Name :   h264_module_reset_processing
 * Description   :   h264 module reset proessing.
 * Input         :   param1 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void h264_module_reset_processing(isil_chip_t *chip)
{
    isil_vd_cross_bus_t *chip_vd_cross_bus;
    isil_h264_phy_video_slot_t  *phy_video_slot;
    int     i;

    chip_vd_cross_bus = &chip->chip_vd_cross_bus;
    for(i=0; i<ISIL_PHY_VD_CHAN_NUMBER; i++){
        phy_video_slot = &chip_vd_cross_bus->phy_video_slot[i];
        h264_module_phy_video_slot_reset_proessing(phy_video_slot);
    }
}

/****************************************************************************************************************
 * Function Name :   h264_module_phy_video_slot_reset_postproessing
 * Description   :   h264 module phy video slot reset postproessing.
 * Input         :   param1 is phy video slot
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void h264_module_phy_video_slot_reset_postproessing(isil_h264_phy_video_slot_t *phy_video_slot)
{
    isil_register_table_t   *opened_logic_chan_table, *logic_chan_table;
    isil_register_node_t    *register_node;
    struct list_head        *list;
    ed_tcb_t                *opened_logic_chan_ed, *logic_chan_ed;
    isil_h264_logic_encode_chan_t   *encode_chan;
    unsigned long           flags1, flags2;

    opened_logic_chan_table = &phy_video_slot->opened_logic_chan_table;
    logic_chan_table = &phy_video_slot->logic_chan_table;
    spin_lock_irqsave(&opened_logic_chan_table->lock, flags1);
    spin_lock_irqsave(&logic_chan_table->lock, flags2);
    if(logic_chan_table->op->get_curr_register_node_number(logic_chan_table)){
        list_for_each(list, &logic_chan_table->register_header){
            register_node = to_get_register_node(list);
            spin_unlock_irqrestore(&logic_chan_table->lock, flags2);
            spin_unlock_irqrestore(&opened_logic_chan_table->lock, flags1);

            logic_chan_ed = to_get_endpoint_tcb_with_register_node(register_node);
            encode_chan = to_get_isil_h264_encode_chan_with_logic_chan_ed(logic_chan_ed);
            if(isil_h264_encode_chan_is_in_robust_processing(encode_chan) == 1){
                opened_logic_chan_ed = &encode_chan->opened_logic_chan_ed;
                reset_isil_h264_encode_chan_postprocessing(opened_logic_chan_ed);
            }

            spin_lock_irqsave(&opened_logic_chan_table->lock, flags1);
            spin_lock_irqsave(&logic_chan_table->lock, flags2);
        }
    }
    spin_unlock_irqrestore(&logic_chan_table->lock, flags2);
    spin_unlock_irqrestore(&opened_logic_chan_table->lock, flags1);
}

/****************************************************************************************************************
 * Function Name :   h264_module_reset_postprocessing
 * Description   :   h264 module reset postproessing.
 * Input         :   param1 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void h264_module_reset_postprocessing(isil_chip_t *chip)
{
    isil_vd_cross_bus_t *chip_vd_cross_bus;
    isil_h264_phy_video_slot_t  *phy_video_slot;
    int     i;

    chip_vd_cross_bus = &chip->chip_vd_cross_bus;
    for(i=0; i<ISIL_PHY_VD_CHAN_NUMBER; i++){
        phy_video_slot = &chip_vd_cross_bus->phy_video_slot[i];
        h264_module_phy_video_slot_reset_postproessing(phy_video_slot);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_module_reset_preprocessing
 * Description   :   audio module reset preproessing.
 * Input         :   param1 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_module_reset_preprocessing(isil_chip_t *chip)
{
    isil_register_table_t   *opened_logic_chan_table, *logic_chan_table;
    isil_register_node_t    *register_node;
    struct list_head        *list;
    ed_tcb_t                *opened_logic_chan_ed;
    isil_audio_driver_t     *audio_chan_driver;
    chip_audio_t            *chip_audio;    
    unsigned long           flags1, flags2;
    int entry_number;

    chip_audio = &chip->chip_audio;
    opened_logic_chan_table = &chip_audio->opened_logic_chan_table;
    logic_chan_table = &chip_audio->logic_chan_table;

    spin_lock_irqsave(&opened_logic_chan_table->lock, flags1);
    spin_lock_irqsave(&logic_chan_table->lock, flags2);
    entry_number = opened_logic_chan_table->op->get_curr_register_node_number(opened_logic_chan_table);
    if(entry_number){
        list_for_each(list, &opened_logic_chan_table->register_header){
            if(entry_number == 0){
                break;
            }
            entry_number--;
            register_node = to_get_register_node(list);
            opened_logic_chan_ed = to_get_endpoint_tcb_with_register_node(register_node);
            audio_chan_driver = to_isil_audio_chan_driver_with_opened_logic_chan_ed(opened_logic_chan_ed);
            if(opened_logic_chan_ed->op->get_state(opened_logic_chan_ed) == ISIL_ED_UNREGISTER){
                continue;
            } else {
                start_audio_chan_robust_process(audio_chan_driver);
            }
            spin_unlock_irqrestore(&logic_chan_table->lock, flags2);
            spin_unlock_irqrestore(&opened_logic_chan_table->lock, flags1);

            if(audio_chan_driver->audio_logic_chan_id < ISIL_AUDIO_IN_CHAN_NUMBER){
                reset_isil_audio_encode_chan_preprocessing(opened_logic_chan_ed);
            } else {
                reset_isil_audio_decode_chan_preprocessing(opened_logic_chan_ed);
            }

            spin_lock_irqsave(&opened_logic_chan_table->lock, flags1);
            spin_lock_irqsave(&logic_chan_table->lock, flags2);
        }
    }
    spin_unlock_irqrestore(&logic_chan_table->lock, flags2);
    spin_unlock_irqrestore(&opened_logic_chan_table->lock, flags1);
}

/****************************************************************************************************************
 * Function Name :   audio_module_reset_processing
 * Description   :   audio module reset proessing.
 * Input         :   param1 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_module_reset_processing(isil_chip_t *chip)
{
    isil_register_table_t   *opened_logic_chan_table, *logic_chan_table;
    isil_register_node_t    *register_node;
    struct list_head        *list;
    ed_tcb_t                *opened_logic_chan_ed, *logic_chan_ed;
    isil_audio_driver_t     *audio_chan_driver;
    chip_audio_t            *chip_audio;    
    unsigned long           flags1, flags2;

    chip_audio = &chip->chip_audio;
    opened_logic_chan_table = &chip_audio->opened_logic_chan_table;
    logic_chan_table = &chip_audio->logic_chan_table;
    spin_lock_irqsave(&opened_logic_chan_table->lock, flags1);
    spin_lock_irqsave(&logic_chan_table->lock, flags2);
    if(logic_chan_table->op->get_curr_register_node_number(logic_chan_table)){
        list_for_each(list, &logic_chan_table->register_header){
            register_node = to_get_register_node(list);
            spin_unlock_irqrestore(&logic_chan_table->lock, flags2);
            spin_unlock_irqrestore(&opened_logic_chan_table->lock, flags1);

            logic_chan_ed = to_get_endpoint_tcb_with_register_node(register_node);
            audio_chan_driver = to_isil_audio_chan_driver_with_logic_chan_ed(logic_chan_ed);
            if(isil_audio_chan_is_in_robust_processing(audio_chan_driver) == 1){
                opened_logic_chan_ed = &audio_chan_driver->opened_logic_chan_ed;
                if(audio_chan_driver->audio_logic_chan_id < ISIL_AUDIO_IN_CHAN_NUMBER){
                    reset_isil_audio_encode_chan_processing(opened_logic_chan_ed);
                } else {
                    reset_isil_audio_decode_chan_processing(opened_logic_chan_ed);
                }
            }

            spin_lock_irqsave(&opened_logic_chan_table->lock, flags1);
            spin_lock_irqsave(&logic_chan_table->lock, flags2);
        }
    }
    spin_unlock_irqrestore(&logic_chan_table->lock, flags2);
    spin_unlock_irqrestore(&opened_logic_chan_table->lock, flags1);
}

/****************************************************************************************************************
 * Function Name :   audio_module_reset_postprocessing
 * Description   :   audio module reset postprocessing.
 * Input         :   param1 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_module_reset_postprocessing(isil_chip_t *chip)
{
    isil_register_table_t   *opened_logic_chan_table, *logic_chan_table;
    isil_register_node_t    *register_node;
    struct list_head        *list;
    ed_tcb_t                *opened_logic_chan_ed, *logic_chan_ed;
    isil_audio_driver_t     *audio_chan_driver;
    chip_audio_t            *chip_audio;    
    unsigned long           flags1, flags2;

    chip_audio = &chip->chip_audio;
    opened_logic_chan_table = &chip_audio->opened_logic_chan_table;
    logic_chan_table = &chip_audio->logic_chan_table;
    spin_lock_irqsave(&opened_logic_chan_table->lock, flags1);
    spin_lock_irqsave(&logic_chan_table->lock, flags2);
    if(logic_chan_table->op->get_curr_register_node_number(logic_chan_table)){
        list_for_each(list, &logic_chan_table->register_header){
            register_node = to_get_register_node(list);
            spin_unlock_irqrestore(&logic_chan_table->lock, flags2);
            spin_unlock_irqrestore(&opened_logic_chan_table->lock, flags1);

            logic_chan_ed = to_get_endpoint_tcb_with_register_node(register_node);
            audio_chan_driver = to_isil_audio_chan_driver_with_logic_chan_ed(logic_chan_ed);
            if(isil_audio_chan_is_in_robust_processing(audio_chan_driver) == 1){
                opened_logic_chan_ed = &audio_chan_driver->opened_logic_chan_ed;
                if(audio_chan_driver->audio_logic_chan_id < ISIL_AUDIO_IN_CHAN_NUMBER){
                    reset_isil_audio_encode_chan_postprocessing(opened_logic_chan_ed);
                } else {
                    reset_isil_audio_decode_chan_postprocessing(opened_logic_chan_ed);
                }
            }

            spin_lock_irqsave(&opened_logic_chan_table->lock, flags1);
            spin_lock_irqsave(&logic_chan_table->lock, flags2);
        }
    }
    spin_unlock_irqrestore(&logic_chan_table->lock, flags2);
    spin_unlock_irqrestore(&opened_logic_chan_table->lock, flags1);
}

#if defined(PCI_PLATFORM)
/****************************************************************************************************************
 * Function Name :   jpeg_module_reset_preprocessing
 * Description   :   jpeg module reset preprocessing.
 * Input         :   param1 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void jpeg_module_reset_preprocessing(isil_chip_t *chip)
{
    if(chip){
        isil_vj_bus_t *vj_bus;
        isil_jpeg_phy_video_slot_t *phy_video_slot;

        vj_bus = &chip->chip_vj_bus;
        for(phy_video_slot = vj_bus->phy_video_slot; (phy_video_slot - vj_bus->phy_video_slot) < ISIL_PHY_VJ_CHAN_NUMBER; phy_video_slot++){
            phy_video_slot->op->touch_each_register_logic_chan(phy_video_slot, (void (*)(void *))reset_isil_jpeg_encode_chan_preprocessing);
        }
    }
}

/****************************************************************************************************************
 * Function Name :   jpeg_module_reset_processing
 * Description   :   jpeg module reset processing.
 * Input         :   param1 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void jpeg_module_reset_processing(isil_chip_t *chip)
{
    if(chip){
        isil_vj_bus_t *vj_bus;
        isil_jpeg_phy_video_slot_t *phy_video_slot;

        vj_bus = &chip->chip_vj_bus;
        for(phy_video_slot = vj_bus->phy_video_slot; (phy_video_slot - vj_bus->phy_video_slot) < ISIL_PHY_VJ_CHAN_NUMBER; phy_video_slot++){
            phy_video_slot->op->touch_each_register_logic_chan(phy_video_slot, (void (*)(void *))reset_isil_jpeg_encode_chan_processing);
        }
    }
}

/****************************************************************************************************************
 * Function Name :   jpeg_module_reset_postprocessing
 * Description   :   jpeg module reset postprocessing.
 * Input         :   param1 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void jpeg_module_reset_postprocessing(isil_chip_t *chip)
{
    if(chip){
        isil_vj_bus_t *vj_bus;
        isil_jpeg_phy_video_slot_t *phy_video_slot;

        vj_bus = &chip->chip_vj_bus;
        for(phy_video_slot = vj_bus->phy_video_slot; (phy_video_slot - vj_bus->phy_video_slot) < ISIL_PHY_VJ_CHAN_NUMBER; phy_video_slot++){
            phy_video_slot->op->touch_each_register_logic_chan(phy_video_slot, (void (*)(void *))reset_isil_jpeg_encode_chan_postprocessing);
        }
    }
}

#ifdef  V4L2_PREVIEW_MODULE
/****************************************************************************************************************
 * Function Name :   prev_module_reset_preprocessing
 * Description   :   preview module reset preprocessing.
 * Input         :   param1 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void prev_module_reset_preprocessing(isil_chip_t *chip)
{
    int i;
    ed_tcb_t    *ed_tcb;
    isil_prev_logic_chan_t *chan;

    if (chip == NULL)
        return;

    for (i = 0; i < ISIL_PHY_VP_CHAN_NUMBER; i++) {
        chan = &chip->prev_chan[i];
        if (chan->enable) {
            ed_tcb = &chan->opened_logic_chan_ed;
            reset_isil_prev_chan_preprocessing(ed_tcb);
        }
    }
}

/****************************************************************************************************************
 * Function Name :   prev_module_reset_processing
 * Description   :   preview module reset processing.
 * Input         :   param1 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void prev_module_reset_processing(isil_chip_t *chip)
{
    int i;
    ed_tcb_t    *ed_tcb;
    isil_prev_logic_chan_t *chan;

    if (chip == NULL)
        return;

    for (i = 0; i < ISIL_PHY_VP_CHAN_NUMBER; i++) {
        chan = &chip->prev_chan[i];
        if (chan->enable) {
            ed_tcb = &chan->opened_logic_chan_ed;
            reset_isil_prev_chan_processing(ed_tcb);
        }
    }
}

/****************************************************************************************************************
 * Function Name :   prev_module_reset_postprocessing
 * Description   :   preview module reset postprocessing.
 * Input         :   param1 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void prev_module_reset_postprocessing(isil_chip_t *chip)
{
    int i;
    ed_tcb_t    *ed_tcb;
    isil_prev_logic_chan_t *chan;

    if (chip == NULL)
        return;

    for (i = 0; i < ISIL_PHY_VP_CHAN_NUMBER; i++) {
        chan = &chip->prev_chan[i];
        if (chan->enable) {
            ed_tcb = &chan->opened_logic_chan_ed;
            reset_isil_prev_chan_postprocessing(ed_tcb);
        }
    }
}
#endif
#endif

/****************************************************************************************************************
 * Function Name :   process_chip_reset_req
 * Description   :   chip reset processing.
 * Input         :   param1 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void process_chip_reset_req(isil_chip_t *chip)
{
    if(chip != NULL){

        //chip's reset pre-processing: include h264_encode, audio_encode, audio_decode, mjpeg_encode, preview
        h264_module_reset_preprocessing(chip);
        audio_module_reset_preprocessing(chip);
#if defined(PCI_PLATFORM)
        jpeg_module_reset_preprocessing(chip);
#ifdef  V4L2_PREVIEW_MODULE
        prev_module_reset_preprocessing(chip);
#endif
#endif

        //chip's reset precossing: include h264_encode, audio_encode, audio_decode, mjpeg_encode, preview
        h264_module_reset_processing(chip);
        audio_module_reset_processing(chip);
#if defined(PCI_PLATFORM)
        jpeg_module_reset_processing(chip);
#ifdef  V4L2_PREVIEW_MODULE
        prev_module_reset_processing(chip);
#endif
#endif

        if(chip->chip_set_valid != NULL){
            chip->chip_set_valid(chip);
        }

        //chip's reset post-precossing: include h264_encode, audio_encode, audio_decode, mjpeg_encode, preview
        h264_module_reset_postprocessing(chip);
        audio_module_reset_postprocessing(chip);
#if defined(PCI_PLATFORM)
        jpeg_module_reset_postprocessing(chip);
#ifdef  V4L2_PREVIEW_MODULE
        prev_module_reset_postprocessing(chip);
        isil_chip_prev_robust_process_done(&chip->chip_vp_bus);
#endif
        isil_chip_jpeg_robust_process_done(&chip->chip_vj_bus);
#endif
        isil_chip_audio_robust_process_done(&chip->chip_audio);
        chip_robust_process_done(chip);
    }
}


/****************************************************************************************************************
 * Function Name :   process_chip_video_standard_changed
 * Description   :   chip video signal standard change processing.
 * Input         :   param1 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void process_chip_video_standard_changed(isil_chip_t *chip)
{
    if(chip){
        isil_chip_vi_driver_t *vi_driver;
        isil_video_bus_t *video_bus;
        enum CHIP_VI_STANDARD video_standard;

        vi_driver = chip->chip_vi_driver;
        video_bus = &chip->chip_driver->video_bus;
        video_standard = vi_driver->op->get_hw_video_standard(vi_driver);

        vi_driver->op->set_video_standard(vi_driver, video_standard);
        if((video_standard == CHIP_VI_STANDARD_NTSC_M) || (video_standard == CHIP_VI_STANDARD_NTSC443)){
            if(video_bus->op->get_video_standard(video_bus) != ISIL_VIDEO_STANDARD_NTSC){
                video_bus->op->set_video_standard(video_bus, ISIL_VIDEO_STANDARD_NTSC);
            }
        }else{
            if(video_bus->op->get_video_standard(video_bus) != ISIL_VIDEO_STANDARD_PAL){
                video_bus->op->set_video_standard(video_bus, ISIL_VIDEO_STANDARD_PAL);
            }
        }
    }
}

/****************************************************************************************************************
 * Function Name :   kthread_algo_agent
 * Description   :   kthread
 * Input         :   param1 is void ptr
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static int	kthread_algo_agent(void *data)
{
    isil_chip_kthread_interface_t   *chip_kthread_interface = (isil_chip_kthread_interface_t*)data;
    isil_kthread_msg_pool_t   *isil_msg_pool;
    isil_kthread_msg_queue_t  *isil_msg_queue;
    int	escape = 0;

    get_msg_pool_header(chip_kthread_interface, &isil_msg_pool);
    get_msg_queue_header(chip_kthread_interface, &isil_msg_queue);
    atomic_set(&chip_kthread_interface->algo_running_flag, ISIL_KTHREAD_RUNNING);
    while(1){
        isil_msg_queue->op->get_curr_consumer_from_queue(isil_msg_queue);
        if(isil_msg_queue->curr_consumer != NULL) {
            switch(isil_msg_queue->curr_consumer->msg_type){
                case REQ_ALGO_CHIP_RESET:
                    process_chip_reset_req((isil_chip_t*)isil_msg_queue->curr_consumer->msg_context);
                    break;
                case REQ_VIDEO_STANDARD_CHANGED:
                    process_chip_video_standard_changed((isil_chip_t*)isil_msg_queue->curr_consumer->msg_context);
                    break;
                case SYSTEM_ESCAPE_MSG:
                    escape = 1;
                    break;
                default:
                    ISIL_DBG(ISIL_DBG_INFO, ": we don't know's msg type = %d\n", isil_msg_queue->curr_consumer->msg_type);
                    break;
            }
        }
        isil_msg_queue->op->release_curr_consumer(isil_msg_queue);
        if(escape){
            break;
        }
    }
    atomic_set(&chip_kthread_interface->algo_running_flag, ISIL_KTHREAD_EXIT);
    return 0;
}

/****************************************************************************************************************
 * Function Name :   isil_chip_kthread_interface_init_isil_kthread_manage
 * Description   :   init all resource of kthread
 * Input         :   null
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void isil_chip_kthread_interface_init_isil_kthread_manage(isil_chip_kthread_interface_t *chip_kthread_interface)
{
    if(chip_kthread_interface != NULL){
        isil_kthread_msg_pool_t   *isil_msg_pool;
        isil_kthread_msg_queue_t  *isil_msg_queue;

        chip_kthread_interface->algo_kthread=NULL;
        atomic_set(&chip_kthread_interface->algo_running_flag, ISIL_KTHREAD_CLOSE);
        get_msg_pool_header(chip_kthread_interface, &isil_msg_pool);
        if(init_isil_kthread_msg_pool(isil_msg_pool) != 0){
            ISIL_DBG(ISIL_DBG_INFO, ": create algo msg pool err\n");
            return;
        }
        get_msg_queue_header(chip_kthread_interface, &isil_msg_queue);
        init_isil_kthread_msg_queue(isil_msg_queue);
        atomic_set(&chip_kthread_interface->algo_running_flag, ISIL_KTHREAD_PREPARE);
    }
}

/****************************************************************************************************************
 * Function Name :   isil_chip_kthread_interface_remove_isil_kthread_manage
 * Description   :   remove all resource of kthread
 * Input         :   null
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void	isil_chip_kthread_interface_remove_isil_kthread_manage(isil_chip_kthread_interface_t *chip_kthread_interface)
{
    if(chip_kthread_interface != NULL){
        if(atomic_read(&chip_kthread_interface->algo_running_flag) >= ISIL_KTHREAD_PREPARE){
            isil_kthread_msg_pool_t   *isil_msg_pool;
            isil_kthread_msg_queue_t  *isil_msg_queue;

            get_msg_queue_header(chip_kthread_interface, &isil_msg_queue);
            remove_isil_kthread_msg_queue(isil_msg_queue);
            get_msg_pool_header(chip_kthread_interface, &isil_msg_pool);
            remove_isil_kthread_msg_pool(isil_msg_pool);
            chip_kthread_interface->algo_kthread=NULL;
            atomic_set(&chip_kthread_interface->algo_running_flag, ISIL_KTHREAD_CLOSE);
        }
    }
}

/****************************************************************************************************************
 * Function Name :   isil_chip_kthread_interface_notify_algo_kthread_kill_myself
 * Description   :   notify himself thread will escape
 * Input         :   null
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void isil_chip_kthread_interface_notify_algo_kthread_kill_myself(isil_chip_kthread_interface_t *chip_kthread_interface)
{
    if(chip_kthread_interface != NULL){
        if(atomic_read(&chip_kthread_interface->algo_running_flag) >= ISIL_KTHREAD_STANDBY){
            isil_kthread_msg_pool_t   *isil_msg_pool;
            isil_kthread_msg_queue_t  *isil_msg_queue;
            isil_kthread_msg_t		  *isil_msg;

            get_msg_queue_header(chip_kthread_interface, &isil_msg_queue);
            get_msg_pool_header(chip_kthread_interface, &isil_msg_pool);
            isil_msg_pool->op->get(isil_msg_pool, &isil_msg);
            if(isil_msg != NULL){
                isil_msg->msg_type = SYSTEM_ESCAPE_MSG;
                isil_msg_queue->op->put(isil_msg_queue, isil_msg);
            }
        }
    }
}

/****************************************************************************************************************
 * Function Name :   isil_chip_kthread_interface_wait_10_second_for_algo_kthread
 * Description   :   maxim wait 10 second for kthread escape
 * Input         :   null
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void isil_chip_kthread_interface_wait_10_second_for_algo_kthread(isil_chip_kthread_interface_t *chip_kthread_interface)
{
    if(chip_kthread_interface != NULL){
        if(atomic_read(&chip_kthread_interface->algo_running_flag) >= ISIL_KTHREAD_STANDBY){
            int count=0;
            while(atomic_read(&chip_kthread_interface->algo_running_flag) != ISIL_KTHREAD_EXIT){
                msleep(10);
                count++;
                if(count > 1000){
                    break;
                }
            }
        }
    }
}

/****************************************************************************************************************
 * Function Name :   force_close_kthread
 * Description   :   force kill kthread
 * Input         :   param1 is task 
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void isil_chip_kthread_interface_force_close_kthread(isil_chip_kthread_interface_t *chip_kthread_interface)
{
    if(chip_kthread_interface != NULL){
        if(atomic_read(&chip_kthread_interface->algo_running_flag) >= ISIL_KTHREAD_STANDBY){
            if(atomic_read(&chip_kthread_interface->algo_running_flag) != ISIL_KTHREAD_EXIT){
                kthread_stop(chip_kthread_interface->algo_kthread);
                atomic_set(&chip_kthread_interface->algo_running_flag, ISIL_KTHREAD_PREPARE);
            }
        }
    }
}

/****************************************************************************************************************
 * Function Name :   start_isil_kthread
 * Description   :   start kthread
 * Input         :   null
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void isil_chip_kthread_interface_start_isil_kthread(isil_chip_kthread_interface_t *chip_kthread_interface)
{
    if(chip_kthread_interface != NULL){
        if(atomic_read(&chip_kthread_interface->algo_running_flag) == ISIL_KTHREAD_PREPARE){
            memset(chip_kthread_interface->name, 0, 64);
            sprintf(chip_kthread_interface->name, "isil_%d", chip_kthread_interface->chip->chip_id);
            atomic_set(&chip_kthread_interface->algo_running_flag, ISIL_KTHREAD_STANDBY);
            chip_kthread_interface->algo_kthread = kthread_run(kthread_algo_agent, chip_kthread_interface, chip_kthread_interface->name);
            if(IS_ERR(chip_kthread_interface->algo_kthread)) {
                ISIL_DBG(ISIL_DBG_INFO, ": start isil_algo_kthread err\n");
                atomic_set(&chip_kthread_interface->algo_running_flag, ISIL_KTHREAD_PREPARE);
            }
        }
    }
}

/****************************************************************************************************************
 * Function Name :   stop_isil_kthread
 * Description   :   stop kthread
 * Input         :   null
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void isil_chip_kthread_interface_stop_isil_kthread(isil_chip_kthread_interface_t *chip_kthread_interface)
{
    if(chip_kthread_interface->algo_kthread != NULL) {
        if(atomic_read(&chip_kthread_interface->algo_running_flag) >= ISIL_KTHREAD_STANDBY){
            chip_kthread_interface->op->notify_algo_kthread_kill_myself(chip_kthread_interface);
            chip_kthread_interface->op->wait_10_second_for_algo_kthread(chip_kthread_interface);
            if((atomic_read(&chip_kthread_interface->algo_running_flag) != ISIL_KTHREAD_EXIT)){
                chip_kthread_interface->op->force_close_kthread(chip_kthread_interface);
                ISIL_DBG(ISIL_DBG_INFO, ": algo_kthread myself kill err, so force kill it\n");
            } else {
                atomic_set(&chip_kthread_interface->algo_running_flag, ISIL_KTHREAD_PREPARE);
            }
        }
    }
}

static struct isil_chip_kthread_interface_operation isil_chip_kthread_interface_op = {
    .init_isil_kthread_manage = isil_chip_kthread_interface_init_isil_kthread_manage,
    .remove_isil_kthread_manage = isil_chip_kthread_interface_remove_isil_kthread_manage,
    .notify_algo_kthread_kill_myself = isil_chip_kthread_interface_notify_algo_kthread_kill_myself,
    .wait_10_second_for_algo_kthread = isil_chip_kthread_interface_wait_10_second_for_algo_kthread,
    .force_close_kthread = isil_chip_kthread_interface_force_close_kthread,
    .start_isil_kthread = isil_chip_kthread_interface_start_isil_kthread,
    .stop_isil_kthread = isil_chip_kthread_interface_stop_isil_kthread,
};

void    init_isil_chip_kthread_interface(isil_chip_t *chip)
{
    isil_chip_kthread_interface_t   *chip_kthread_interface;

    chip_kthread_interface = &chip->chip_kthread_interface;
    chip_kthread_interface->chip = chip;
    chip_kthread_interface->op = &isil_chip_kthread_interface_op;
    chip_kthread_interface->op->init_isil_kthread_manage(chip_kthread_interface);
}

void	remove_isil_chip_kthread_interface(isil_chip_t *chip)
{
    isil_chip_kthread_interface_t   *chip_kthread_interface;

    chip_kthread_interface = &chip->chip_kthread_interface;
    if(chip_kthread_interface->op != NULL){
        chip_kthread_interface->op->remove_isil_kthread_manage(chip_kthread_interface);
    }
}

void    start_isil_chip_kthread_interface(isil_chip_t *chip)
{
    isil_chip_kthread_interface_t   *chip_kthread_interface;

    chip_kthread_interface = &chip->chip_kthread_interface;
    if(chip_kthread_interface->op != NULL){
        chip_kthread_interface->op->start_isil_kthread(chip_kthread_interface);
    }
}

void    stop_isil_chip_kthread_interface(isil_chip_t *chip)
{
    isil_chip_kthread_interface_t   *chip_kthread_interface;

    chip_kthread_interface = &chip->chip_kthread_interface;
    if(chip_kthread_interface->op != NULL){
        chip_kthread_interface->op->stop_isil_kthread(chip_kthread_interface);
    }
}

