/***************************************** (C) COPYRIGHT 2010 Intersil *****************************************
 * File Name         : isil_audio_iop.c
 * Author            : tang.sc@intersil.com
 * Revision          : 1.2.1.1275
 * Date              : 2012-03-29
 * Description       : audio driver interface with hardware
 * Change log        : 
 *                   
 ****************************************************************************************************************
 * THE PRESENT SOFTWARE IS DISTRIBUTED IN THE HOPE THAT IT WILL PROVIDE CUSTOMERS 
 * WITH CODING INFORMATION TO SPEED UP THEIR PRODUCT DEVELOPMENT, BUT WITHOUT
 * ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF MERCHANTABILITY OR FITNESS
 * FOR A PARTICULAR PURPOSE. AS A RESULT, TECHWELL SHALL NOT BE HELD LIABLE FOR ANY
 * DIRECT, INDIRECT OR CONSEQUENTIAL DAMAGES WITH RESPECT TO ANY CLAIMS ARISING
 * FROM THE CONTENT OF SUCH SOFTWARE AND/OR THE USE MADE BY CUSTOMERS OF THE
 * CODING INFORMATION CONTAINED HEREIN IN CONNECTION WITH THEIR PRODUCTS.
 ****************************************************************************************************************/
#include	<isil5864/isil_common.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE > KERNEL_VERSION(4,11,0)
#include <linux/sched/signal.h>
#else
#include <linux/sched.h>
#endif
isil_audio_config_descriptor_t    default_audio_encode_config_descriptor = {
    .type = ISIL_AUDIO_ADPCM_32K,
    .sample_rate = ISIL_AUDIO_8000,
    .bit_wide = ISIL_AUDIO_16BIT,
};

isil_audio_config_descriptor_t    default_audio_decode_config_descriptor = {
    .type = ISIL_AUDIO_ADPCM_32K,
    .sample_rate = ISIL_AUDIO_8000,
    .bit_wide = ISIL_AUDIO_16BIT,
};

/****************************************************************************************************************
 * Function Name :   get_power_base
 * Description   :   this function is finding 2's power base.
 * Input         :   param1 is value
 * Output        :   null
 * Return        :   find value base.
 ****************************************************************************************************************/
int  get_power_base(u32 value)
{
    int position, number;
    position = number = 0;
    while(value){
        if(value&1){
            number++;
        }
        position++;
        value >>= 1;
    }
    if(number==1){
        position--;
    }
    return position;
}

/****************************************************************************************************************
 * Function Name :   audio_section_descriptor_init_audio_descriptor
 * Description   :   update audio descriptor's position.
 * Input         :   param1 is audio descriptor management struct,
 *                   param2 is audio descriptor's position
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_section_descriptor_init_audio_descriptor(isil_audio_descriptor_t *audio_descriptor, void *descriptor_addr)
{
    if(descriptor_addr != NULL){
        audio_descriptor->descriptor = (audio_section_descriptor_t*)descriptor_addr;
    } else {
        audio_descriptor->descriptor = NULL;
    }
}

/****************************************************************************************************************
 * Function Name :   audio_section_descriptor_get_audio_descriptor_timestamp
 * Description   :   get audio data's timestamp.
 * Input         :   param1 is audio descriptor management struct,
 * Output        :   null
 * Return        :   timestamp
 ****************************************************************************************************************/
static u32  audio_section_descriptor_get_audio_descriptor_timestamp(isil_audio_descriptor_t *audio_descriptor)
{
    if(audio_descriptor->descriptor != NULL){
        return audio_descriptor->descriptor->timestamp;
    } else {
        return 0;
    }
}

/****************************************************************************************************************
 * Function Name :   audio_section_descriptor_set_audio_descriptor_timestamp
 * Description   :   update audio data's timestamp.
 * Input         :   param1 is audio descriptor management struct,
 *                   param2 is timestamp
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_section_descriptor_set_audio_descriptor_timestamp(isil_audio_descriptor_t *audio_descriptor, u32 timestamp)
{
    if(audio_descriptor->descriptor != NULL){
        audio_descriptor->descriptor->timestamp = timestamp;
    }
}

/****************************************************************************************************************
 * Function Name :   audio_section_descriptor_get_audio_descriptor_sample_rate
 * Description   :   get audio data's sample rate.
 * Input         :   param1 is audio descriptor management struct,
 * Output        :   null
 * Return        :   audio's sample rate
 ****************************************************************************************************************/
static int  audio_section_descriptor_get_audio_descriptor_sample_rate(isil_audio_descriptor_t *audio_descriptor)
{
    if(audio_descriptor->descriptor != NULL){
        return audio_descriptor->descriptor->sample_rate;
    } else {
        return ISIL_AUDIO_8000;
    }
}

/****************************************************************************************************************
 * Function Name :   audio_section_descriptor_set_audio_descriptor_sample_rate
 * Description   :   set audio data's sample rate.
 * Input         :   param1 is audio descriptor management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_section_descriptor_set_audio_descriptor_sample_rate(isil_audio_descriptor_t *audio_descriptor, int sample_rate)
{
    if(audio_descriptor->descriptor != NULL){
        audio_descriptor->descriptor->sample_rate = sample_rate;
    }
}

/****************************************************************************************************************
 * Function Name :   audio_section_descriptor_get_audio_descriptor_type
 * Description   :   get audio data's audio type.
 * Input         :   param1 is audio descriptor management struct,
 * Output        :   null
 * Return        :   audio type
 ****************************************************************************************************************/
static int  audio_section_descriptor_get_audio_descriptor_type(isil_audio_descriptor_t *audio_descriptor)
{
    if(audio_descriptor->descriptor != NULL){
        return audio_descriptor->descriptor->type;
    } else {
        return ISIL_AUDIO_ADPCM_32K;
    }
}

/****************************************************************************************************************
 * Function Name :   audio_section_descriptor_set_audio_descriptor_type
 * Description   :   set audio data's audio type.
 * Input         :   param1 is audio descriptor management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_section_descriptor_set_audio_descriptor_type(isil_audio_descriptor_t *audio_descriptor, int type)
{
    if(audio_descriptor->descriptor != NULL){
        audio_descriptor->descriptor->type= type;
    }
}

/****************************************************************************************************************
 * Function Name :   audio_section_descriptor_get_audio_descriptor_bit_wide
 * Description   :   get audio data's bit_wide.
 * Input         :   param1 is audio descriptor management struct,
 * Output        :   null
 * Return        :   bit_wide
 ****************************************************************************************************************/
static int  audio_section_descriptor_get_audio_descriptor_bit_wide(isil_audio_descriptor_t *audio_descriptor)
{
    if(audio_descriptor->descriptor != NULL){
        return audio_descriptor->descriptor->bit_wide;
    } else {
        return ISIL_AUDIO_16BIT;
    }
}

/****************************************************************************************************************
 * Function Name :   audio_section_descriptor_get_audio_descriptor_bit_wide
 * Description   :   get audio data's bit_wide.
 * Input         :   param1 is audio descriptor management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_section_descriptor_set_audio_descriptor_bit_wide(isil_audio_descriptor_t *audio_descriptor, int bit_wide)
{
    if(audio_descriptor->descriptor != NULL){
        audio_descriptor->descriptor->bit_wide = bit_wide;
    }
}

/****************************************************************************************************************
 * Function Name :   audio_section_descriptor_get_audio_descriptor_valid_len
 * Description   :   get audio data's valid len.
 * Input         :   param1 is audio descriptor management struct,
 * Output        :   null
 * Return        :   valid len
 ****************************************************************************************************************/
static u32  audio_section_descriptor_get_audio_descriptor_valid_len(isil_audio_descriptor_t *audio_descriptor)
{
    if(audio_descriptor->descriptor != NULL){
        return audio_descriptor->descriptor->valid_len;
    } else {
        return 0;
    }
}

/****************************************************************************************************************
 * Function Name :   audio_section_descriptor_set_audio_descriptor_valid_len
 * Description   :   set audio data's valid len.
 * Input         :   param1 is audio descriptor management struct,
 *                   param2 is valid len
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_section_descriptor_set_audio_descriptor_valid_len(isil_audio_descriptor_t *audio_descriptor, u32 valid_len)
{
    if(audio_descriptor->descriptor != NULL){
        audio_descriptor->descriptor->valid_len = valid_len;
    }
}

/****************************************************************************************************************
 * Veriable Name :   audio_section_descriptor_op
 * Description   :   for helping op audio descriptor, we define general interface
 ****************************************************************************************************************/
static struct isil_audio_section_descriptor_operation audio_section_descriptor_op = {
    .init_audio_descriptor = audio_section_descriptor_init_audio_descriptor,
    .get_audio_descriptor_timestamp = audio_section_descriptor_get_audio_descriptor_timestamp,
    .set_audio_descriptor_timestamp = audio_section_descriptor_set_audio_descriptor_timestamp,
    .get_audio_descriptor_sample_rate = audio_section_descriptor_get_audio_descriptor_sample_rate,
    .set_audio_descriptor_sample_rate = audio_section_descriptor_set_audio_descriptor_sample_rate,
    .get_audio_descriptor_type = audio_section_descriptor_get_audio_descriptor_type,
    .set_audio_descriptor_type = audio_section_descriptor_set_audio_descriptor_type,
    .get_audio_descriptor_bit_wide = audio_section_descriptor_get_audio_descriptor_bit_wide,
    .set_audio_descriptor_bit_wide = audio_section_descriptor_set_audio_descriptor_bit_wide,
    .get_audio_descriptor_valid_len = audio_section_descriptor_get_audio_descriptor_valid_len,
    .set_audio_descriptor_valid_len = audio_section_descriptor_set_audio_descriptor_valid_len,
};

/****************************************************************************************************************
 * Function Name :   init_isil_audio_descriptor
 * Description   :   init audio descriptor.
 * Input         :   param1 is audio descriptor management struct,
 *                   param2 is valid len
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void	init_isil_audio_descriptor(isil_audio_descriptor_t *audio_descriptor, void *audio_descriptor_addr)
{
    if(audio_descriptor != NULL){
        audio_descriptor->op = &audio_section_descriptor_op;
        audio_descriptor->op->init_audio_descriptor(audio_descriptor, audio_descriptor_addr);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_encode_config_param_update_init
 * Description   :   audio property param init
 * Input         :   param1 is audio's running property param
 *                   param2 is audio's config property param
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_encode_config_param_update_init(isil_audio_param_t *running_param, isil_audio_config_descriptor_t *param)
{
    if(param == NULL){
        param = &default_audio_encode_config_descriptor;
    }
    running_param->property.type = param->type;
    running_param->property.sample_rate = param->sample_rate;
    running_param->property.bit_wide = param->bit_wide;
}

/****************************************************************************************************************
 * Function Name :   audio_encode_config_param_update_update_sample_rate
 * Description   :   update sample rate
 * Input         :   param1 is audio's running property param
 *                   param2 is audio's sample rate
 * Output        :   null
 * Return        :   ISIL_OK is meaning update ok, other is fail
 ****************************************************************************************************************/
static int  audio_encode_config_param_update_update_sample_rate(isil_audio_param_t *running_param, int sample_rate)
{
    chip_audio_encode_t     *audio_encode;
    audio_encode_control_t  *audio_control;
    chip_audio_t    *chip_audio;
    isil_chip_t     *chip;

    if((sample_rate!=ISIL_AUDIO_8000) && (sample_rate!=ISIL_AUDIO_16000)){
        ISIL_DBG(ISIL_DBG_ERR, "only support 8K, 16K\n");
        return ISIL_ERR;
    }
    running_param->property.sample_rate = sample_rate;
    audio_encode = to_get_chip_audio_encode_with_running_param(running_param);
    chip_audio = to_get_chip_audio_with_audio_encode(audio_encode);
    chip = chip_audio->chip;
    audio_control = &audio_encode->audio_control;
    if(audio_control->op != NULL){
        audio_control->op->set_audio_sample_rate(audio_control, sample_rate, chip);
    }
    return ISIL_OK;
}

/****************************************************************************************************************
 * Function Name :   audio_encode_config_param_update_get_sample_rate
 * Description   :   get sample rate
 * Input         :   param1 is audio's running property param
 * Output        :   null
 * Return        :   sample rate
 ****************************************************************************************************************/
static int  audio_encode_config_param_update_get_sample_rate(isil_audio_param_t *running_param)
{
    return running_param->property.sample_rate;
}

/****************************************************************************************************************
 * Function Name :   audio_encode_config_param_update_update_bit_wide
 * Description   :   update bit_wide
 * Input         :   param1 is audio's running property param
 *                   param2 is bit_wide
 * Output        :   null
 * Return        :   ISIL_OK is meaning update ok, other is fail
 ****************************************************************************************************************/
static int  audio_encode_config_param_update_update_bit_wide(isil_audio_param_t *running_param, int bit_wide)
{
    chip_audio_encode_t     *audio_encode;
    audio_encode_control_t  *audio_control;
    chip_audio_t    *chip_audio;
    isil_chip_t     *chip;

    if((bit_wide!=ISIL_AUDIO_16BIT) && (bit_wide!=ISIL_AUDIO_8BIT)){
        ISIL_DBG(ISIL_DBG_ERR, "only support 8bit or 16bit\n");
        return ISIL_ERR;
    }
    running_param->property.bit_wide = bit_wide;
    audio_encode = to_get_chip_audio_encode_with_running_param(running_param);
    chip_audio = to_get_chip_audio_with_audio_encode(audio_encode);
    chip = chip_audio->chip;
    audio_control = &audio_encode->audio_control;
    if(audio_control->op != NULL){
        audio_control->op->set_bit_wide(audio_control, bit_wide, chip);
    }
    return ISIL_OK;
}

/****************************************************************************************************************
 * Function Name :   audio_encode_config_param_update_get_bit_wide
 * Description   :   get bit wide
 * Input         :   param1 is audio's running property param
 * Output        :   null
 * Return        :   bit wide
 ****************************************************************************************************************/
static int  audio_encode_config_param_update_get_bit_wide(isil_audio_param_t *running_param)
{
    return running_param->property.bit_wide;
}

/****************************************************************************************************************
 * Function Name :   audio_encode_config_param_update_update_type
 * Description   :   update audio encode type
 * Input         :   param1 is audio's running property param
 *                   param2 is type
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static int  audio_encode_config_param_update_update_type(isil_audio_param_t *running_param, int type)
{
    chip_audio_encode_t     *audio_encode;
    audio_encode_control_t  *audio_control;
    chip_audio_t    *chip_audio;
    isil_chip_t     *chip;

    if((type != ISIL_AUDIO_ADPCM_32K) && (type != ISIL_AUDIO_PCM)){
        ISIL_DBG(ISIL_DBG_ERR, "only support ISIL_AUDIO_ADPCM_32K and ISIL_AUDIO_PCM\n");
        return ISIL_ERR;
    }
    running_param->property.type = type;
    audio_encode = to_get_chip_audio_encode_with_running_param(running_param);
    chip_audio = to_get_chip_audio_with_audio_encode(audio_encode);
    chip = chip_audio->chip;
    audio_control = &audio_encode->audio_control;
    if(audio_control->op != NULL){
        audio_control->op->set_audio_type(audio_control, type, chip);
    }
    return ISIL_OK;
}

/****************************************************************************************************************
 * Function Name :   audio_encode_config_param_update_get_type
 * Description   :   get type
 * Input         :   param1 is audio's running property param
 * Output        :   null
 * Return        :   type
 ****************************************************************************************************************/
static int  audio_encode_config_param_update_get_type(isil_audio_param_t *running_param)
{
    return running_param->property.type;
}

/****************************************************************************************************************
 * Veriable Name :   audio_encode_running_param_update_op
 * Description   :   for helping op audio running param, we define general interface
 ****************************************************************************************************************/
static struct audio_property_update_operation   audio_encode_running_param_update_op = {
    .init = audio_encode_config_param_update_init,
    .update_sample_rate = audio_encode_config_param_update_update_sample_rate,
    .get_sample_rate = audio_encode_config_param_update_get_sample_rate,
    .update_bit_wide = audio_encode_config_param_update_update_bit_wide,
    .get_bit_wide = audio_encode_config_param_update_get_bit_wide,
    .update_type = audio_encode_config_param_update_update_type,
    .get_type = audio_encode_config_param_update_get_type,
};

/****************************************************************************************************************
 * Function Name :   init_audio_encode_property
 * Description   :   init audio encode property
 * Input         :   param1 is audio's running property param
 *                   param2 is audio config descriptor
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void    init_audio_encode_property(isil_audio_param_t *running_param, isil_audio_config_descriptor_t *config)
{
    if(running_param != NULL){
        running_param->op = &audio_encode_running_param_update_op;
        running_param->op->init(running_param, config);
    }
}

/****************************************************************************************************************
 * Function Name :   init_audio_decode_property
 * Description   :   init audio decode property
 * Input         :   param1 is audio's running property param
 *                   param2 is audio config descriptor
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_decode_config_param_update_init(isil_audio_param_t *running_param, isil_audio_config_descriptor_t *param)
{
    if(param == NULL){
        param = &default_audio_decode_config_descriptor;
    }
    if(running_param->op != NULL){
        running_param->op->update_bit_wide(running_param, (int)param->bit_wide);
        running_param->op->update_sample_rate(running_param, (int)param->sample_rate);
        running_param->op->update_type(running_param, (int)param->type);
    } else {
        running_param->property.type = param->type;
        running_param->property.sample_rate = param->sample_rate;
        running_param->property.bit_wide = param->bit_wide;
    }
}

/****************************************************************************************************************
 * Function Name :   audio_decode_config_param_update_update_sample_rate
 * Description   :   update decode sample rate
 * Input         :   param1 is audio's running property param
 *                   param2 is sample rate
 * Output        :   null
 * Return        :   ISIL_OK is meaning update ok, other is meaning fail
 ****************************************************************************************************************/
static int  audio_decode_config_param_update_update_sample_rate(isil_audio_param_t *running_param, int sample_rate)
{
    if((sample_rate!=ISIL_AUDIO_8000) && (sample_rate!=ISIL_AUDIO_16000)){
        ISIL_DBG(ISIL_DBG_ERR, "only support 8K, 16K\n");
        return ISIL_ERR;
    }
    if(running_param == NULL){
        return ISIL_ERR;
    }
    running_param->property.sample_rate = sample_rate;
    return ISIL_OK;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_config_param_update_get_sample_rate
 * Description   :   get decode sample rate
 * Input         :   param1 is audio's running property param
 * Output        :   null
 * Return        :   sample rate
 ****************************************************************************************************************/
static int  audio_decode_config_param_update_get_sample_rate(isil_audio_param_t *running_param)
{
    return running_param->property.sample_rate;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_config_param_update_update_bit_wide
 * Description   :   update decode bit wide
 * Input         :   param1 is audio's running property param
 *                   param2 is bit_wide
 * Output        :   null
 * Return        :   ISIL_OK is meaning update ok, other is meaning fail
 ****************************************************************************************************************/
static int  audio_decode_config_param_update_update_bit_wide(isil_audio_param_t *running_param, int bit_wide)
{
    chip_audio_decode_t     *audio_decode;
    audio_decode_control_t  *audio_control;
    chip_audio_t    *chip_audio;
    isil_chip_t     *chip;

    if((bit_wide!=ISIL_AUDIO_16BIT) && (bit_wide!=ISIL_AUDIO_8BIT)){
        ISIL_DBG(ISIL_DBG_ERR, "only support 8bit or 16bit\n");
        return ISIL_ERR;
    }
    running_param->property.bit_wide = bit_wide;
    audio_decode = to_get_chip_audio_decode_with_running_param(running_param);
    chip_audio = to_get_chip_audio_with_audio_decode(audio_decode);
    chip = chip_audio->chip;
    audio_control = &audio_decode->audio_control;
    if(audio_control->op != NULL){
        audio_control->op->set_decode_chan_bit_wide(audio_control, ISIL_AUDIO_OUT_SPEAKER_ID, bit_wide, chip);
        audio_control->op->set_decode_chan_bit_wide(audio_control, ISIL_AUDIO_OUT_PLAYBACK_ID, bit_wide, chip);
    }
    return ISIL_OK;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_config_param_update_get_bit_wide
 * Description   :   get decode bit wide
 * Input         :   param1 is audio's running property param
 * Output        :   null
 * Return        :   bit wide
 ****************************************************************************************************************/
static int  audio_decode_config_param_update_get_bit_wide(isil_audio_param_t *running_param)
{
    return running_param->property.bit_wide;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_config_param_update_update_type
 * Description   :   update decode type
 * Input         :   param1 is audio's running property param
 *                   param2 is type
 * Output        :   null
 * Return        :   ISIL_OK is meaning update ok, other is meaning fail
 ****************************************************************************************************************/
static int  audio_decode_config_param_update_update_type(isil_audio_param_t *running_param, int type)
{
    if((type != ISIL_AUDIO_ADPCM_32K)  && (type != ISIL_AUDIO_PCM)){
        ISIL_DBG(ISIL_DBG_ERR, "only support ISIL_AUDIO_ADPCM_32K and ISIL_AUDIO_PCM\n");
        return ISIL_ERR;
    }
    running_param->property.type = type;
    return ISIL_OK;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_config_param_update_get_type
 * Description   :   get decode type
 * Input         :   param1 is audio's running property param
 * Output        :   null
 * Return        :   type
 ****************************************************************************************************************/
static int  audio_decode_config_param_update_get_type(isil_audio_param_t *running_param)
{
    return running_param->property.type;
}

/****************************************************************************************************************
 * Veriable Name :   audio_decode_running_param_update_op
 * Description   :   for helping op audio decode running param, we define general interface
 ****************************************************************************************************************/
static struct audio_property_update_operation   audio_decode_running_param_update_op = {
    .init = audio_decode_config_param_update_init,
    .update_sample_rate = audio_decode_config_param_update_update_sample_rate,
    .get_sample_rate = audio_decode_config_param_update_get_sample_rate,
    .update_bit_wide = audio_decode_config_param_update_update_bit_wide,
    .get_bit_wide = audio_decode_config_param_update_get_bit_wide,
    .update_type = audio_decode_config_param_update_update_type,
    .get_type = audio_decode_config_param_update_get_type,
};

/****************************************************************************************************************
 * Function Name :   init_audio_decode_property
 * Description   :   init audio decode property
 * Input         :   param1 is audio's running property param
 *                   param2 is audio descriptor
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void    init_audio_decode_property(isil_audio_param_t *running_param, isil_audio_config_descriptor_t *config)
{
    if(running_param != NULL){
        running_param->op = &audio_decode_running_param_update_op;
        running_param->op->init(running_param, config);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_enable_get_adpcm_decode_state
 * Description   :   get enable adpcm decode's state
 * Input         :   param1 is audio enable control management struct
 * Output        :   null
 * Return        :   real state
 ****************************************************************************************************************/
static int  audio_enable_get_adpcm_decode_state(audio_enable_control_t *audio_en)
{
    return audio_en->value.bit_value.adpcm_decode_enable;
}

/****************************************************************************************************************
 * Function Name :   audio_enable_update_enable_adpcm_decode
 * Description   :   update enable adpcm decode's state
 * Input         :   param1 is audio enable control management struct
 *                   param2 is enable value
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_enable_update_enable_adpcm_decode(audio_enable_control_t *audio_en, int en)
{
    if(en == AUDIO_DISABLE){
        chip_audio_t        *chip_aduio = to_get_chip_audio_with_audio_enable(audio_en);
        chip_audio_decode_t *audio_decode = &chip_aduio->audio_decode;
        audio_decode_control_t  *audio_control = &audio_decode->audio_control;
        if((audio_control->value.bit_value.audio_out_chan0_enable == AUDIO_DISABLE) &&
                (audio_control->value.bit_value.audio_out_chan1_enable == AUDIO_DISABLE)){
            audio_en->value.bit_value.adpcm_decode_enable = AUDIO_DISABLE;
        }
    } else {
        audio_en->value.bit_value.adpcm_decode_enable = AUDIO_ENABLE;
    }
}

/****************************************************************************************************************
 * Function Name :   audio_enable_set_enable_adpcm_decode
 * Description   :   update enable adpcm decode's state, and update hardware reg's content
 * Input         :   param1 is audio enable control management struct
 *                   param2 is enable value
 *                   param3 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_enable_set_enable_adpcm_decode(audio_enable_control_t *audio_en, int en, isil_chip_t *chip)
{
    if(audio_en->op != NULL){
        audio_en->op->update_enable_adpcm_decode(audio_en, en);
    } else {
        audio_enable_update_enable_adpcm_decode(audio_en, en);    
    }
    if(chip != NULL){
        chip->io_op->chip_write32(chip, audio_en->reg_offset, audio_en->value.value);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_enable_get_pcm_encode_state
 * Description   :   get enable pcm encode's state
 * Input         :   param1 is audio enable control management struct
 * Output        :   null
 * Return        :   real state
 ****************************************************************************************************************/
static int  audio_enable_get_pcm_encode_state(audio_enable_control_t *audio_en)
{
    return audio_en->value.bit_value.pcm_encode_enable;
}

/****************************************************************************************************************
 * Function Name :   audio_enable_update_enable_pcm_encode
 * Description   :   update enable pcm encode's state
 * Input         :   param1 is audio enable control management struct
 *                   param2 is enable value
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_enable_update_enable_pcm_encode(audio_enable_control_t *audio_en, int en)
{
    chip_audio_t            *chip_audio = to_get_chip_audio_with_audio_enable(audio_en);
    chip_audio_encode_t     *audio_encode = &chip_audio->audio_encode;
    audio_encode_control_t  *audio_control = &audio_encode->audio_control;

    if(en == AUDIO_DISABLE){
        if((audio_control->pcm_value.bit_value.audio_chan0_en == AUDIO_DISABLE) &&
                (audio_control->pcm_value.bit_value.audio_chan1_en == AUDIO_DISABLE) &&
                (audio_control->pcm_value.bit_value.audio_chan2_en == AUDIO_DISABLE) &&
                (audio_control->pcm_value.bit_value.audio_chan3_en == AUDIO_DISABLE) &&
                (audio_control->pcm_value.bit_value.audio_chan4_en == AUDIO_DISABLE) &&
                (audio_control->pcm_value.bit_value.audio_chan5_en == AUDIO_DISABLE) &&
                (audio_control->pcm_value.bit_value.audio_chan6_en == AUDIO_DISABLE) &&
                (audio_control->pcm_value.bit_value.audio_chan7_en == AUDIO_DISABLE) &&
                (audio_control->pcm_value.bit_value.audio_chan8_en == AUDIO_DISABLE) &&
                (audio_control->pcm_value.bit_value.audio_chan9_en == AUDIO_DISABLE) &&
                (audio_control->pcm_value.bit_value.audio_chan10_en == AUDIO_DISABLE) &&
                (audio_control->pcm_value.bit_value.audio_chan11_en == AUDIO_DISABLE) &&
                (audio_control->pcm_value.bit_value.audio_chan12_en == AUDIO_DISABLE) &&
                (audio_control->pcm_value.bit_value.audio_chan13_en == AUDIO_DISABLE) &&
                (audio_control->pcm_value.bit_value.audio_chan14_en == AUDIO_DISABLE) &&
                (audio_control->pcm_value.bit_value.audio_chan15_en == AUDIO_DISABLE) &&
                (audio_control->pcm_value.bit_value.audio_chan16_en == AUDIO_DISABLE)){
            audio_en->value.bit_value.pcm_encode_enable = AUDIO_DISABLE; 
            audio_en->value.bit_value.adpcm_encode_enable = AUDIO_DISABLE;
        }
    } else {
        audio_en->value.bit_value.pcm_encode_enable = AUDIO_ENABLE;
    } 
}

/****************************************************************************************************************
 * Function Name :   audio_enable_set_enable_pcm_encode
 * Description   :   update enable pcm encode's state, and update hardware reg's content
 * Input         :   param1 is audio enable control management struct
 *                   param2 is enable value
 *                   param3 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_enable_set_enable_pcm_encode(audio_enable_control_t *audio_en, int en, isil_chip_t *chip)
{
    if(audio_en->op != NULL){
        audio_en->op->update_enable_pcm_encode(audio_en, en);
    } else {
        audio_enable_update_enable_pcm_encode(audio_en, en);    
    }
    if(chip != NULL){
        chip->io_op->chip_write32(chip, audio_en->reg_offset, audio_en->value.value);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_enable_get_adpcm_encode_state
 * Description   :   get enable adpcm encode's state
 * Input         :   param1 is audio enable control management struct
 * Output        :   null
 * Return        :   real state
 ****************************************************************************************************************/
static int  audio_enable_get_adpcm_encode_state(audio_enable_control_t *audio_en)
{
    return audio_en->value.bit_value.adpcm_encode_enable;
}

/****************************************************************************************************************
 * Function Name :   audio_enable_update_enable_adpcm_encode
 * Description   :   update enable adpcm encode's state
 * Input         :   param1 is audio enable control management struct
 *                   param2 is enable value
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_enable_update_enable_adpcm_encode(audio_enable_control_t *audio_en, int en)
{
    chip_audio_t            *chip_audio = to_get_chip_audio_with_audio_enable(audio_en);
    chip_audio_encode_t     *audio_encode = &chip_audio->audio_encode;
    audio_encode_control_t  *audio_control = &audio_encode->audio_control;

    if(en == AUDIO_DISABLE){
        if((audio_control->adpcm_value.bit_value.audio_chan0_en == AUDIO_DISABLE) &&
                (audio_control->adpcm_value.bit_value.audio_chan1_en == AUDIO_DISABLE) &&
                (audio_control->adpcm_value.bit_value.audio_chan2_en == AUDIO_DISABLE) &&
                (audio_control->adpcm_value.bit_value.audio_chan3_en == AUDIO_DISABLE) &&
                (audio_control->adpcm_value.bit_value.audio_chan4_en == AUDIO_DISABLE) &&
                (audio_control->adpcm_value.bit_value.audio_chan5_en == AUDIO_DISABLE) &&
                (audio_control->adpcm_value.bit_value.audio_chan6_en == AUDIO_DISABLE) &&
                (audio_control->adpcm_value.bit_value.audio_chan7_en == AUDIO_DISABLE) &&
                (audio_control->adpcm_value.bit_value.audio_chan8_en == AUDIO_DISABLE) &&
                (audio_control->adpcm_value.bit_value.audio_chan9_en == AUDIO_DISABLE) &&
                (audio_control->adpcm_value.bit_value.audio_chan10_en == AUDIO_DISABLE) &&
                (audio_control->adpcm_value.bit_value.audio_chan11_en == AUDIO_DISABLE) &&
                (audio_control->adpcm_value.bit_value.audio_chan12_en == AUDIO_DISABLE) &&
                (audio_control->adpcm_value.bit_value.audio_chan13_en == AUDIO_DISABLE) &&
                (audio_control->adpcm_value.bit_value.audio_chan14_en == AUDIO_DISABLE) &&
                (audio_control->adpcm_value.bit_value.audio_chan15_en == AUDIO_DISABLE) &&
                (audio_control->adpcm_value.bit_value.audio_chan16_en == AUDIO_DISABLE)){
            audio_en->value.bit_value.adpcm_encode_enable = AUDIO_DISABLE;
        }
    } else {
        audio_en->value.bit_value.adpcm_encode_enable = AUDIO_ENABLE;
    } 
}

/****************************************************************************************************************
 * Function Name :   audio_enable_set_enable_adpcm_encode
 * Description   :   update enable adpcm encode's state, and update hardware reg's content
 * Input         :   param1 is audio enable control management struct
 *                   param2 is enable value
 *                   param3 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_enable_set_enable_adpcm_encode(audio_enable_control_t *audio_en, int en, isil_chip_t *chip)
{
    if(audio_en->op != NULL){
        audio_en->op->update_enable_adpcm_encode(audio_en, en);
    } else {
        audio_enable_update_enable_adpcm_encode(audio_en, en);    
    }
    if(chip != NULL){
        chip->io_op->chip_write32(chip, audio_en->reg_offset, audio_en->value.value);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_enable_init_enable_control_reg
 * Description   :   audio enable control management struct init
 * Input         :   param1 is audio enable control management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_enable_init_enable_control_reg(audio_enable_control_t *audio_en)
{
    audio_en->reg_offset = AUDIO_EN_CONTROL_REG;
    audio_en->value.value = 0;
}

/****************************************************************************************************************
 * Function Name :   audio_enable_get_enable_control_param
 * Description   :   get enable controll's state from hardware
 * Input         :   param1 is audio enable control management struct
 *                   param2 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_enable_get_enable_control_param(audio_enable_control_t *audio_en, isil_chip_t *chip)
{
    if(chip != NULL){
        audio_en->value.value = chip->io_op->chip_read32(chip, audio_en->reg_offset);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_enable_set_enable_control_param
 * Description   :   set enable controll's state to hardware
 * Input         :   param1 is audio enable control management struct
 *                   param2 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_enable_set_enable_control_param(audio_enable_control_t *audio_en, isil_chip_t *chip)
{
    if(chip != NULL){
        chip->io_op->chip_write32(chip, audio_en->reg_offset, audio_en->value.value);
    }
}

/****************************************************************************************************************
 * Veriable Name :   audio_enable_op
 * Description   :   for helping op audio enable control, we define general interface
 ****************************************************************************************************************/
static struct audio_enable_control_operation	audio_enable_op = {
    .get_adpcm_decode_state = audio_enable_get_adpcm_decode_state,
    .update_enable_adpcm_decode = audio_enable_update_enable_adpcm_decode,
    .set_enable_adpcm_decode = audio_enable_set_enable_adpcm_decode,

    .get_pcm_encode_state = audio_enable_get_pcm_encode_state,
    .update_enable_pcm_encode = audio_enable_update_enable_pcm_encode,
    .set_enable_pcm_encode = audio_enable_set_enable_pcm_encode,

    .get_adpcm_encode_state = audio_enable_get_adpcm_encode_state,
    .update_enable_adpcm_encode = audio_enable_update_enable_adpcm_encode,
    .set_enable_adpcm_encode = audio_enable_set_enable_adpcm_encode,

    .init = audio_enable_init_enable_control_reg,
    .get_enable_control_param = audio_enable_get_enable_control_param,
    .set_enable_control_param = audio_enable_set_enable_control_param,
};

/****************************************************************************************************************
 * Function Name :   init_audio_enable_control
 * Description   :   init audio enable control
 * Input         :   param1 is audio enable control management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void    init_audio_enable_control(audio_enable_control_t *audio_en)
{
    audio_en->op = &audio_enable_op;
    audio_en->op->init(audio_en);
}

/****************************************************************************************************************
 * Function Name :   audio_pcm_encode_wr_rd_ptr_init
 * Description   :   audio pcm encode's wr&rd ptr init
 * Input         :   param1 is audio pcm encode's data position management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_pcm_encode_wr_rd_ptr_init(audio_pcm_encode_ptr_control_t *pcm_encode_ptr)
{
    pcm_encode_ptr->ch0_ch7_wr_ptr_value.value = 0;
    pcm_encode_ptr->ch0_ch7_wr_ptr_reg_offset = PCM_ENCODE_WR_PTR_CH0_CH7_REG;
    pcm_encode_ptr->ch8_ch15_wr_ptr_value.value = 0;
    pcm_encode_ptr->ch8_ch15_wr_ptr_reg_offset = PCM_ENCODE_WR_PTR_CH8_CH15_REG;
    pcm_encode_ptr->ch16_wr_ptr_value.value = 0;
    pcm_encode_ptr->ch16_wr_ptr_reg_offset = PCM_ENCODE_WR_PTR_CH16_REG;
    pcm_encode_ptr->ch0_ch7_rd_ptr_value.value = 0;
    pcm_encode_ptr->ch0_ch7_rd_ptr_reg_offset = PCM_ENCODE_RD_PTR_CH0_CH7_REG;
    pcm_encode_ptr->ch8_ch15_rd_ptr_value.value = 0;
    pcm_encode_ptr->ch8_ch15_rd_ptr_reg_offset = PCM_ENCODE_RD_PTR_CH8_CH15_REG;
    pcm_encode_ptr->ch16_rd_ptr_value.value = 0;
    pcm_encode_ptr->ch16_rd_ptr_reg_offset = PCM_ENCODE_RD_PTR_CH16_REG;
    pcm_encode_ptr->rd_finish_number_reg_offset = HOST_BLOCK_RD_PCM_FINISH_REG;
    pcm_encode_ptr->section_number = AUDIO_IN_PCM_CHAN_SECTION_NUMBER;
}

/****************************************************************************************************************
 * Function Name :   audio_pcm_encode_wr_rd_ptr_get_pcm_wr_rd_ptr
 * Description   :   get audio pcm encode's wr&rd ptr from hardware
 * Input         :   param1 is audio pcm encode's data position management struct
 *                   param2 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_pcm_encode_wr_rd_ptr_get_pcm_wr_rd_ptr(audio_pcm_encode_ptr_control_t *pcm_encode_ptr, isil_chip_t *chip)
{
    if(chip != NULL){
        pcm_encode_ptr->ch0_ch7_wr_ptr_value.value = chip->io_op->chip_read32(chip, pcm_encode_ptr->ch0_ch7_wr_ptr_reg_offset);
        pcm_encode_ptr->ch8_ch15_wr_ptr_value.value = chip->io_op->chip_read32(chip, pcm_encode_ptr->ch8_ch15_wr_ptr_reg_offset);
        pcm_encode_ptr->ch16_wr_ptr_value.value = chip->io_op->chip_read32(chip, pcm_encode_ptr->ch16_wr_ptr_reg_offset);

        pcm_encode_ptr->ch0_ch7_rd_ptr_value.value = chip->io_op->chip_read32(chip, pcm_encode_ptr->ch0_ch7_rd_ptr_reg_offset);
        pcm_encode_ptr->ch8_ch15_rd_ptr_value.value = chip->io_op->chip_read32(chip, pcm_encode_ptr->ch8_ch15_rd_ptr_reg_offset);
        pcm_encode_ptr->ch16_rd_ptr_value.value = chip->io_op->chip_read32(chip, pcm_encode_ptr->ch16_rd_ptr_reg_offset);

        //ISIL_DBG(ISIL_DBG_FATAL, "%08x, %08x\n", pcm_encode_ptr->ch0_ch7_wr_ptr_value.value, pcm_encode_ptr->ch0_ch7_rd_ptr_value.value);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_pcm_encode_wr_rd_ptr_get_chan_wr_ptr
 * Description   :   get audio chan's pcm encode's wr ptr
 * Input         :   param1 is audio pcm encode's data position management struct
 *                   param2 is chan id
 * Output        :   ext_flag
 * Return        :   wr ptr value
 ****************************************************************************************************************/
static int  audio_pcm_encode_wr_rd_ptr_get_chan_wr_ptr(audio_pcm_encode_ptr_control_t *pcm_encode_ptr, int chan_id, int *ext_flag)
{
    int	ret = INVALID_ISIL_AUDIO_SECTION_ID;
    switch(chan_id){
        case ISIL_AUDIO_IN_CHAN0_ID:
            ret = pcm_encode_ptr->ch0_ch7_wr_ptr_value.bit_value.audio_chan0_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN1_ID:
            ret = pcm_encode_ptr->ch0_ch7_wr_ptr_value.bit_value.audio_chan1_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN2_ID:
            ret = pcm_encode_ptr->ch0_ch7_wr_ptr_value.bit_value.audio_chan2_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN3_ID:
            ret = pcm_encode_ptr->ch0_ch7_wr_ptr_value.bit_value.audio_chan3_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN4_ID:
            ret = pcm_encode_ptr->ch0_ch7_wr_ptr_value.bit_value.audio_chan4_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN5_ID:
            ret = pcm_encode_ptr->ch0_ch7_wr_ptr_value.bit_value.audio_chan5_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN6_ID:
            ret = pcm_encode_ptr->ch0_ch7_wr_ptr_value.bit_value.audio_chan6_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN7_ID:
            ret = pcm_encode_ptr->ch0_ch7_wr_ptr_value.bit_value.audio_chan7_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN8_ID:
            ret = pcm_encode_ptr->ch8_ch15_wr_ptr_value.bit_value.audio_chan08_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN9_ID:
            ret = pcm_encode_ptr->ch8_ch15_wr_ptr_value.bit_value.audio_chan09_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN10_ID:
            ret = pcm_encode_ptr->ch8_ch15_wr_ptr_value.bit_value.audio_chan10_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN11_ID:
            ret = pcm_encode_ptr->ch8_ch15_wr_ptr_value.bit_value.audio_chan11_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN12_ID:
            ret = pcm_encode_ptr->ch8_ch15_wr_ptr_value.bit_value.audio_chan12_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN13_ID:
            ret = pcm_encode_ptr->ch8_ch15_wr_ptr_value.bit_value.audio_chan13_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN14_ID:
            ret = pcm_encode_ptr->ch8_ch15_wr_ptr_value.bit_value.audio_chan14_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN15_ID:
            ret = pcm_encode_ptr->ch8_ch15_wr_ptr_value.bit_value.audio_chan15_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN16_ID:
            ret = pcm_encode_ptr->ch16_wr_ptr_value.bit_value.audio_chan16_ptr;
            break;
        default:
            ISIL_DBG(ISIL_DBG_INFO, ": err audio chan id %d\n", chan_id);
            break;
    }
    *ext_flag = ret>>(get_power_base((u32)AUDIO_IN_PCM_CHAN_SECTION_NUMBER));
    ret &= ((1<<(get_power_base((u32)AUDIO_IN_PCM_CHAN_SECTION_NUMBER)))-1);
    return ret;
}

/****************************************************************************************************************
 * Function Name :   audio_pcm_encode_wr_rd_ptr_get_chan_rd_ptr
 * Description   :   get audio chan's pcm encode's rd ptr
 * Input         :   param1 is audio pcm encode's data position management struct
 *                   param2 is chan id
 * Output        :   ext_flag
 * Return        :   wr ptr value
 ****************************************************************************************************************/
static int  audio_pcm_encode_wr_rd_ptr_get_chan_rd_ptr(audio_pcm_encode_ptr_control_t *pcm_encode_ptr, int chan_id, int *ext_flag)
{
    int	ret = INVALID_ISIL_AUDIO_SECTION_ID;
    switch(chan_id){
        case ISIL_AUDIO_IN_CHAN0_ID:
            ret = pcm_encode_ptr->ch0_ch7_rd_ptr_value.bit_value.audio_chan0_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN1_ID:
            ret = pcm_encode_ptr->ch0_ch7_rd_ptr_value.bit_value.audio_chan1_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN2_ID:
            ret = pcm_encode_ptr->ch0_ch7_rd_ptr_value.bit_value.audio_chan2_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN3_ID:
            ret = pcm_encode_ptr->ch0_ch7_rd_ptr_value.bit_value.audio_chan3_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN4_ID:
            ret = pcm_encode_ptr->ch0_ch7_rd_ptr_value.bit_value.audio_chan4_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN5_ID:
            ret = pcm_encode_ptr->ch0_ch7_rd_ptr_value.bit_value.audio_chan5_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN6_ID:
            ret = pcm_encode_ptr->ch0_ch7_rd_ptr_value.bit_value.audio_chan6_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN7_ID:
            ret = pcm_encode_ptr->ch0_ch7_rd_ptr_value.bit_value.audio_chan7_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN8_ID:
            ret = pcm_encode_ptr->ch8_ch15_rd_ptr_value.bit_value.audio_chan08_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN9_ID:
            ret = pcm_encode_ptr->ch8_ch15_rd_ptr_value.bit_value.audio_chan09_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN10_ID:
            ret = pcm_encode_ptr->ch8_ch15_rd_ptr_value.bit_value.audio_chan10_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN11_ID:
            ret = pcm_encode_ptr->ch8_ch15_rd_ptr_value.bit_value.audio_chan11_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN12_ID:
            ret = pcm_encode_ptr->ch8_ch15_rd_ptr_value.bit_value.audio_chan12_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN13_ID:
            ret = pcm_encode_ptr->ch8_ch15_rd_ptr_value.bit_value.audio_chan13_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN14_ID:
            ret = pcm_encode_ptr->ch8_ch15_rd_ptr_value.bit_value.audio_chan14_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN15_ID:
            ret = pcm_encode_ptr->ch8_ch15_rd_ptr_value.bit_value.audio_chan15_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN16_ID:
            ret = pcm_encode_ptr->ch16_rd_ptr_value.bit_value.audio_chan16_ptr;
            break;
        default:
            ISIL_DBG(ISIL_DBG_INFO, ": err audio chan id %d\n", chan_id);
            break;
    }
    *ext_flag = ret>>(get_power_base((u32)AUDIO_IN_PCM_CHAN_SECTION_NUMBER));
    ret &= ((1<<(get_power_base((u32)AUDIO_IN_PCM_CHAN_SECTION_NUMBER)))-1);
    return ret;
}

/****************************************************************************************************************
 * Function Name :   audio_pcm_encode_wr_rd_get_chan_section_number
 * Description   :   get audio chan section's number
 * Input         :   param1 is audio pcm encode's data position management struct
 * Output        :   null
 * Return        :   section number
 ****************************************************************************************************************/
static int  audio_pcm_encode_wr_rd_get_chan_section_number(audio_pcm_encode_ptr_control_t *pcm_encode_ptr)
{
    return pcm_encode_ptr->section_number;
}

/****************************************************************************************************************
 * Function Name :   audio_pcm_encode_wr_rd_ptr_set_rd_finish_number
 * Description   :   update hardware's section id
 * Input         :   param1 is audio pcm encode's data position management struct
 *                   param2 is section id
 *                   param3 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_pcm_encode_wr_rd_ptr_set_rd_finish_number(audio_pcm_encode_ptr_control_t *pcm_encode_ptr, int section_id, isil_chip_t *chip)
{
    chip->io_op->chip_write32(chip, pcm_encode_ptr->rd_finish_number_reg_offset, section_id);
}

/****************************************************************************************************************
 * Veriable Name :   audio_pcm_encode_wr_rd_ptr_op
 * Description   :   for helping op audio pcm encode wr rd ptr, we define general interface
 ****************************************************************************************************************/
static struct audio_pcm_encode_wr_rd_ptr_operation  audio_pcm_encode_wr_rd_ptr_op = {
    .init = audio_pcm_encode_wr_rd_ptr_init,
    .get_pcm_wr_rd_ptr = audio_pcm_encode_wr_rd_ptr_get_pcm_wr_rd_ptr,
    .get_chan_wr_ptr = audio_pcm_encode_wr_rd_ptr_get_chan_wr_ptr,
    .get_chan_rd_ptr = audio_pcm_encode_wr_rd_ptr_get_chan_rd_ptr,
    .get_chan_section_number = audio_pcm_encode_wr_rd_get_chan_section_number,
    .set_rd_finish_number = audio_pcm_encode_wr_rd_ptr_set_rd_finish_number,
};

/****************************************************************************************************************
 * Function Name :   init_audio_pcm_encode_ptr_control
 * Description   :   init audio pcm encode wr&rd ptr control
 * Input         :   param1 is audio pcm encode's data position management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void    init_audio_pcm_encode_ptr_control(audio_pcm_encode_ptr_control_t *pcm_encode_ptr)
{
    pcm_encode_ptr->op = &audio_pcm_encode_wr_rd_ptr_op;
    pcm_encode_ptr->op->init(pcm_encode_ptr);
}

/****************************************************************************************************************
 * Function Name :   audio_pcm_encode_data_addr_set_reg_bit_section_offset
 * Description   :   update pcm encode data addr's section offset
 * Input         :   param1 is audio pcm encode's data address management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_pcm_encode_data_addr_set_reg_bit_section_offset(audio_pcm_encode_data_addr_t *pcm_encode_data_addr, u32 section_offset)
{
    pcm_encode_data_addr->value.bit_value.section_offset = section_offset;
}

/****************************************************************************************************************
 * Function Name :   audio_pcm_encode_data_addr_get_reg_bit_section_offset
 * Description   :   get pcm encode data addr's section offset
 * Input         :   param1 is audio pcm encode's data address management struct
 * Output        :   null
 * Return        :   section offset
 ****************************************************************************************************************/
static u32  audio_pcm_encode_data_addr_get_reg_bit_section_offset(audio_pcm_encode_data_addr_t *pcm_encode_data_addr)
{
    return pcm_encode_data_addr->value.bit_value.section_offset;
}

/****************************************************************************************************************
 * Function Name :   audio_pcm_encode_data_addr_set_reg_bit_section_id
 * Description   :   update pcm encode data addr's section id
 * Input         :   param1 is audio pcm encode's data address management struct
 *                   param2 is section id
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_pcm_encode_data_addr_set_reg_bit_section_id(audio_pcm_encode_data_addr_t *pcm_encode_data_addr, u32 section_id)
{
    pcm_encode_data_addr->value.bit_value.section_id = section_id;
}

/****************************************************************************************************************
 * Function Name :   audio_pcm_encode_data_addr_get_reg_bit_section_id
 * Description   :   get pcm encode data addr's section id
 * Input         :   param1 is audio pcm encode's data address management struct
 * Output        :   null
 * Return        :   section_id
 ****************************************************************************************************************/
static u32  audio_pcm_encode_data_addr_get_reg_bit_section_id(audio_pcm_encode_data_addr_t *pcm_encode_data_addr)
{
    return pcm_encode_data_addr->value.bit_value.section_id;
}

/****************************************************************************************************************
 * Function Name :   audio_pcm_encode_data_addr_set_reg_bit_chan_id
 * Description   :   update pcm encode data addr's chan id
 * Input         :   param1 is audio pcm encode's data address management struct
 *                   param2 is chan id
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_pcm_encode_data_addr_set_reg_bit_chan_id(audio_pcm_encode_data_addr_t *pcm_encode_data_addr, u32 chan_id)
{
    pcm_encode_data_addr->value.bit_value.chan_id = chan_id;
}

/****************************************************************************************************************
 * Function Name :   audio_pcm_encode_data_addr_set_reg_bit_chan_id
 * Description   :   get pcm encode data addr's chan id
 * Input         :   param1 is audio pcm encode's data address management struct
 * Output        :   null
 * Return        :   chan id
 ****************************************************************************************************************/
static u32  audio_pcm_encode_data_addr_get_reg_bit_chan_id(audio_pcm_encode_data_addr_t *pcm_encode_data_addr)
{
    return pcm_encode_data_addr->value.bit_value.chan_id;
}

/****************************************************************************************************************
 * Function Name :   audio_pcm_encode_data_addr_set_reg_bit_pcm_type
 * Description   :   update pcm encode data addr's pcm type
 * Input         :   param1 is audio pcm encode's data address management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_pcm_encode_data_addr_set_reg_bit_pcm_type(audio_pcm_encode_data_addr_t *pcm_encode_data_addr)
{
    pcm_encode_data_addr->value.bit_value.pcm_type = ADUIO_DATA_ADDR_PCM_ENCODE_TYPE;
}

/****************************************************************************************************************
 * Function Name :   audio_pcm_encode_data_addr_get_reg_bit_pcm_type
 * Description   :   get pcm encode data addr's pcm type
 * Input         :   param1 is audio pcm encode's data address management struct
 * Output        :   null
 * Return        :   ADUIO_DATA_ADDR_PCM_ENCODE_TYPE
 ****************************************************************************************************************/
static u32  audio_pcm_encode_data_addr_get_reg_bit_pcm_type(audio_pcm_encode_data_addr_t *pcm_encode_data_addr)
{
    return ADUIO_DATA_ADDR_PCM_ENCODE_TYPE;
}

/****************************************************************************************************************
 * Function Name :   audio_pcm_encode_data_addr_get_pcm_encode_data_addr
 * Description   :   get pcm encode data addr
 * Input         :   param1 is audio pcm encode's data address management struct
 * Output        :   null
 * Return        :   data addr
 ****************************************************************************************************************/
static u32  audio_pcm_encode_data_addr_get_pcm_encode_data_addr(audio_pcm_encode_data_addr_t *pcm_encode_data_addr)
{
    return pcm_encode_data_addr->value.value;
}

/****************************************************************************************************************
 * Function Name :   audio_pcm_encode_data_addr_get_buf_page_id
 * Description   :   get pcm encode data's buf page id
 * Input         :   param1 is audio pcm encode's data address management struct
 * Output        :   null
 * Return        :   page id
 ****************************************************************************************************************/
static int  audio_pcm_encode_data_addr_get_buf_page_id(audio_pcm_encode_data_addr_t *pcm_encode_data_addr)
{
    return pcm_encode_data_addr->buf_page_id;
}

/****************************************************************************************************************
 * Function Name :   audio_pcm_encode_data_addr_get_buf_page_id
 * Description   :   get pcm encode data in chip A or B
 * Input         :   param1 is audio pcm encode's data address management struct
 * Output        :   null
 * Return        :   chip_a_or_b
 ****************************************************************************************************************/
static int  audio_pcm_encode_data_addr_get_buf_in_chip_a_or_b(audio_pcm_encode_data_addr_t *pcm_encode_data_addr)
{
    return pcm_encode_data_addr->chip_a_or_b;
}

/****************************************************************************************************************
 * Function Name :   audio_pcm_encode_data_addr_get_pcm_encode_data_addr_in_host_end_offset
 * Description   :   get pcm encode data addr at host end
 * Input         :   param1 is audio pcm encode's data address management struct
 * Output        :   null
 * Return        :   host_end_addr_offset
 ****************************************************************************************************************/
static u32  audio_pcm_encode_data_addr_get_pcm_encode_data_addr_in_host_end_offset(audio_pcm_encode_data_addr_t *pcm_encode_data_addr)
{
    u32 host_end_offset;
    host_end_offset = DDRBASE;
    host_end_offset |= (pcm_encode_data_addr->value.value<<2);
    return host_end_offset;
}

/****************************************************************************************************************
 * Function Name :   audio_pcm_encode_data_addr_get_pcm_encode_data_addr_in_ddr_end_offset
 * Description   :   get pcm encode data addr at ddr end
 * Input         :   param1 is audio pcm encode's data address management struct
 * Output        :   null
 * Return        :   ddr_end_addr_offset
 ****************************************************************************************************************/
static u32  audio_pcm_encode_data_addr_get_pcm_encode_data_addr_in_ddr_end_offset(audio_pcm_encode_data_addr_t *pcm_encode_data_addr)
{
    u32	ddr_end_offset;
    ddr_end_offset = (pcm_encode_data_addr->buf_page_id<<19);
    ddr_end_offset |= (pcm_encode_data_addr->value.value<<2);
    return ddr_end_offset;
}

/****************************************************************************************************************
 * Function Name :   audio_pcm_encode_data_addr_init
 * Description   :   audio pcm encode data addr maintain struct init
 * Input         :   param1 is audio pcm encode's data address management struct
 *                   param2 is page_id
 *                   param3 is chip_a_or_b
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_pcm_encode_data_addr_init(audio_pcm_encode_data_addr_t *pcm_encode_data_addr, int page_id, int chip_a_or_b)
{
    pcm_encode_data_addr->value.value = 0;
    pcm_encode_data_addr->value.bit_value.pcm_type = ADUIO_DATA_ADDR_PCM_ENCODE_TYPE;
    pcm_encode_data_addr->buf_page_id = page_id;
    pcm_encode_data_addr->chip_a_or_b = chip_a_or_b;
}

/****************************************************************************************************************
 * Veriable Name :   audio_pcm_encode_data_addr_op
 * Description   :   for helping op audio pcm encode data addr calc, we define general interface
 ****************************************************************************************************************/
static struct audio_pcm_encode_data_addr_operation  audio_pcm_encode_data_addr_op = {
    .set_reg_bit_section_offset = audio_pcm_encode_data_addr_set_reg_bit_section_offset,
    .get_reg_bit_section_offset = audio_pcm_encode_data_addr_get_reg_bit_section_offset,
    .set_reg_bit_section_id = audio_pcm_encode_data_addr_set_reg_bit_section_id,
    .get_reg_bit_section_id = audio_pcm_encode_data_addr_get_reg_bit_section_id,
    .set_reg_bit_chan_id = audio_pcm_encode_data_addr_set_reg_bit_chan_id,
    .get_reg_bit_chan_id = audio_pcm_encode_data_addr_get_reg_bit_chan_id,
    .set_reg_bit_pcm_type = audio_pcm_encode_data_addr_set_reg_bit_pcm_type,
    .get_reg_bit_pcm_type = audio_pcm_encode_data_addr_get_reg_bit_pcm_type,
    .get_pcm_encode_data_addr = audio_pcm_encode_data_addr_get_pcm_encode_data_addr,
    .get_buf_page_id = audio_pcm_encode_data_addr_get_buf_page_id,
    .get_buf_in_chip_a_or_b = audio_pcm_encode_data_addr_get_buf_in_chip_a_or_b,
    .get_pcm_encode_data_addr_in_host_end_offset = audio_pcm_encode_data_addr_get_pcm_encode_data_addr_in_host_end_offset,
    .get_pcm_encode_data_addr_in_ddr_end_offset = audio_pcm_encode_data_addr_get_pcm_encode_data_addr_in_ddr_end_offset,
    .init = audio_pcm_encode_data_addr_init,
};

/****************************************************************************************************************
 * Function Name :   init_audio_pcm_encode_data_addr
 * Description   :   init audio pcm encode data addr maintain struct
 * Input         :   param1 is audio pcm encode's data address management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void	init_audio_pcm_encode_data_addr(audio_pcm_encode_data_addr_t *pcm_encode_data_addr)
{
    if(pcm_encode_data_addr != NULL){
        pcm_encode_data_addr->op = &audio_pcm_encode_data_addr_op;
        pcm_encode_data_addr->op->init(pcm_encode_data_addr, ISIL_AUDIO_BUF_PAGE_ID, ISIL_AUDIO_BUF_OF_DDR_ID);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_adpcm_encode_wr_rd_ptr_init
 * Description   :   audio adpcm encode's wr&rd ptr init
 * Input         :   param1 is audio adpcm encode's data position management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_adpcm_encode_wr_rd_ptr_init(audio_adpcm_encode_ptr_control_t *adpcm_encode_ptr)
{
    adpcm_encode_ptr->ch0_ch9_wr_ptr_value.value = 0;
    adpcm_encode_ptr->ch0_ch9_wr_ptr_reg_offset = ADPCM_ENCODE_WR_PTR_CH0_CH9_REG;
    adpcm_encode_ptr->ch10_ch16_wr_ptr_value.value = 0;
    adpcm_encode_ptr->ch10_ch16_wr_ptr_reg_offset = ADPCM_ENCODE_WR_PTR_CH10_CH16_REG;

    adpcm_encode_ptr->ch0_ch9_rd_ptr_value.value = 0;
    adpcm_encode_ptr->ch0_ch9_rd_ptr_reg_offset = ADPCM_ENCODE_RD_PTR_CH0_CH9_REG;
    adpcm_encode_ptr->ch10_ch16_rd_ptr_value.value = 0;
    adpcm_encode_ptr->ch10_ch16_rd_ptr_reg_offset = ADPCM_ENCODE_RD_PTR_CH10_CH16_REG;

    adpcm_encode_ptr->rd_finish_number_reg_offset = HOST_BLOCK_RD_ADPCM_FINISH_REG;
    adpcm_encode_ptr->section_number = AUDIO_IN_ADPCM_CHAN_SECTION_NUMBER;
}

/****************************************************************************************************************
 * Function Name :   audio_adpcm_encode_wr_rd_ptr_get_adpcm_wr_rd_ptr
 * Description   :   get audio adpcm encode's wr&rd ptr from hardware
 * Input         :   param1 is audio adpcm encode's data position management struct
 *                   param2 is chip management
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_adpcm_encode_wr_rd_ptr_get_adpcm_wr_rd_ptr(audio_adpcm_encode_ptr_control_t *adpcm_encode_ptr, isil_chip_t *chip)
{
    if(chip != NULL){
        adpcm_encode_ptr->ch0_ch9_wr_ptr_value.value = chip->io_op->chip_read32(chip, adpcm_encode_ptr->ch0_ch9_wr_ptr_reg_offset);
        adpcm_encode_ptr->ch10_ch16_wr_ptr_value.value = chip->io_op->chip_read32(chip, adpcm_encode_ptr->ch10_ch16_wr_ptr_reg_offset);

        adpcm_encode_ptr->ch0_ch9_rd_ptr_value.value = chip->io_op->chip_read32(chip, adpcm_encode_ptr->ch0_ch9_rd_ptr_reg_offset);
        adpcm_encode_ptr->ch10_ch16_rd_ptr_value.value = chip->io_op->chip_read32(chip, adpcm_encode_ptr->ch10_ch16_rd_ptr_reg_offset);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_adpcm_encode_wr_rd_ptr_get_chan_wr_ptr
 * Description   :   get audio adpcm encode chan's wr ptr
 * Input         :   param1 is audio adpcm encode's data position management struct
 *                   param2 is chip id
 * Output        :   ext_flag
 * Return        :   wr_ptr
 ****************************************************************************************************************/
static int  audio_adpcm_encode_wr_rd_ptr_get_chan_wr_ptr(audio_adpcm_encode_ptr_control_t *adpcm_encode_ptr, int chan_id, int *ext_flag)
{
    int	ret = INVALID_ISIL_AUDIO_SECTION_ID;
    switch(chan_id){
        case ISIL_AUDIO_IN_CHAN0_ID:
            ret = adpcm_encode_ptr->ch0_ch9_wr_ptr_value.bit_value.audio_chan0_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN1_ID:
            ret = adpcm_encode_ptr->ch0_ch9_wr_ptr_value.bit_value.audio_chan1_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN2_ID:
            ret = adpcm_encode_ptr->ch0_ch9_wr_ptr_value.bit_value.audio_chan2_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN3_ID:
            ret = adpcm_encode_ptr->ch0_ch9_wr_ptr_value.bit_value.audio_chan3_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN4_ID:
            ret = adpcm_encode_ptr->ch0_ch9_wr_ptr_value.bit_value.audio_chan4_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN5_ID:
            ret = adpcm_encode_ptr->ch0_ch9_wr_ptr_value.bit_value.audio_chan5_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN6_ID:
            ret = adpcm_encode_ptr->ch0_ch9_wr_ptr_value.bit_value.audio_chan6_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN7_ID:
            ret = adpcm_encode_ptr->ch0_ch9_wr_ptr_value.bit_value.audio_chan7_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN8_ID:
            ret = adpcm_encode_ptr->ch0_ch9_wr_ptr_value.bit_value.audio_chan8_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN9_ID:
            ret = adpcm_encode_ptr->ch0_ch9_wr_ptr_value.bit_value.audio_chan9_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN10_ID:
            ret = adpcm_encode_ptr->ch10_ch16_wr_ptr_value.bit_value.audio_chan10_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN11_ID:
            ret = adpcm_encode_ptr->ch10_ch16_wr_ptr_value.bit_value.audio_chan11_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN12_ID:
            ret = adpcm_encode_ptr->ch10_ch16_wr_ptr_value.bit_value.audio_chan12_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN13_ID:
            ret = adpcm_encode_ptr->ch10_ch16_wr_ptr_value.bit_value.audio_chan13_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN14_ID:
            ret = adpcm_encode_ptr->ch10_ch16_wr_ptr_value.bit_value.audio_chan14_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN15_ID:
            ret = adpcm_encode_ptr->ch10_ch16_wr_ptr_value.bit_value.audio_chan15_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN16_ID:
            ret = adpcm_encode_ptr->ch10_ch16_wr_ptr_value.bit_value.audio_chan16_ptr;
            break;
        default:
            ISIL_DBG(ISIL_DBG_INFO, ": err audio chan id %d\n", chan_id);
            break;
    }
    *ext_flag = ret>>(get_power_base((u32)AUDIO_IN_ADPCM_CHAN_SECTION_NUMBER));
    ret &= ((1<<(get_power_base((u32)AUDIO_IN_ADPCM_CHAN_SECTION_NUMBER)))-1);
    return ret;
}

/****************************************************************************************************************
 * Function Name :   audio_adpcm_encode_wr_rd_ptr_get_chan_rd_ptr
 * Description   :   get audio adpcm encode chan's rd ptr
 * Input         :   param1 is audio adpcm encode's data position management struct
 *                   param2 is chip id
 * Output        :   ext_flag
 * Return        :   rd_ptr
 ****************************************************************************************************************/
static int  audio_adpcm_encode_wr_rd_ptr_get_chan_rd_ptr(audio_adpcm_encode_ptr_control_t *adpcm_encode_ptr, int chan_id, int *ext_flag)
{
    int	ret = INVALID_ISIL_AUDIO_SECTION_ID;
    switch(chan_id){
        case ISIL_AUDIO_IN_CHAN0_ID:
            ret = adpcm_encode_ptr->ch0_ch9_rd_ptr_value.bit_value.audio_chan0_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN1_ID:
            ret = adpcm_encode_ptr->ch0_ch9_rd_ptr_value.bit_value.audio_chan1_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN2_ID:
            ret = adpcm_encode_ptr->ch0_ch9_rd_ptr_value.bit_value.audio_chan2_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN3_ID:
            ret = adpcm_encode_ptr->ch0_ch9_rd_ptr_value.bit_value.audio_chan3_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN4_ID:
            ret = adpcm_encode_ptr->ch0_ch9_rd_ptr_value.bit_value.audio_chan4_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN5_ID:
            ret = adpcm_encode_ptr->ch0_ch9_rd_ptr_value.bit_value.audio_chan5_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN6_ID:
            ret = adpcm_encode_ptr->ch0_ch9_rd_ptr_value.bit_value.audio_chan6_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN7_ID:
            ret = adpcm_encode_ptr->ch0_ch9_rd_ptr_value.bit_value.audio_chan7_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN8_ID:
            ret = adpcm_encode_ptr->ch0_ch9_rd_ptr_value.bit_value.audio_chan8_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN9_ID:
            ret = adpcm_encode_ptr->ch0_ch9_rd_ptr_value.bit_value.audio_chan9_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN10_ID:
            ret = adpcm_encode_ptr->ch10_ch16_rd_ptr_value.bit_value.audio_chan10_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN11_ID:
            ret = adpcm_encode_ptr->ch10_ch16_rd_ptr_value.bit_value.audio_chan11_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN12_ID:
            ret = adpcm_encode_ptr->ch10_ch16_rd_ptr_value.bit_value.audio_chan12_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN13_ID:
            ret = adpcm_encode_ptr->ch10_ch16_rd_ptr_value.bit_value.audio_chan13_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN14_ID:
            ret = adpcm_encode_ptr->ch10_ch16_rd_ptr_value.bit_value.audio_chan14_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN15_ID:
            ret = adpcm_encode_ptr->ch10_ch16_rd_ptr_value.bit_value.audio_chan15_ptr;
            break;
        case ISIL_AUDIO_IN_CHAN16_ID:
            ret = adpcm_encode_ptr->ch10_ch16_rd_ptr_value.bit_value.audio_chan16_ptr;
            break;
        default:
            ISIL_DBG(ISIL_DBG_INFO, ": err audio chan id %d\n", chan_id);
            break;
    }
    *ext_flag = ret>>(get_power_base((u32)AUDIO_IN_ADPCM_CHAN_SECTION_NUMBER));
    ret &= ((1<<(get_power_base((u32)AUDIO_IN_ADPCM_CHAN_SECTION_NUMBER)))-1);
    return ret;
}

/****************************************************************************************************************
 * Function Name :   audio_adpcm_encode_wr_rd_ptr_get_chan_section_number
 * Description   :   get audio adpcm encode chan's section number
 * Input         :   param1 is audio adpcm encode's data position management struct
 * Output        :   null
 * Return        :   section number
 ****************************************************************************************************************/
static int  audio_adpcm_encode_wr_rd_ptr_get_chan_section_number(audio_adpcm_encode_ptr_control_t *adpcm_encode_ptr)
{
    return adpcm_encode_ptr->section_number;
}

/****************************************************************************************************************
 * Function Name :   audio_adpcm_encode_wr_rd_ptr_set_rd_finish_number
 * Description   :   get audio adpcm encode chan's finish number
 * Input         :   param1 is audio adpcm encode's data position management struct
 *                   param2 is section id
 *                   param3 is chip management struct
 * Output        :   null
 * Return        :   section number
 ****************************************************************************************************************/
static void audio_adpcm_encode_wr_rd_ptr_set_rd_finish_number(audio_adpcm_encode_ptr_control_t *adpcm_encode_ptr, int section_id, isil_chip_t *chip)
{
    if(chip != NULL){
        chip->io_op->chip_write32(chip, adpcm_encode_ptr->rd_finish_number_reg_offset, section_id);
    }
}

/****************************************************************************************************************
 * Veriable Name :   audio_adpcm_encode_wr_rd_ptr_op
 * Description   :   for helping op audio adpcm encode wr&rd ptr, we define general interface
 ****************************************************************************************************************/
static struct audio_adpcm_encode_wr_rd_ptr_operation	audio_adpcm_encode_wr_rd_ptr_op = {
    .init = audio_adpcm_encode_wr_rd_ptr_init,
    .get_adpcm_wr_rd_ptr = audio_adpcm_encode_wr_rd_ptr_get_adpcm_wr_rd_ptr,

    .get_chan_wr_ptr = audio_adpcm_encode_wr_rd_ptr_get_chan_wr_ptr,
    .get_chan_rd_ptr = audio_adpcm_encode_wr_rd_ptr_get_chan_rd_ptr,
    .get_chan_section_number = audio_adpcm_encode_wr_rd_ptr_get_chan_section_number,
    .set_rd_finish_number = audio_adpcm_encode_wr_rd_ptr_set_rd_finish_number,
};

/****************************************************************************************************************
 * Function Name :   init_audio_adpcm_encode_ptr_control
 * Description   :   init audio adpcm encode's wr&rd ptr
 * Input         :   param1 is audio adpcm encode's data position management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void	init_audio_adpcm_encode_ptr_control(audio_adpcm_encode_ptr_control_t *adpcm_encode_ptr)
{
    adpcm_encode_ptr->op = &audio_adpcm_encode_wr_rd_ptr_op;
    adpcm_encode_ptr->op->init(adpcm_encode_ptr);
}

/****************************************************************************************************************
 * Function Name :   audio_adpcm_encode_data_addr_set_reg_bit_section_offset
 * Description   :   update adpcm encode data addr's section offset
 * Input         :   param1 is audio adpcm encode's data address management struct
 *                   param2 is section offset
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_adpcm_encode_data_addr_set_reg_bit_section_offset(audio_adpcm_encode_data_addr_t *adpcm_encode_data_addr, u32 section_offset)
{
    adpcm_encode_data_addr->value.bit_value.section_offset = section_offset;
}

/****************************************************************************************************************
 * Function Name :   audio_adpcm_encode_data_addr_set_reg_bit_section_offset
 * Description   :   get adpcm encode data addr's section offset
 * Input         :   param1 is audio adpcm encode's data address management struct
 * Output        :   null
 * Return        :   section offset
 ****************************************************************************************************************/
static u32  audio_adpcm_encode_data_addr_get_reg_bit_section_offset(audio_adpcm_encode_data_addr_t *adpcm_encode_data_addr)
{
    return adpcm_encode_data_addr->value.bit_value.section_offset;
}

/****************************************************************************************************************
 * Function Name :   audio_adpcm_encode_data_addr_set_reg_bit_section_id
 * Description   :   update adpcm encode data addr's section id
 * Input         :   param1 is audio adpcm encode's data address management struct
 *                   param2 is section offset
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_adpcm_encode_data_addr_set_reg_bit_section_id(audio_adpcm_encode_data_addr_t *adpcm_encode_data_addr, u32 section_id)
{
    adpcm_encode_data_addr->value.bit_value.section_id = section_id;
}

/****************************************************************************************************************
 * Function Name :   audio_adpcm_encode_data_addr_get_reg_bit_section_id
 * Description   :   get adpcm encode data addr's section id
 * Input         :   param1 is audio adpcm encode's data address management struct
 * Output        :   null
 * Return        :   section id
 ****************************************************************************************************************/
static u32  audio_adpcm_encode_data_addr_get_reg_bit_section_id(audio_adpcm_encode_data_addr_t *adpcm_encode_data_addr)
{
    return adpcm_encode_data_addr->value.bit_value.section_id;
}

/****************************************************************************************************************
 * Function Name :   audio_adpcm_encode_data_addr_set_reg_bit_chan_id
 * Description   :   update adpcm encode data addr's chan id
 * Input         :   param1 is audio adpcm encode's data address management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_adpcm_encode_data_addr_set_reg_bit_chan_id(audio_adpcm_encode_data_addr_t *adpcm_encode_data_addr, u32 chan_id)
{
    adpcm_encode_data_addr->value.bit_value.chan_id = chan_id;
}

/****************************************************************************************************************
 * Function Name :   audio_adpcm_encode_data_addr_get_reg_bit_chan_id
 * Description   :   get adpcm encode data addr's chan id
 * Input         :   param1 is audio adpcm encode's data address management struct
 * Output        :   null
 * Return        :   chan_id
 ****************************************************************************************************************/
static u32  audio_adpcm_encode_data_addr_get_reg_bit_chan_id(audio_adpcm_encode_data_addr_t *adpcm_encode_data_addr)
{
    return adpcm_encode_data_addr->value.bit_value.chan_id;
}

/****************************************************************************************************************
 * Function Name :   audio_adpcm_encode_data_addr_set_reg_bit_adpcm_type
 * Description   :   update adpcm encode data addr's adpcm encode type
 * Input         :   param1 is audio adpcm encode's data address management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_adpcm_encode_data_addr_set_reg_bit_adpcm_type(audio_adpcm_encode_data_addr_t *adpcm_encode_data_addr)
{
    adpcm_encode_data_addr->value.bit_value.adpcm_encode_type = ADUIO_DATA_ADDR_ADPCM_ENCODE_TYPE;
}

/****************************************************************************************************************
 * Function Name :   audio_adpcm_encode_data_addr_get_reg_bit_adpcm_type
 * Description   :   get adpcm encode data addr's adpcm encode type
 * Input         :   param1 is audio adpcm encode's data address management struct
 * Output        :   null
 * Return        :   adpcm encode type
 ****************************************************************************************************************/
static u32  audio_adpcm_encode_data_addr_get_reg_bit_adpcm_type(audio_adpcm_encode_data_addr_t *adpcm_encode_data_addr)
{
    return adpcm_encode_data_addr->value.bit_value.adpcm_encode_type;
}

/****************************************************************************************************************
 * Function Name :   audio_adpcm_encode_data_addr_get_adpcm_encode_data_addr
 * Description   :   get adpcm encode data addr
 * Input         :   param1 is audio adpcm encode's data address management struct
 * Output        :   null
 * Return        :   encode data addr
 ****************************************************************************************************************/
static u32  audio_adpcm_encode_data_addr_get_adpcm_encode_data_addr(audio_adpcm_encode_data_addr_t *adpcm_encode_data_addr)
{
    return adpcm_encode_data_addr->value.value;
}

/****************************************************************************************************************
 * Function Name :   audio_adpcm_encode_data_addr_get_buf_page_id
 * Description   :   get adpcm encode data buf page id
 * Input         :   param1 is audio adpcm encode's data address management struct
 * Output        :   null
 * Return        :   buf page id
 ****************************************************************************************************************/
static int  audio_adpcm_encode_data_addr_get_buf_page_id(audio_adpcm_encode_data_addr_t *adpcm_encode_data_addr)
{
    return adpcm_encode_data_addr->buf_page_id;
}

/****************************************************************************************************************
 * Function Name :   audio_adpcm_encode_data_addr_get_buf_in_chip_a_or_b
 * Description   :   get adpcm encode data buf chip_a_or_b
 * Input         :   param1 is audio adpcm encode's data address management struct
 * Output        :   null
 * Return        :   chip_a_or_b
 ****************************************************************************************************************/
static int  audio_adpcm_encode_data_addr_get_buf_in_chip_a_or_b(audio_adpcm_encode_data_addr_t *adpcm_encode_data_addr)
{
    return adpcm_encode_data_addr->chip_a_or_b;
}

/****************************************************************************************************************
 * Function Name :   audio_adpcm_encode_data_addr_get_adpcm_encode_data_addr_in_host_end_offset
 * Description   :   get adpcm encode data addr of host end
 * Input         :   param1 is audio adpcm encode's data address management struct
 * Output        :   null
 * Return        :   addr of host end
 ****************************************************************************************************************/
static u32  audio_adpcm_encode_data_addr_get_adpcm_encode_data_addr_in_host_end_offset(audio_adpcm_encode_data_addr_t *adpcm_encode_data_addr)
{
    u32	host_end_offset;
    host_end_offset = DDRBASE;
    host_end_offset |= (adpcm_encode_data_addr->value.value<<2);
    return host_end_offset;
}

/****************************************************************************************************************
 * Function Name :   audio_adpcm_encode_data_addr_get_adpcm_encode_data_addr_in_ddr_end_offset
 * Description   :   get adpcm encode data addr of ddr end
 * Input         :   param1 is audio adpcm encode's data address management struct
 * Output        :   null
 * Return        :   addr of ddr end
 ****************************************************************************************************************/
static u32  audio_adpcm_encode_data_addr_get_adpcm_encode_data_addr_in_ddr_end_offset(audio_adpcm_encode_data_addr_t *adpcm_encode_data_addr)
{
    u32	ddr_end_offset;
    ddr_end_offset = (adpcm_encode_data_addr->buf_page_id<<19);
    ddr_end_offset |= (adpcm_encode_data_addr->value.value<<2);
    return ddr_end_offset;
}

/****************************************************************************************************************
 * Function Name :   audio_adpcm_encode_data_addr_init
 * Description   :   audio adpcm encode data addr calc init
 * Input         :   param1 is audio adpcm encode's data address management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_adpcm_encode_data_addr_init(audio_adpcm_encode_data_addr_t *adpcm_encode_data_addr, int page_id, int chip_a_or_b)
{
    adpcm_encode_data_addr->value.value = 0;
    adpcm_encode_data_addr->value.bit_value.adpcm_encode_type = ADUIO_DATA_ADDR_ADPCM_ENCODE_TYPE;
    adpcm_encode_data_addr->buf_page_id = page_id;
    adpcm_encode_data_addr->chip_a_or_b = chip_a_or_b;
}

/****************************************************************************************************************
 * Veriable Name :   audio_adpcm_encode_data_addr_op
 * Description   :   for helping op audio adpcm encode data addr calc, we define general interface
 ****************************************************************************************************************/
static struct audio_adpcm_encode_data_addr_operation	audio_adpcm_encode_data_addr_op = {
    .set_reg_bit_section_offset = audio_adpcm_encode_data_addr_set_reg_bit_section_offset,
    .get_reg_bit_section_offset = audio_adpcm_encode_data_addr_get_reg_bit_section_offset,
    .set_reg_bit_section_id = audio_adpcm_encode_data_addr_set_reg_bit_section_id,
    .get_reg_bit_section_id = audio_adpcm_encode_data_addr_get_reg_bit_section_id,
    .set_reg_bit_chan_id = audio_adpcm_encode_data_addr_set_reg_bit_chan_id,
    .get_reg_bit_chan_id = audio_adpcm_encode_data_addr_get_reg_bit_chan_id,
    .set_reg_bit_adpcm_type = audio_adpcm_encode_data_addr_set_reg_bit_adpcm_type,
    .get_reg_bit_adpcm_type = audio_adpcm_encode_data_addr_get_reg_bit_adpcm_type,
    .get_adpcm_encode_data_addr = audio_adpcm_encode_data_addr_get_adpcm_encode_data_addr,
    .get_buf_page_id = audio_adpcm_encode_data_addr_get_buf_page_id,
    .get_buf_in_chip_a_or_b = audio_adpcm_encode_data_addr_get_buf_in_chip_a_or_b,
    .get_adpcm_encode_data_addr_in_host_end_offset = audio_adpcm_encode_data_addr_get_adpcm_encode_data_addr_in_host_end_offset,
    .get_adpcm_encode_data_addr_in_ddr_end_offset = audio_adpcm_encode_data_addr_get_adpcm_encode_data_addr_in_ddr_end_offset,
    .init = audio_adpcm_encode_data_addr_init,
};

/****************************************************************************************************************
 * Function Name :   init_audio_adpcm_encode_data_addr
 * Description   :   init audio adpcm encode data addr calc management struct
 * Input         :   param1 is audio adpcm encode's data address management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void    init_audio_adpcm_encode_data_addr(audio_adpcm_encode_data_addr_t *adpcm_encode_data_addr)
{
    adpcm_encode_data_addr->op = &audio_adpcm_encode_data_addr_op;
    adpcm_encode_data_addr->op->init(adpcm_encode_data_addr, ISIL_AUDIO_BUF_PAGE_ID, ISIL_AUDIO_BUF_OF_DDR_ID);
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_enable_pcm_chan
 * Description   :   enable audio chan's pcm encode
 * Input         :   param1 is audio encode control management struct
 *                   param2 is chan id
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_encode_control_enable_pcm_chan(audio_encode_control_t *encode_control, int phy_chan_id)
{
    if(phy_chan_id < ISIL_PHY_AUDIO_IN_CHAN_NUMBER){
        encode_control->pcm_value.value &= ~(AUDIO_DISABLE<<phy_chan_id);
        encode_control->pcm_value.value |= (AUDIO_ENABLE<<phy_chan_id);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_disable_pcm_chan
 * Description   :   disable audio chan's pcm encode
 * Input         :   param1 is audio encode control management struct
 *                   param2 is chan id
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_encode_control_disable_pcm_chan(audio_encode_control_t *encode_control, int phy_chan_id)
{
    if(phy_chan_id < ISIL_PHY_AUDIO_IN_CHAN_NUMBER){
        encode_control->pcm_value.value &= ~(AUDIO_ENABLE<<phy_chan_id);
        encode_control->pcm_value.value |= (AUDIO_DISABLE<<phy_chan_id);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_set_enable_pcm_chan
 * Description   :   set audio chan's pcm enable
 * Input         :   param1 is audio encode control management struct
 *                   param2 is chan id
 *                   param3 is enable
 *                   param4 is mode
 *                   param5 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_encode_control_set_enable_pcm_chan(audio_encode_control_t *encode_control, int phy_chan_id, int en, int mode, isil_chip_t *chip)
{
    if(phy_chan_id < ISIL_PHY_AUDIO_IN_CHAN_NUMBER){
        if(en == AUDIO_ENABLE){
            encode_control->pcm_value.value &= ~(AUDIO_DISABLE<<phy_chan_id);
            encode_control->pcm_value.value |= (AUDIO_ENABLE<<phy_chan_id);
        } else {
            encode_control->pcm_value.value &= ~(AUDIO_ENABLE<<phy_chan_id);
            encode_control->pcm_value.value |= (AUDIO_DISABLE<<phy_chan_id);
        }
        if(mode == READ_BITSTREAM_BY_PCI){
            encode_control->pcm_value.bit_value.arm_mode_or_pci_mod = 1;
        } else {
            encode_control->pcm_value.bit_value.arm_mode_or_pci_mod = 0;
        }
        if(chip != NULL){
            chip->io_op->chip_write32(chip, encode_control->pcm_reg_offset, encode_control->pcm_value.value);
        }
    }
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_get_pcm_chan_state
 * Description   :   get audio chan's pcm enable state
 * Input         :   param1 is audio encode control management struct
 *                   param2 is chan id
 * Output        :   null
 * Return        :   real state
 ****************************************************************************************************************/
static int  audio_encode_control_get_pcm_chan_state(audio_encode_control_t *encode_control, int phy_chan_id)
{
    int	ret = AUDIO_DISABLE;
    if(phy_chan_id < ISIL_PHY_AUDIO_IN_CHAN_NUMBER){
        if(phy_chan_id>=0){
            ret = (((encode_control->pcm_value.value)>>phy_chan_id)&0x1);
        } else {
            ISIL_DBG(ISIL_DBG_INFO, ": impossible, err chan_id = %d\n", phy_chan_id);
        }
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_init_audio_pcm
 * Description   :   audio encode control init chan pcm enable control management struct
 * Input         :   param1 is audio encode control management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_encode_control_init_audio_pcm(audio_encode_control_t *encode_control)
{
    encode_control->pcm_value.value = 0;
    encode_control->pcm_reg_offset = AUDIO_IN_ORIG_EN_CONTROL_REG;
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_enable_adpcm_chan
 * Description   :   audio encode control init chan adpcm enable control management struct
 * Input         :   param1 is audio encode control management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_encode_control_enable_adpcm_chan(audio_encode_control_t *encode_control, int logic_chan_id)
{
    if(logic_chan_id <= ISIL_AUDIO_IN_CHAN16_ID){
        encode_control->adpcm_value.value &= ~(AUDIO_DISABLE<<logic_chan_id);
        encode_control->adpcm_value.value |= (AUDIO_ENABLE<<logic_chan_id);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_disable_adpcm_chan
 * Description   :   audio encode control disable chan adpcm encode
 * Input         :   param1 is audio encode control management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_encode_control_disable_adpcm_chan(audio_encode_control_t *encode_control, int logic_chan_id)
{
    if(logic_chan_id <= ISIL_AUDIO_IN_CHAN16_ID){
        encode_control->adpcm_value.value &= ~(AUDIO_ENABLE<<logic_chan_id);
        encode_control->adpcm_value.value |= (AUDIO_DISABLE<<logic_chan_id);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_set_enable_adpcm_chan
 * Description   :   update audio encode control chan enable control and update hardware
 * Input         :   param1 is audio encode control management struct
 *                   param2 is logic chan id
 *                   param3 is enable
 *                   param4 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_encode_control_set_enable_adpcm_chan(audio_encode_control_t *encode_control, int logic_chan_id, int en, isil_chip_t *chip)
{
    if(logic_chan_id <= ISIL_AUDIO_IN_CHAN16_ID){
        if(en == AUDIO_ENABLE){
            encode_control->adpcm_value.value &= ~(AUDIO_DISABLE<<logic_chan_id);
            encode_control->adpcm_value.value |= (AUDIO_ENABLE<<logic_chan_id);
        } else {
            encode_control->adpcm_value.value &= ~(AUDIO_ENABLE<<logic_chan_id);
            encode_control->adpcm_value.value |= (AUDIO_DISABLE<<logic_chan_id);
        }
        if(chip != NULL){
            chip->io_op->chip_write32(chip, encode_control->adpcm_reg_offset, encode_control->adpcm_value.value);
        }
    }
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_get_adpcm_chan_state
 * Description   :   get audio chan's adpcm enable state
 * Input         :   param1 is audio encode control management struct
 *                   param2 is chan id
 * Output        :   null
 * Return        :   real state
 ****************************************************************************************************************/
static int  audio_encode_control_get_adpcm_chan_state(audio_encode_control_t *encode_control, int logic_chan_id)
{
    int	ret = AUDIO_DISABLE;
    if(logic_chan_id <= ISIL_AUDIO_IN_CHAN16_ID){
        if(logic_chan_id>=0){
            ret = (((encode_control->adpcm_value.value)>>logic_chan_id)&0x1);
        } else {
            ISIL_DBG(ISIL_DBG_INFO, ": impossible, err chan_id = %d\n", logic_chan_id);
        }
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_init_audio_adpcm
 * Description   :   audio encode control init chan adpcm enable control management struct
 * Input         :   param1 is audio encode control management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_encode_control_init_audio_adpcm(audio_encode_control_t *encode_control)
{
    encode_control->adpcm_value.value = 0;
    encode_control->adpcm_reg_offset = AUDIO_IN_ADPCM_EN_CONTROL_REG;
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_update_audio_sample_rate
 * Description   :   update audio encode's sample rate param
 * Input         :   param1 is audio encode control management struct
 *                   param2 is sample rate
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_encode_control_update_audio_sample_rate(audio_encode_control_t *encode_control, int sample_rate)
{
    encode_control->pcm_value.bit_value.audio_sample_rate = sample_rate;
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_get_audio_sample_rate
 * Description   :   get audio encode's sample rate param
 * Input         :   param1 is audio encode control management struct
 * Output        :   null
 * Return        :   audio sample rate
 ****************************************************************************************************************/
static int  audio_encode_control_get_audio_sample_rate(audio_encode_control_t *encode_control)
{
    return encode_control->pcm_value.bit_value.audio_sample_rate;
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_set_audio_sample_rate
 * Description   :   set audio encode's sample rate param to hardware
 * Input         :   param1 is audio encode control management struct
 *                   param2 is sample rate
 *                   param3 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_encode_control_set_audio_sample_rate(audio_encode_control_t *encode_control, int sample_rate, isil_chip_t *chip)
{
    encode_control->pcm_value.bit_value.audio_sample_rate = sample_rate;
    if(chip != NULL){
        chip->io_op->chip_write32(chip, encode_control->pcm_reg_offset, encode_control->pcm_value.value);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_update_audio_type
 * Description   :   update audio encode's type param
 * Input         :   param1 is audio encode control management struct
 *                   param2 is audio type
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_encode_control_update_audio_type(audio_encode_control_t *encode_control, int audio_type)
{
    encode_control->pcm_value.bit_value.audio_type = audio_type;
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_update_audio_type
 * Description   :   get audio encode's type param
 * Input         :   param1 is audio encode control management struct
 * Output        :   null
 * Return        :   audio type
 ****************************************************************************************************************/
static int  audio_encode_control_get_audio_type(audio_encode_control_t *encode_control)
{
    return encode_control->pcm_value.bit_value.audio_type;
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_set_audio_type
 * Description   :   set audio encode's type param to hardware
 * Input         :   param1 is audio encode control management struct
 *                   param2 is audio type
 *                   param3 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_encode_control_set_audio_type(audio_encode_control_t *encode_control, int audio_type, isil_chip_t *chip)
{
    encode_control->pcm_value.bit_value.audio_type = audio_type;
    if(chip != NULL){
        chip->io_op->chip_write32(chip, encode_control->pcm_reg_offset, encode_control->pcm_value.value);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_update_bit_wide
 * Description   :   update audio encode's bit wide param
 * Input         :   param1 is audio encode control management struct
 *                   param2 is audio's bit wide
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_encode_control_update_bit_wide(audio_encode_control_t *encode_control, int bit_wide)
{
    encode_control->pcm_value.bit_value.AD_BIT_WIDE = bit_wide;
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_get_bit_wide
 * Description   :   get audio encode's bit wide param
 * Input         :   param1 is audio encode control management struct
 * Output        :   null
 * Return        :   bit wide
 ****************************************************************************************************************/
static int  audio_encode_control_get_bit_wide(audio_encode_control_t *encode_control)
{
    return encode_control->pcm_value.bit_value.AD_BIT_WIDE;
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_set_bit_wide
 * Description   :   set audio encode's bit wide param to hardware
 * Input         :   param1 is audio encode control management struct
 *                   param2 is bit wide
 *                   param3 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_encode_control_set_bit_wide(audio_encode_control_t *encode_control, int bit_wide, isil_chip_t *chip)
{
    encode_control->pcm_value.bit_value.AD_BIT_WIDE = bit_wide;
    if(chip != NULL){
        chip->io_op->chip_write32(chip, encode_control->pcm_reg_offset, encode_control->pcm_value.value);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_pci_sel_pcm_or_adpcm
 * Description   :   select audio encode work in pci push mode or host master read mode
 * Input         :   param1 is audio encode control management struct
 *                   param2 is chan id
 *                   param3 is type
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void	audio_encode_control_pci_sel_pcm_or_adpcm(audio_encode_control_t *encode_control, int logic_chan_id, int type)
{
    if(type == ISIL_AUDIO_PCM){
        encode_control->sel_pcm_or_adpcm_based_pci_push.value &= ~(0xffff);
    } else {
        encode_control->sel_pcm_or_adpcm_based_pci_push.value |= (0xffff);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_set_pci_sel_pcm_or_adpcm
 * Description   :   select audio encode type is pcm or adpcm based pci push mode
 * Input         :   param1 is audio encode control management struct
 *                   param2 is chan id
 *                   param3 is type
 *                   param4 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void	audio_encode_control_set_pci_sel_pcm_or_adpcm(audio_encode_control_t *encode_control, int logic_chan_id, int type, isil_chip_t *chip)
{
    if(type == ISIL_AUDIO_PCM){
        encode_control->sel_pcm_or_adpcm_based_pci_push.value &= ~(0xffff);
    } else {
        encode_control->sel_pcm_or_adpcm_based_pci_push.value |= (0xffff);
    }

    if(chip != NULL){
        chip->io_op->chip_write32(chip, encode_control->sel_pci_pcm_adpcm_reg_offset, encode_control->sel_pcm_or_adpcm_based_pci_push.value);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_set_toggle_pci_flowcn
 * Description   :   send toggle signal to hardware for flowing control
 * Input         :   param1 is audio encode control management struct
 *                   param2 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void     audio_encode_control_set_toggle_pci_flowcn(audio_encode_control_t *encode_control, isil_chip_t *chip)
{
    if(chip != NULL){
        encode_control->sel_pcm_or_adpcm_based_pci_push.bit_value.pci_audio_flow_en = 1;
        encode_control->sel_pcm_or_adpcm_based_pci_push.bit_value.pci_audio_frame_toggle = 1;
        chip->io_op->chip_write32(chip, encode_control->sel_pci_pcm_adpcm_reg_offset, encode_control->sel_pcm_or_adpcm_based_pci_push.value);
        encode_control->sel_pcm_or_adpcm_based_pci_push.bit_value.pci_audio_frame_toggle = 0;
        chip->io_op->chip_write32(chip, encode_control->sel_pci_pcm_adpcm_reg_offset, encode_control->sel_pcm_or_adpcm_based_pci_push.value);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_init_audio_pci_sel
 * Description   :   init audio pci sel management struct
 * Input         :   param1 is audio encode control management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void	audio_encode_control_init_audio_pci_sel(audio_encode_control_t *encode_control)
{
    encode_control->sel_pci_pcm_adpcm_reg_offset = PCI_AUDIO_CHAN_SELECT_REG;
    encode_control->sel_pcm_or_adpcm_based_pci_push.value = 0xffff;
    encode_control->sel_pcm_or_adpcm_based_pci_push.bit_value.pci_audio_flow_en = 1;
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_set_audio_param
 * Description   :   set audio encode param to hardware
 * Input         :   param1 is audio encode control management struct
 *                   param2 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_encode_control_set_audio_param(audio_encode_control_t *encode_control, isil_chip_t *chip)
{
    if(chip != NULL){
        chip->io_op->chip_write32(chip, encode_control->pcm_reg_offset, encode_control->pcm_value.value);
        chip->io_op->chip_write32(chip, encode_control->adpcm_reg_offset, encode_control->adpcm_value.value);
        chip->io_op->chip_write32(chip, encode_control->sel_pci_pcm_adpcm_reg_offset, encode_control->sel_pcm_or_adpcm_based_pci_push.value);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_encode_control_get_audio_param
 * Description   :   get audio encode param to hardware
 * Input         :   param1 is audio encode control management struct
 *                   param2 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_encode_control_get_audio_param(audio_encode_control_t *encode_control, isil_chip_t *chip)
{
    if(chip != NULL){
        //encode_control->pcm_value.value = chip->io_op->chip_read32(chip, encode_control->pcm_reg_offset);
        //encode_control->adpcm_value.value = chip->io_op->chip_read32(chip, encode_control->adpcm_reg_offset);
    }
}

/****************************************************************************************************************
 * Veriable Name :   audio_encode_control_op
 * Description   :   for helping op audio encode control, we define general interface
 ****************************************************************************************************************/
static struct audio_encode_control_operation    audio_encode_control_op = {
    .enable_pcm_chan = audio_encode_control_enable_pcm_chan,
    .disable_pcm_chan = audio_encode_control_disable_pcm_chan,
    .set_enable_pcm_chan = audio_encode_control_set_enable_pcm_chan,
    .get_pcm_chan_state = audio_encode_control_get_pcm_chan_state,
    .init_audio_pcm = audio_encode_control_init_audio_pcm,

    .enable_adpcm_chan = audio_encode_control_enable_adpcm_chan,
    .disable_adpcm_chan = audio_encode_control_disable_adpcm_chan,
    .set_enable_adpcm_chan = audio_encode_control_set_enable_adpcm_chan,
    .get_adpcm_chan_state = audio_encode_control_get_adpcm_chan_state,
    .init_audio_adpcm = audio_encode_control_init_audio_adpcm,

    .update_audio_sample_rate = audio_encode_control_update_audio_sample_rate,
    .get_audio_sample_rate = audio_encode_control_get_audio_sample_rate,
    .set_audio_sample_rate = audio_encode_control_set_audio_sample_rate,
    .update_audio_type = audio_encode_control_update_audio_type,
    .get_audio_type = audio_encode_control_get_audio_type,
    .set_audio_type = audio_encode_control_set_audio_type,
    .update_bit_wide = audio_encode_control_update_bit_wide,
    .get_bit_wide= audio_encode_control_get_bit_wide,
    .set_bit_wide = audio_encode_control_set_bit_wide,

    .pci_sel_pcm_or_adpcm = audio_encode_control_pci_sel_pcm_or_adpcm,
    .set_enable_pci_sel_pcm_or_adpcm = audio_encode_control_set_pci_sel_pcm_or_adpcm,
    .toggle_pci_flowcn = audio_encode_control_set_toggle_pci_flowcn,
    .init_audio_pci_sel = audio_encode_control_init_audio_pci_sel,

    .set_audio_param = audio_encode_control_set_audio_param,
    .get_audio_param = audio_encode_control_get_audio_param,
};

/****************************************************************************************************************
 * Function Name :   init_audio_encode_control
 * Description   :   init audio encode control
 * Input         :   param1 is audio encode control management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void	init_audio_encode_control(audio_encode_control_t *encode_control)
{
    encode_control->op = &audio_encode_control_op;
    encode_control->op->init_audio_pcm(encode_control);
    encode_control->op->init_audio_adpcm(encode_control);
    encode_control->op->init_audio_pci_sel(encode_control);
}

/****************************************************************************************************************
 * Function Name :   audio_decode_control_init
 * Description   :   audio decode control's init
 * Input         :   param1 is audio decode control management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_decode_control_init(audio_decode_control_t *decode_control)
{
    decode_control->value.value = 0;
    decode_control->reg_offset = AUDIO_OUT_CONTROL_REG;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_control_update_decode_chan_mute
 * Description   :   update audio decode chan's mute
 * Input         :   param1 is audio decode control management struct
 *                   param2 is chan id
 *                   param3 is mute
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_decode_control_update_decode_chan_mute(audio_decode_control_t *decode_control, int chan_id, int mute)
{
    switch(chan_id){
        case ISIL_AUDIO_OUT_SPEAKER_ID:
            decode_control->value.bit_value.audio_out_chan0_mute = mute;
            break;
        case ISIL_AUDIO_OUT_PLAYBACK_ID:
            decode_control->value.bit_value.audio_out_chan1_mute = mute;
            break;
        default:
            break;
    }
}

/****************************************************************************************************************
 * Function Name :   audio_decode_control_get_decode_chan_mute
 * Description   :   get audio decode chan's mute
 * Input         :   param1 is audio decode control management struct
 *                   param2 is chan id
 * Output        :   null
 * Return        :   mute
 ****************************************************************************************************************/
static int  audio_decode_control_get_decode_chan_mute(audio_decode_control_t *decode_control, int chan_id)
{
    int	ret = AUDIO_MUTE_ENABLE;
    switch(chan_id){
        case ISIL_AUDIO_OUT_SPEAKER_ID:
            ret = decode_control->value.bit_value.audio_out_chan0_mute;
            break;
        case ISIL_AUDIO_OUT_PLAYBACK_ID:
            ret = decode_control->value.bit_value.audio_out_chan1_mute;
            break;
        default:
            break;
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_control_set_decode_chan_mute
 * Description   :   set audio decode chan's mute
 * Input         :   param1 is audio decode control management struct
 *                   param2 is chan id
 *                   param3 is mute
 *                   param4 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_decode_control_set_decode_chan_mute(audio_decode_control_t *decode_control, int chan_id, int mute, isil_chip_t *chip)
{
    if(chip != NULL){
        switch(chan_id){
            case ISIL_AUDIO_OUT_SPEAKER_ID:
                decode_control->value.bit_value.audio_out_chan0_mute = mute;
                break;
            case ISIL_AUDIO_OUT_PLAYBACK_ID:
                decode_control->value.bit_value.audio_out_chan1_mute = mute;
                break;
            default:
                return;
        }
        chip->io_op->chip_write32(chip, decode_control->reg_offset, decode_control->value.value);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_decode_control_update_decode_chan_bit_wide
 * Description   :   update audio decode chan's bit wide
 * Input         :   param1 is audio decode control management struct
 *                   param2 is chan id
 *                   param3 is bit wide
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_decode_control_update_decode_chan_bit_wide(audio_decode_control_t *decode_control, int chan_id, int bit_wide)
{
    switch(chan_id){
        case ISIL_AUDIO_OUT_SPEAKER_ID:
            decode_control->value.bit_value.audio_out_chan0_DA_BIT_WIDE = bit_wide;
            break;
        case ISIL_AUDIO_OUT_PLAYBACK_ID:
            decode_control->value.bit_value.audio_out_chan1_DA_BIT_WIDE = bit_wide;
            break;
        default:
            break;
    }
}

/****************************************************************************************************************
 * Function Name :   audio_decode_control_get_decode_chan_bit_wide
 * Description   :   get audio decode chan's bit wide
 * Input         :   param1 is audio decode control management struct
 *                   param2 is chan id
 * Output        :   null
 * Return        :   bit wide
 ****************************************************************************************************************/
static int  audio_decode_control_get_decode_chan_bit_wide(audio_decode_control_t *decode_control, int chan_id)
{
    int	ret = ISIL_AUDIO_16BIT;
    switch(chan_id){
        case ISIL_AUDIO_OUT_SPEAKER_ID:
            ret = decode_control->value.bit_value.audio_out_chan0_DA_BIT_WIDE;
            break;
        case ISIL_AUDIO_OUT_PLAYBACK_ID:
            ret = decode_control->value.bit_value.audio_out_chan1_DA_BIT_WIDE;
            break;
        default:
            break;
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_control_set_decode_chan_bit_wide
 * Description   :   set audio decode chan's bit wide
 * Input         :   param1 is audio decode control management struct
 *                   param2 is chan id
 *                   param3 is bit_wide
 *                   param4 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_decode_control_set_decode_chan_bit_wide(audio_decode_control_t *decode_control, int chan_id, int bit_wide, isil_chip_t *chip)
{
    if(chip != NULL){
        switch(chan_id){
            case ISIL_AUDIO_OUT_SPEAKER_ID:
                decode_control->value.bit_value.audio_out_chan0_DA_BIT_WIDE = bit_wide;
                break;
            case ISIL_AUDIO_OUT_PLAYBACK_ID:
                decode_control->value.bit_value.audio_out_chan1_DA_BIT_WIDE = bit_wide;
                break;
            default:
                return;
        }
        chip->io_op->chip_write32(chip, decode_control->reg_offset, decode_control->value.value);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_decode_control_update_decode_chan_enable
 * Description   :   update audio decode chan's enable
 * Input         :   param1 is audio decode control management struct
 *                   param2 is chan id
 *                   param3 is enable
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_decode_control_update_decode_chan_enable(audio_decode_control_t *decode_control, int chan_id, int enable)
{
    switch(chan_id){
        case ISIL_AUDIO_OUT_SPEAKER_ID:
            decode_control->value.bit_value.audio_out_chan0_enable = enable;
            break;
        case ISIL_AUDIO_OUT_PLAYBACK_ID:
            decode_control->value.bit_value.audio_out_chan1_enable = enable;
            break;
        default:
            break;
    }
}

/****************************************************************************************************************
 * Function Name :   audio_decode_control_get_decode_chan_enable
 * Description   :   get audio decode chan's enable
 * Input         :   param1 is audio decode control management struct
 *                   param2 is chan id
 * Output        :   null
 * Return        :   enable
 ****************************************************************************************************************/
static int  audio_decode_control_get_decode_chan_enable(audio_decode_control_t *decode_control, int chan_id)
{
    int	ret = AUDIO_DISABLE;
    switch(chan_id){
        case ISIL_AUDIO_OUT_SPEAKER_ID:
            ret = decode_control->value.bit_value.audio_out_chan0_enable;
            break;
        case ISIL_AUDIO_OUT_PLAYBACK_ID:
            ret = decode_control->value.bit_value.audio_out_chan1_enable;
            break;
        default:
            break;
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_control_set_decode_chan_enable
 * Description   :   set audio decode chan's enable
 * Input         :   param1 is audio decode control management struct
 *                   param2 is chan id
 *                   param3 is enable
 *                   param4 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_decode_control_set_decode_chan_enable(audio_decode_control_t *decode_control, int chan_id, int enable, isil_chip_t *chip)
{
    if(chip != NULL){
        switch(chan_id){
            case ISIL_AUDIO_OUT_SPEAKER_ID:
                decode_control->value.bit_value.audio_out_chan0_enable = enable;
                break;
            case ISIL_AUDIO_OUT_PLAYBACK_ID:
                decode_control->value.bit_value.audio_out_chan1_enable = enable;
                break;
            default:
                return;
        }
        chip->io_op->chip_write32(chip, decode_control->reg_offset, decode_control->value.value);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_decode_control_set_decode_param
 * Description   :   set audio decode param
 * Input         :   param1 is audio decode control management struct
 *                   param2 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_decode_control_set_decode_param(audio_decode_control_t *decode_control, isil_chip_t *chip)
{
    if(chip != NULL){
        chip->io_op->chip_write32(chip, decode_control->reg_offset, decode_control->value.value);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_decode_control_set_decode_param
 * Description   :   get audio decode param
 * Input         :   param1 is audio decode control management struct
 *                   param2 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_decode_control_get_decode_param(audio_decode_control_t *decode_control, isil_chip_t *chip)
{
    if(chip != NULL){
        //decode_control->value.value = chip->io_op->chip_read32(chip, decode_control->reg_offset);
    }
}

/****************************************************************************************************************
 * Veriable Name :   audio_decode_control_op
 * Description   :   for helping op audio decode control, we define general interface
 ****************************************************************************************************************/
static struct audio_decode_control_operation	audio_decode_control_op = {
    .init = audio_decode_control_init,
    .update_decode_chan_mute = audio_decode_control_update_decode_chan_mute,
    .get_decode_chan_mute = audio_decode_control_get_decode_chan_mute,
    .set_decode_chan_mute = audio_decode_control_set_decode_chan_mute,
    .update_decode_chan_bit_wide = audio_decode_control_update_decode_chan_bit_wide,
    .get_decode_chan_bit_wide = audio_decode_control_get_decode_chan_bit_wide,
    .set_decode_chan_bit_wide = audio_decode_control_set_decode_chan_bit_wide,
    .update_decode_chan_enable = audio_decode_control_update_decode_chan_enable,
    .get_decode_chan_enable = audio_decode_control_get_decode_chan_enable,
    .set_decode_chan_enable = audio_decode_control_set_decode_chan_enable,

    .set_decode_param = audio_decode_control_set_decode_param,
    .get_decode_param = audio_decode_control_get_decode_param,
};

/****************************************************************************************************************
 * Function Name :   init_audio_decode_control
 * Description   :   init audio decode control
 * Input         :   param1 is audio decode control management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void	init_audio_decode_control(audio_decode_control_t *decode_control)
{
    decode_control->op = &audio_decode_control_op;
    decode_control->op->init(decode_control);
}

/****************************************************************************************************************
 * Function Name :   audio_decode_ptr_control_init
 * Description   :   audio decode ptr control management init
 * Input         :   param1 is audio decode ptr control management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_decode_ptr_control_init(audio_decode_ptr_control_t *decode_ptr)
{
    decode_ptr->wr_rd_value.value = 0;
    decode_ptr->wr_rd_reg_offset = ADPCM_DECODE_RD_WR_PTR_REG;
    decode_ptr->section_number = AUDIO_OUT_CHAN_SECTION_NUMBER;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_ptr_control_get_adpcm_wr_rd_ptr
 * Description   :   get audio decode wr&rd ptr
 * Input         :   param1 is audio decode ptr control management struct
 *                   param2 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_decode_ptr_control_get_adpcm_wr_rd_ptr(audio_decode_ptr_control_t *decode_ptr, isil_chip_t *chip)
{
    if(chip != NULL){
        decode_ptr->wr_rd_value.value = chip->io_op->chip_read32(chip, decode_ptr->wr_rd_reg_offset);

        //ISIL_DBG(ISIL_DBG_FATAL, "%08x\n", decode_ptr->wr_rd_value.value);
    }
}

/****************************************************************************************************************
 * Function Name :   audio_decode_ptr_control_get_chan_wr_ptr
 * Description   :   get audio decode wr ptr
 * Input         :   param1 is audio decode ptr control management struct
 *                   param2 is chan id
 * Output        :   ext flag
 * Return        :   wr ptr
 ****************************************************************************************************************/
static int  audio_decode_ptr_control_get_chan_wr_ptr(audio_decode_ptr_control_t *decode_ptr, int chan_id, int *ext_flag)
{
    int	ret = INVALID_ISIL_AUDIO_SECTION_ID;
    switch(chan_id){
        case ISIL_AUDIO_OUT_SPEAKER_ID:
            ret = decode_ptr->wr_rd_value.bit_value.chan0_wr_ptr;
            break;
        case ISIL_AUDIO_OUT_PLAYBACK_ID:
            ret = decode_ptr->wr_rd_value.bit_value.chan1_wr_ptr;
            break;
        default:
            ISIL_DBG(ISIL_DBG_INFO, ": err audio chan id %d\n", chan_id);
            break;
    }
    *ext_flag = ret>>(get_power_base((u32)AUDIO_OUT_CHAN_SECTION_NUMBER));
    ret &= ((1<<(get_power_base((u32)AUDIO_OUT_CHAN_SECTION_NUMBER)))-1);
    return ret;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_ptr_control_get_chan_rd_ptr
 * Description   :   get audio decode rd ptr
 * Input         :   param1 is audio decode ptr control management struct
 *                   param2 is chan id
 * Output        :   ext flag
 * Return        :   rd ptr
 ****************************************************************************************************************/
static int  audio_decode_ptr_control_get_chan_rd_ptr(audio_decode_ptr_control_t *decode_ptr, int chan_id, int *ext_flag)
{
    int	ret = INVALID_ISIL_AUDIO_SECTION_ID;
    switch(chan_id){
        case ISIL_AUDIO_OUT_SPEAKER_ID:
            ret = decode_ptr->wr_rd_value.bit_value.chan0_rd_ptr;
            break;
        case ISIL_AUDIO_OUT_PLAYBACK_ID:
            ret = decode_ptr->wr_rd_value.bit_value.chan1_rd_ptr;
            break;
        default:
            ISIL_DBG(ISIL_DBG_INFO, ": err audio chan id %d\n", chan_id);
            break;
    }
    *ext_flag = ret>>(get_power_base((u32)AUDIO_OUT_CHAN_SECTION_NUMBER));    
    ret &= ((1<<(get_power_base((u32)AUDIO_OUT_CHAN_SECTION_NUMBER)))-1);
    return ret;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_ptr_control_get_chan_section_number
 * Description   :   get audio decode ptr section number
 * Input         :   param1 is audio decode ptr control management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static int  audio_decode_ptr_control_get_chan_section_number(audio_decode_ptr_control_t *decode_ptr)
{
    return decode_ptr->section_number;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_ptr_control_set_finish_wr_ptr
 * Description   :   update audio decode chan's finish id
 * Input         :   param1 is audio decode ptr control management struct
 *                   param2 is finish id
 *                   param3 is chip management struct
 *                   param4 is chan id
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_decode_ptr_control_set_finish_wr_ptr(audio_decode_ptr_control_t *decode_ptr, int finish_id, isil_chip_t *chip, int chan_id)
{
    if(chip != NULL){
        finish_id++;
        if(finish_id >= (decode_ptr->section_number<<1)){
            finish_id = 0;
        }
        if(chan_id == ISIL_AUDIO_OUT_SPEAKER_ID){
            decode_ptr->wr_rd_value.bit_value.chan0_wr_ptr = finish_id;
        } else {
            decode_ptr->wr_rd_value.bit_value.chan1_wr_ptr = finish_id;       
        }
        chip->io_op->chip_write32(chip, decode_ptr->wr_rd_reg_offset, decode_ptr->wr_rd_value.value);
    }
}

/****************************************************************************************************************
 * Veriable Name :   audio_decode_ptr_control_op
 * Description   :   for helping op audio decode ptr control, we define general interface
 ****************************************************************************************************************/
struct audio_decode_ptr_control_operation   audio_decode_ptr_control_op = {
    .init = audio_decode_ptr_control_init,
    .get_adpcm_wr_rd_ptr = audio_decode_ptr_control_get_adpcm_wr_rd_ptr,

    .get_chan_wr_ptr = audio_decode_ptr_control_get_chan_wr_ptr,
    .get_chan_rd_ptr = audio_decode_ptr_control_get_chan_rd_ptr,
    .get_chan_section_number = audio_decode_ptr_control_get_chan_section_number,
    .set_finish_wr_ptr = audio_decode_ptr_control_set_finish_wr_ptr,
};

/****************************************************************************************************************
 * Function Name :   init_audio_decode_ptr_control
 * Description   :   init audio decode ptr control
 * Input         :   param1 is audio decode ptr control management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void	init_audio_decode_ptr_control(audio_decode_ptr_control_t *decode_ptr)
{
    decode_ptr->op = &audio_decode_ptr_control_op;
    decode_ptr->op->init(decode_ptr);
}

/****************************************************************************************************************
 * Function Name :   audio_decode_data_addr_set_reg_bit_section_offset
 * Description   :   update decode data addr's section offset
 * Input         :   param1 is audio decode data addr calc management struct,
 *                   param2 is section offset
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_decode_data_addr_set_reg_bit_section_offset(audio_decode_data_addr_t *decode_data_addr, u32 section_offset)
{
    decode_data_addr->value.bit_value.section_offset = section_offset;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_data_addr_get_reg_bit_section_offset
 * Description   :   get decode data addr's section offset
 * Input         :   param1 is audio decode data addr calc management struct,
 * Output        :   null
 * Return        :   section offset
 ****************************************************************************************************************/
static u32  audio_decode_data_addr_get_reg_bit_section_offset(audio_decode_data_addr_t *decode_data_addr)
{
    return decode_data_addr->value.bit_value.section_offset;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_data_addr_set_reg_bit_section_id
 * Description   :   update decode data addr's section id
 * Input         :   param1 is audio decode data addr calc management struct,
 *                   param2 is section id
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_decode_data_addr_set_reg_bit_section_id(audio_decode_data_addr_t *decode_data_addr, u32 section_id)
{
    decode_data_addr->value.bit_value.section_id = section_id;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_data_addr_get_reg_bit_section_id
 * Description   :   update decode data addr's section id
 * Input         :   param1 is audio decode data addr calc management struct,
 * Output        :   null
 * Return        :   section id
 ****************************************************************************************************************/
static u32  audio_decode_data_addr_get_reg_bit_section_id(audio_decode_data_addr_t *decode_data_addr)
{
    return decode_data_addr->value.bit_value.section_id;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_data_addr_set_reg_bit_section_id
 * Description   :   update decode data addr's section id
 * Input         :   param1 is audio decode data addr calc management struct,
 *                   param2 is section id
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_decode_data_addr_set_reg_bit_chan_id(audio_decode_data_addr_t *decode_data_addr, u32 chan_id)
{
    decode_data_addr->value.bit_value.chan_id = chan_id;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_data_addr_get_reg_bit_chan_id
 * Description   :   get decode data addr's section id
 * Input         :   param1 is audio decode data addr calc management struct,
 * Output        :   null
 * Return        :   chan id
 ****************************************************************************************************************/
static u32  audio_decode_data_addr_get_reg_bit_chan_id(audio_decode_data_addr_t *decode_data_addr)
{
    return decode_data_addr->value.bit_value.chan_id;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_data_addr_set_reg_bit_decode_type
 * Description   :   set decode data type
 * Input         :   param1 is audio decode data addr calc management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_decode_data_addr_set_reg_bit_decode_type(audio_decode_data_addr_t *decode_data_addr)
{
    decode_data_addr->value.bit_value.decode_type = AUDIO_DATA_ADDR_ADPCM_DECODE_TYPE;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_data_addr_set_reg_bit_decode_type
 * Description   :   get decode data type
 * Input         :   param1 is audio decode data addr calc management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static u32  audio_decode_data_addr_get_reg_bit_decode_type(audio_decode_data_addr_t *decode_data_addr)
{
    return decode_data_addr->value.bit_value.decode_type;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_data_addr_get_decode_data_addr
 * Description   :   get decode data addr
 * Input         :   param1 is audio decode data addr calc management struct,
 * Output        :   null
 * Return        :   data addr
 ****************************************************************************************************************/
static u32  audio_decode_data_addr_get_decode_data_addr(audio_decode_data_addr_t *decode_data_addr)
{
    return decode_data_addr->value.value;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_data_addr_get_buf_page_id
 * Description   :   get decode data's page id
 * Input         :   param1 is audio decode data addr calc management struct,
 * Output        :   null
 * Return        :   buf page id
 ****************************************************************************************************************/
static int  audio_decode_data_addr_get_buf_page_id(audio_decode_data_addr_t *decode_data_addr)
{
    return decode_data_addr->buf_page_id;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_data_addr_get_buf_in_chip_a_or_b
 * Description   :   get decode data's in chip a or b
 * Input         :   param1 is audio decode data addr calc management struct,
 * Output        :   null
 * Return        :   chip_a_or_b
 ****************************************************************************************************************/
static int  audio_decode_data_addr_get_buf_in_chip_a_or_b(audio_decode_data_addr_t *decode_data_addr)
{
    return decode_data_addr->chip_a_or_b;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_data_addr_get_adpcm_decode_data_addr_in_host_end_offset
 * Description   :   get decode data's host end addr
 * Input         :   param1 is audio decode data addr calc management struct,
 * Output        :   null
 * Return        :   host end offset
 ****************************************************************************************************************/
static u32  audio_decode_data_addr_get_adpcm_decode_data_addr_in_host_end_offset(audio_decode_data_addr_t *decode_data_addr)
{
    u32	host_end_offset;
    host_end_offset = DDRBASE;
    host_end_offset |= (decode_data_addr->value.value<<2);
    return host_end_offset;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_data_addr_get_adpcm_decode_data_addr_in_ddr_end_offset
 * Description   :   get decode data's ddr end addr
 * Input         :   param1 is audio decode data addr calc management struct,
 * Output        :   null
 * Return        :   ddr end offset
 ****************************************************************************************************************/
static u32  audio_decode_data_addr_get_adpcm_decode_data_addr_in_ddr_end_offset(audio_decode_data_addr_t *decode_data_addr)
{
    u32	ddr_end_offset;
    ddr_end_offset = (decode_data_addr->buf_page_id<<19);
    ddr_end_offset |= (decode_data_addr->value.value<<2);

    return ddr_end_offset;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_data_addr_init
 * Description   :   audio decode data addr init
 * Input         :   param1 is audio decode data addr calc management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void audio_decode_data_addr_init(audio_decode_data_addr_t *decode_data_addr)
{
    if(decode_data_addr != NULL){
        decode_data_addr->value.value = 0;
        decode_data_addr->value.bit_value.decode_type = AUDIO_DATA_ADDR_ADPCM_DECODE_TYPE;
        decode_data_addr->buf_page_id = ISIL_AUDIO_BUF_PAGE_ID;
        decode_data_addr->chip_a_or_b = ISIL_AUDIO_BUF_OF_DDR_ID;
    }
}

/****************************************************************************************************************
 * Veriable Name :   audio_decode_data_addr_op
 * Description   :   for helping op audio decode data addr, we define general interface
 ****************************************************************************************************************/
static struct audio_decode_data_addr_operation	audio_decode_data_addr_op = {
    .set_reg_bit_section_offset = audio_decode_data_addr_set_reg_bit_section_offset,
    .get_reg_bit_section_offset = audio_decode_data_addr_get_reg_bit_section_offset,
    .set_reg_bit_section_id = audio_decode_data_addr_set_reg_bit_section_id,
    .get_reg_bit_section_id = audio_decode_data_addr_get_reg_bit_section_id,
    .set_reg_bit_chan_id = audio_decode_data_addr_set_reg_bit_chan_id,
    .get_reg_bit_chan_id = audio_decode_data_addr_get_reg_bit_chan_id,
    .set_reg_bit_decode_type = audio_decode_data_addr_set_reg_bit_decode_type,
    .get_reg_bit_decode_type = audio_decode_data_addr_get_reg_bit_decode_type,
    .get_decode_data_addr = audio_decode_data_addr_get_decode_data_addr,
    .get_buf_page_id = audio_decode_data_addr_get_buf_page_id,
    .get_buf_in_chip_a_or_b = audio_decode_data_addr_get_buf_in_chip_a_or_b,
    .get_adpcm_decode_data_addr_in_host_end_offset = audio_decode_data_addr_get_adpcm_decode_data_addr_in_host_end_offset,
    .get_adpcm_decode_data_addr_in_ddr_end_offset = audio_decode_data_addr_get_adpcm_decode_data_addr_in_ddr_end_offset,
    .init = audio_decode_data_addr_init,
};

/****************************************************************************************************************
 * Function Name :   init_audio_decode_data_addr
 * Description   :   init audio decode data addr
 * Input         :   param1 is audio decode data addr calc management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void	init_audio_decode_data_addr(audio_decode_data_addr_t *decode_data_addr)
{
    decode_data_addr->op = &audio_decode_data_addr_op;
    decode_data_addr->op->init(decode_data_addr);
}

/****************************************************************************************************************
 * Function Name :   chip_audio_encode_polling_task
 * Description   :   chip audio encode polling task
 * Input         :   param1 is chip audio encode management struct,
 * Output        :   null
 * Return        :   always ISIL_OK
 ****************************************************************************************************************/
static int	chip_audio_encode_polling_task(void *context)
{
    chip_audio_encode_t *audio_encode = (chip_audio_encode_t*)context;
    if(audio_encode != NULL){
        chip_audio_t    *chip_audio;

        chip_audio = to_get_chip_audio_with_audio_encode(audio_encode);
        chip_audio->op->process_audio_encode(chip_audio);
    }
    return ISIL_OK;
}

/*******************************************************************************
 * Function Name  : add_chip_audio_encode_polling_task
 * Description	 : add chip audio encode polling task
 * Input 		 : param1 is chip audio encode management struct
 * Output		 : NULL
 * Return		 : NULL
 *******************************************************************************/
static void add_chip_audio_encode_polling_task(chip_audio_encode_t *audio_encode)
{
    if(audio_encode != NULL){
        chip_audio_t    *chip_audio;
        isil_chip_t		*chip;
        isil_chip_timer_controller_t	*chip_timer_cont;

        chip_audio = to_get_chip_audio_with_audio_encode(audio_encode);
        chip = chip_audio->chip;
        chip_timer_cont = &chip->chip_timer_cont;
        audio_encode->delete_chip_audio_encode_polling_task(audio_encode);
        audio_encode->chip_aduio_encode_polling_timeid = chip_timer_cont->op->AddForFireTimerJob(chip_timer_cont, CHIP_POLLING_TASK_COUNTER, chip_audio_encode_polling_task, audio_encode);
        if(audio_encode->chip_aduio_encode_polling_timeid == ADDJOBERROR){
            ISIL_DBG(ISIL_DBG_INFO, ": add timer hook err\n");
        }
    }
}

/****************************************************************************************************************
 * Function Name :	delete_chip_audio_encode_polling_task
 * Description	:  delete chip audio encode polling task
 * Input 		 : param1 is chip audio encode management struct
 * Output		:	null
 * Return		:	null
 ****************************************************************************************************************/
static void delete_chip_audio_encode_polling_task(chip_audio_encode_t *audio_encode)
{
    if(audio_encode != NULL){
        chip_audio_t    *chip_audio;
        isil_chip_t		*chip;
        isil_chip_timer_controller_t	*chip_timer_cont;

        chip_audio = to_get_chip_audio_with_audio_encode(audio_encode);
        chip = chip_audio->chip;
        chip_timer_cont = &chip->chip_timer_cont;
        chip_timer_cont->op->DeleteForFireTimerJob(chip_timer_cont, audio_encode->chip_aduio_encode_polling_timeid);
        audio_encode->chip_aduio_encode_polling_timeid = INVALIDTIMERID;
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_encode_init
 * Description   :   init chip audio encode
 * Input         :   param1 is chip audio encode management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_encode_init(chip_audio_encode_t *audio_encode)
{
    if(audio_encode != NULL){
        atomic_set(&audio_encode->need_sync_param, 1);
        atomic_set(&audio_encode->opened_encode_chan_number, 0);
        memcpy(&audio_encode->config_param, &default_audio_encode_config_descriptor, sizeof(isil_audio_config_descriptor_t));
        init_audio_encode_property(&audio_encode->running_param, &default_audio_encode_config_descriptor);
        init_audio_encode_control(&audio_encode->audio_control);
        init_audio_pcm_encode_ptr_control(&audio_encode->pcm_audio_ptr);
        init_audio_pcm_encode_data_addr(&audio_encode->pcm_audio_data_base);
        init_audio_adpcm_encode_ptr_control(&audio_encode->adpcm_audio_ptr);
        init_audio_adpcm_encode_data_addr(&audio_encode->adpcm_audio_data_base);
        audio_encode->chip_aduio_encode_polling_timeid = INVALIDTIMERID;
        audio_encode->add_chip_audio_encode_polling_task = add_chip_audio_encode_polling_task;
        audio_encode->delete_chip_audio_encode_polling_task = delete_chip_audio_encode_polling_task;
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_encode_set_enable_audio_encode_chan
 * Description   :   chip audio encode enable audio encode chan
 * Input         :   param1 is chip audio encode management struct,
 *                   param2 is chan_id,
 *                   param3 is enable,
 *                   param4 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_encode_set_enable_audio_encode_chan(chip_audio_encode_t *audio_encode, int chan_id, int enable, isil_chip_t *chip)
{
    audio_encode_control_t  *audio_control;
    isil_audio_param_t      *running_param;
    chip_audio_t            *chip_audio;
    isil_audio_driver_t     *audio_driver;
    int     phy_chan_id;

    audio_control = &audio_encode->audio_control;
    chip_audio = to_get_chip_audio_with_audio_encode(audio_encode);
    chip_audio->op->get_audio_chan_driver(chip_audio, chan_id, &audio_driver);
    if(audio_driver != NULL){
        phy_chan_id = audio_driver->audio_phy_chan_id;
    } else {
        phy_chan_id = chan_id;
    }
    running_param = &audio_encode->running_param;
    if(running_param->op != NULL){
        audio_enable_control_t  *audio_enable = &chip_audio->audio_enable;
        if((audio_enable->op!=NULL) && (running_param->op!=NULL)){
            switch(running_param->op->get_type(running_param)){
                case ISIL_AUDIO_PCM:
                    if(chip_audio->read_bitstream_mode == READ_BITSTREAM_BY_PCI){
                        audio_control->op->set_enable_pci_sel_pcm_or_adpcm(audio_control, chan_id, ISIL_AUDIO_PCM, chip);
                        if(enable == AUDIO_ENABLE){
                            audio_enable->op->set_enable_pcm_encode(audio_enable, AUDIO_ENABLE, chip);
                            audio_control->op->set_enable_pcm_chan(audio_control, phy_chan_id, AUDIO_ENABLE, chip_audio->read_bitstream_mode, chip);
                            audio_enable->op->set_enable_adpcm_encode(audio_enable, AUDIO_ENABLE, chip);
                            audio_control->op->set_enable_adpcm_chan(audio_control, chan_id, AUDIO_ENABLE, chip);
                        } else if(enable == AUDIO_DISABLE){
                            audio_control->op->set_enable_adpcm_chan(audio_control, chan_id, AUDIO_DISABLE, chip);
                            audio_enable->op->set_enable_adpcm_encode(audio_enable, AUDIO_DISABLE, chip);
                            audio_control->op->set_enable_pcm_chan(audio_control, phy_chan_id, AUDIO_DISABLE, chip_audio->read_bitstream_mode, chip);
                            audio_enable->op->set_enable_pcm_encode(audio_enable, AUDIO_DISABLE, chip);
                        }
                    } else {
                        audio_enable->op->get_enable_control_param(audio_enable, chip);
                        audio_enable->op->set_enable_adpcm_encode(audio_enable, AUDIO_DISABLE, chip);
                        audio_control->op->set_enable_adpcm_chan(audio_control, chan_id, AUDIO_DISABLE, chip);
                        if(enable == AUDIO_ENABLE){
                            audio_enable->op->set_enable_pcm_encode(audio_enable, AUDIO_ENABLE, chip);
                            audio_control->op->set_enable_pcm_chan(audio_control, phy_chan_id, AUDIO_ENABLE, chip_audio->read_bitstream_mode, chip);
                        } else if(enable == AUDIO_DISABLE){
                            audio_control->op->set_enable_pcm_chan(audio_control, phy_chan_id, chip_audio->read_bitstream_mode, AUDIO_DISABLE, chip);
                            audio_enable->op->set_enable_pcm_encode(audio_enable, AUDIO_DISABLE, chip);
                        }                    
                    }
                    break;
                case ISIL_AUDIO_ADPCM_32K:
                    if(chip_audio->read_bitstream_mode == READ_BITSTREAM_BY_PCI){
                        audio_control->op->set_enable_pci_sel_pcm_or_adpcm(audio_control, chan_id, ISIL_AUDIO_ADPCM_32K, chip);
                    }
                    if(enable == AUDIO_ENABLE){
                        audio_enable->op->set_enable_pcm_encode(audio_enable, AUDIO_ENABLE, chip);
                        audio_control->op->set_enable_pcm_chan(audio_control, phy_chan_id, AUDIO_ENABLE, chip_audio->read_bitstream_mode, chip);
                        audio_enable->op->set_enable_adpcm_encode(audio_enable, AUDIO_ENABLE, chip);
                        audio_control->op->set_enable_adpcm_chan(audio_control, chan_id, AUDIO_ENABLE, chip);
                    } else if(enable == AUDIO_DISABLE){
                        audio_control->op->set_enable_adpcm_chan(audio_control, chan_id, AUDIO_DISABLE, chip);
                        audio_enable->op->set_enable_adpcm_encode(audio_enable, AUDIO_DISABLE, chip);
                        audio_control->op->set_enable_pcm_chan(audio_control, phy_chan_id, AUDIO_DISABLE, chip_audio->read_bitstream_mode, chip);
                        audio_enable->op->set_enable_pcm_encode(audio_enable, AUDIO_DISABLE, chip);
                    }
                    break;
                default:
                    ISIL_DBG(ISIL_DBG_INFO, ": no surpport %d type\n", running_param->op->get_type(running_param));
                    break;
            }
        }
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_encode_update_enable_audio_encode_chan
 * Description   :   update chip audio encode enable audio encode chan
 * Input         :   param1 is chip audio encode management struct,
 *                   param2 is chan_id,
 *                   param3 is enable,
 *                   param4 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_encode_update_enable_audio_encode_chan(chip_audio_encode_t *audio_encode, int chan_id, int enable)
{
    audio_encode_control_t  *audio_control;
    isil_audio_param_t      *running_param;
    chip_audio_t            *chip_audio;
    isil_audio_driver_t     *audio_driver;
    int     phy_chan_id;

    audio_control = &audio_encode->audio_control;
    chip_audio = to_get_chip_audio_with_audio_encode(audio_encode);
    chip_audio->op->get_audio_chan_driver(chip_audio, chan_id, &audio_driver);
    if(audio_driver != NULL){
        phy_chan_id = audio_driver->audio_phy_chan_id;
    } else {
        phy_chan_id = chan_id;
    }
    running_param = &audio_encode->running_param;
    if(audio_control->op != NULL){
        audio_enable_control_t  *audio_enable = &chip_audio->audio_enable;
        if((audio_enable->op!=NULL) && (running_param->op!=NULL)){
            switch(running_param->op->get_type(running_param)){
                case ISIL_AUDIO_PCM:
                    if(chip_audio->read_bitstream_mode == READ_BITSTREAM_BY_PCI){
                        audio_control->op->pci_sel_pcm_or_adpcm(audio_control, chan_id, ISIL_AUDIO_PCM);
                        if(enable == AUDIO_ENABLE){
                            audio_control->op->enable_pcm_chan(audio_control, phy_chan_id);
                            audio_enable->op->update_enable_pcm_encode(audio_enable, AUDIO_ENABLE);
                            audio_control->op->enable_adpcm_chan(audio_control, chan_id);
                            audio_enable->op->update_enable_adpcm_encode(audio_enable, AUDIO_ENABLE);
                        } else if(enable == AUDIO_DISABLE){
                            audio_control->op->disable_adpcm_chan(audio_control, chan_id);
                            audio_enable->op->update_enable_adpcm_encode(audio_enable, AUDIO_DISABLE);
                            audio_control->op->disable_pcm_chan(audio_control, phy_chan_id);
                            audio_enable->op->update_enable_pcm_encode(audio_enable, AUDIO_DISABLE);
                        }
                    } else {
                        audio_control->op->disable_adpcm_chan(audio_control, phy_chan_id);
                        audio_enable->op->update_enable_adpcm_encode(audio_enable, AUDIO_DISABLE);
                        if(enable == AUDIO_ENABLE){
                            audio_control->op->enable_pcm_chan(audio_control, phy_chan_id);
                            audio_enable->op->update_enable_pcm_encode(audio_enable, AUDIO_ENABLE);
                        } else if(enable == AUDIO_DISABLE){
                            audio_control->op->disable_pcm_chan(audio_control, phy_chan_id);
                            audio_enable->op->update_enable_pcm_encode(audio_enable, AUDIO_DISABLE);
                        }                    
                    }
                    break;
                case ISIL_AUDIO_ADPCM_32K:
                    if(chip_audio->read_bitstream_mode == READ_BITSTREAM_BY_PCI){
                        audio_control->op->pci_sel_pcm_or_adpcm(audio_control, chan_id, ISIL_AUDIO_ADPCM_32K);
                    }
                    if(enable == AUDIO_ENABLE){
                        audio_control->op->enable_pcm_chan(audio_control, phy_chan_id);
                        audio_enable->op->update_enable_pcm_encode(audio_enable, AUDIO_ENABLE);
                        audio_control->op->enable_adpcm_chan(audio_control, chan_id);
                        audio_enable->op->update_enable_adpcm_encode(audio_enable, AUDIO_ENABLE);
                    } else if(enable == AUDIO_DISABLE){
                        audio_control->op->disable_adpcm_chan(audio_control, chan_id);
                        audio_enable->op->update_enable_adpcm_encode(audio_enable, AUDIO_DISABLE);
                        audio_control->op->disable_pcm_chan(audio_control, phy_chan_id);
                        audio_enable->op->update_enable_pcm_encode(audio_enable, AUDIO_DISABLE);
                    }
                    break;
                default:
                    ISIL_DBG(ISIL_DBG_INFO, ": no surpport %d type\n", running_param->op->get_type(running_param));
                    break;
            }
        } 
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_encode_get_audio_encode_chan_enable_state
 * Description   :   get chip audio encode enable audio encode chan
 * Input         :   param1 is chip audio encode management struct,
 *                   param2 is chan_id,
 * Output        :   null
 * Return        :   enable state
 ****************************************************************************************************************/
static int  chip_audio_encode_get_audio_encode_chan_enable_state(chip_audio_encode_t *audio_encode, int chan_id)
{
    audio_encode_control_t  *audio_control;
    isil_audio_param_t      *running_param;
    int	ret = AUDIO_DISABLE;

    audio_control = &audio_encode->audio_control;
    running_param = &audio_encode->running_param;
    if(audio_control->op != NULL){
        chip_audio_t    *chip_audio = to_get_chip_audio_with_audio_encode(audio_encode);
        audio_enable_control_t  *audio_enable = &chip_audio->audio_enable;
        if((audio_enable->op!=NULL) && (running_param->op!=NULL)){
            switch(running_param->op->get_type(running_param)){
                case ISIL_AUDIO_PCM:
                    ret = audio_enable->op->get_pcm_encode_state(audio_enable);
                    break;
                case ISIL_AUDIO_ADPCM_32K:
                    ret = audio_enable->op->get_adpcm_encode_state(audio_enable);
                    break;
                default:
                    ISIL_DBG(ISIL_DBG_INFO, ": no surpport %d type\n", running_param->op->get_type(running_param));
                    break;
            }
        }
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   chip_audio_encode_get_audio_encode_chan_rd_ptr
 * Description   :   get chip audio encode chan's rd ptr
 * Input         :   param1 is chip audio encode management struct,
 *                   param2 is chan_id,
 *                   param3 is ext flag
 * Output        :   null
 * Return        :   encode chan's rd ptr
 ****************************************************************************************************************/
static int  chip_audio_encode_get_audio_encode_chan_rd_ptr(chip_audio_encode_t *audio_encode, int chan_id, int *ext_flag, int *queue_size)
{
    audio_adpcm_encode_ptr_control_t    *adpcm_audio_ptr;
    audio_pcm_encode_ptr_control_t      *pcm_audio_ptr;
    isil_audio_param_t                  *running_param;
    int	ret = INVALID_ISIL_AUDIO_SECTION_ID;

    running_param = &audio_encode->running_param;
    *ext_flag = 0;
    *queue_size = 0;
    if(running_param->op != NULL){
        switch(running_param->op->get_type(running_param)){
            case ISIL_AUDIO_PCM:
                pcm_audio_ptr = &audio_encode->pcm_audio_ptr;
                if(pcm_audio_ptr->op != NULL){
                    ret = pcm_audio_ptr->op->get_chan_rd_ptr(pcm_audio_ptr, chan_id, ext_flag);
                    *queue_size = pcm_audio_ptr->op->get_chan_section_number(pcm_audio_ptr);
                }
                break;
            case ISIL_AUDIO_ADPCM_32K:
                adpcm_audio_ptr = &audio_encode->adpcm_audio_ptr;
                if(adpcm_audio_ptr->op != NULL){
                    ret = adpcm_audio_ptr->op->get_chan_rd_ptr(adpcm_audio_ptr, chan_id, ext_flag);
                    *queue_size = adpcm_audio_ptr->op->get_chan_section_number(adpcm_audio_ptr);
                }
                break;
            default:
                ISIL_DBG(ISIL_DBG_INFO, ": no surpport %d type\n", running_param->op->get_type(running_param));
                break;
        }
    } 
    return ret;
}

/****************************************************************************************************************
 * Function Name :   chip_audio_encode_get_audio_encode_chan_wr_ptr
 * Description   :   get chip audio encode chan's wr ptr
 * Input         :   param1 is chip audio encode management struct,
 *                   param2 is chan_id,
 *                   param3 is ext_flag
 * Output        :   null
 * Return        :   encode chan's rd ptr
 ****************************************************************************************************************/
static int  chip_audio_encode_get_audio_encode_chan_wr_ptr(chip_audio_encode_t *audio_encode, int chan_id, int *ext_flag, int *queue_size)
{
    audio_adpcm_encode_ptr_control_t    *adpcm_audio_ptr;
    audio_pcm_encode_ptr_control_t      *pcm_audio_ptr;
    isil_audio_param_t                  *running_param;
    int	ret = INVALID_ISIL_AUDIO_SECTION_ID;

    running_param = &audio_encode->running_param;
    *ext_flag = 0;
    *queue_size = 0;
    if(running_param->op != NULL){
        switch(running_param->op->get_type(running_param)){
            case ISIL_AUDIO_PCM:
                pcm_audio_ptr = &audio_encode->pcm_audio_ptr;
                if(pcm_audio_ptr->op != NULL){
                    ret = pcm_audio_ptr->op->get_chan_wr_ptr(pcm_audio_ptr, chan_id, ext_flag);
                    *queue_size = pcm_audio_ptr->op->get_chan_section_number(pcm_audio_ptr);
                }
                break;
            case ISIL_AUDIO_ADPCM_32K:
                adpcm_audio_ptr = &audio_encode->adpcm_audio_ptr;
                if(adpcm_audio_ptr->op != NULL){
                    ret = adpcm_audio_ptr->op->get_chan_wr_ptr(adpcm_audio_ptr, chan_id, ext_flag);
                    *queue_size = adpcm_audio_ptr->op->get_chan_section_number(adpcm_audio_ptr);
                }
                break;
            default:
                ISIL_DBG(ISIL_DBG_INFO, ": no surpport %d type\n", running_param->op->get_type(running_param));
                break;
        }
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   chip_audio_encode_get_audio_encode_param
 * Description   :   get chip audio encode hardware param
 * Input         :   param1 is chip audio encode management struct,
 *                   param2 is chip
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_encode_get_audio_encode_param(chip_audio_encode_t *audio_encode, isil_chip_t *chip)
{
    if(chip != NULL){
        audio_encode->audio_control.op->get_audio_param(&audio_encode->audio_control, chip);
        audio_encode->pcm_audio_ptr.op->get_pcm_wr_rd_ptr(&audio_encode->pcm_audio_ptr, chip);
        audio_encode->adpcm_audio_ptr.op->get_adpcm_wr_rd_ptr(&audio_encode->adpcm_audio_ptr, chip);
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_encode_set_audio_encode_param
 * Description   :   set chip audio encode hardware param
 * Input         :   param1 is chip audio encode management struct,
 *                   param2 is chip
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_encode_set_audio_encode_param(chip_audio_encode_t *audio_encode, isil_chip_t *chip)
{
    if(chip != NULL){
        audio_encode->audio_control.op->set_audio_param(&audio_encode->audio_control, chip);
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_encode_set_finish_rd_ptr
 * Description   :   set chip audio encode chan's finish id
 * Input         :   param1 is chip audio encode management struct,
 *                   param2 is finish id
 *                   param3 is chip
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_encode_set_finish_rd_ptr(chip_audio_encode_t *audio_encode, int finish_rd_ptr, isil_chip_t *chip)
{
    audio_adpcm_encode_ptr_control_t    *adpcm_audio_ptr;
    audio_pcm_encode_ptr_control_t      *pcm_audio_ptr;
    isil_audio_param_t                  *running_param;

    running_param = &audio_encode->running_param;
    finish_rd_ptr &= 0x1f;
    if(running_param->op != NULL){
        switch(running_param->op->get_type(running_param)){
            case ISIL_AUDIO_PCM:
                pcm_audio_ptr = &audio_encode->pcm_audio_ptr;
                if(pcm_audio_ptr->op != NULL){
                    pcm_audio_ptr->op->set_rd_finish_number(pcm_audio_ptr, finish_rd_ptr, chip);
                }
                break;
            case ISIL_AUDIO_ADPCM_32K:
                adpcm_audio_ptr = &audio_encode->adpcm_audio_ptr;
                if(adpcm_audio_ptr->op != NULL){
                    adpcm_audio_ptr->op->set_rd_finish_number(adpcm_audio_ptr, finish_rd_ptr, chip);
                }
                break;
            default:
                ISIL_DBG(ISIL_DBG_INFO, ": no surpport %d type\n", running_param->op->get_type(running_param));
                break;
        }
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_encode_hardware_resource_open
 * Description   :   open chip encode resource
 * Input         :   param1 is chip audio encode management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void	chip_audio_encode_hardware_resource_open(chip_audio_encode_t *audio_encode)
{
    if(audio_encode != NULL){
        atomic_inc(&audio_encode->opened_encode_chan_number);
        if(atomic_read(&audio_encode->opened_encode_chan_number) == 1){
			chip_audio_t	*chip_audio;

			chip_audio = to_get_chip_audio_with_audio_encode(audio_encode);
			if(chip_audio->read_bitstream_mode == READ_BITSTREAM_BY_PCI){
#if	defined(PCI_PLATFORM)
				chip_audio_push_isr_interface_t *audio_push_interface;
				audio_push_interface = &chip_audio->audio_push_interface;
				audio_push_interface->op->open_chip_audio(audio_push_interface);
#endif
			}
            if(audio_encode->add_chip_audio_encode_polling_task) {
                audio_encode->add_chip_audio_encode_polling_task(audio_encode);
            }
        }
        ISIL_DBG(ISIL_DBG_INFO, "audio_encode->opened_necode_chan_number = %d\n", atomic_read(&audio_encode->opened_encode_chan_number));
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_encode_hardware_resource_close
 * Description   :   close chip encode resource
 * Input         :   param1 is chip audio encode management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void	chip_audio_encode_hardware_resource_close(chip_audio_encode_t *audio_encode)
{
    if(audio_encode != NULL){
        if(atomic_read(&audio_encode->opened_encode_chan_number) > 0){
            if(atomic_dec_and_test(&audio_encode->opened_encode_chan_number)){
				chip_audio_t	*chip_audio;
				chip_audio = to_get_chip_audio_with_audio_encode(audio_encode);
				if(chip_audio->read_bitstream_mode == READ_BITSTREAM_BY_PCI){
#if	defined(PCI_PLATFORM)
					chip_audio_push_isr_interface_t *audio_push_interface;
                    audio_push_interface = &chip_audio->audio_push_interface;
                    audio_push_interface->op->close_chip_audio(audio_push_interface);
#endif
                }
                if(audio_encode->delete_chip_audio_encode_polling_task) {
                    audio_encode->delete_chip_audio_encode_polling_task(audio_encode);
                }
            }
        }
    }
}

/****************************************************************************************************************
 * Veriable Name :   chip_audio_encode_op
 * Description   :   for helping op audio encode, we define general interface
 ****************************************************************************************************************/
static struct chip_audio_encode_operation   chip_audio_encode_op = {
    .init = chip_audio_encode_init,

    .set_enable_audio_encode_chan = chip_audio_encode_set_enable_audio_encode_chan,
    .update_enable_audio_encode_chan = chip_audio_encode_update_enable_audio_encode_chan,
    .get_audio_encode_chan_enable_state = chip_audio_encode_get_audio_encode_chan_enable_state,

    .get_audio_encode_chan_rd_ptr = chip_audio_encode_get_audio_encode_chan_rd_ptr,
    .get_audio_encode_chan_wr_ptr = chip_audio_encode_get_audio_encode_chan_wr_ptr,

    .get_audio_encode_param = chip_audio_encode_get_audio_encode_param,
    .set_audio_encode_param = chip_audio_encode_set_audio_encode_param,
    .set_finish_rd_ptr = chip_audio_encode_set_finish_rd_ptr,


    .audio_encode_open = chip_audio_encode_hardware_resource_open,
    .audio_encode_close = chip_audio_encode_hardware_resource_close,
};

void	init_chip_audio_encode(chip_audio_encode_t *audio_encode)
{
    audio_encode->op = &chip_audio_encode_op;
    audio_encode->op->init(audio_encode);
}

/****************************************************************************************************************
 * Function Name :   chip_audio_decode_polling_task
 * Description   :   chip audio decode polling task
 * Input         :   param1 is chip audio decode management struct,
 * Output        :   null
 * Return        :   always ISIL_OK
 ****************************************************************************************************************/
static int	chip_audio_decode_polling_task(void *context)
{
    chip_audio_decode_t *audio_decode = (chip_audio_decode_t*)context;
    if(audio_decode != NULL){
        chip_audio_t    *chip_audio;

        chip_audio = to_get_chip_audio_with_audio_decode(audio_decode);
        chip_audio->op->process_audio_decode(chip_audio);
    }
    return ISIL_OK;
}

/*******************************************************************************
 * Function Name  : add_chip_audio_decodepolling_task
 * Description	 : add chip audio decode polling task
 * Input 		 : param1 is chip audio decode management struct
 * Output		 : NULL
 * Return		 : NULL
 *******************************************************************************/
static void add_chip_audio_decode_polling_task(chip_audio_decode_t *audio_decode)
{
    if(audio_decode != NULL){
        chip_audio_t    *chip_audio;
        isil_chip_t		*chip;
        isil_chip_timer_controller_t	*chip_timer_cont;

        chip_audio = to_get_chip_audio_with_audio_decode(audio_decode);
        chip = chip_audio->chip;
        chip_timer_cont = &chip->chip_timer_cont;
        audio_decode->delete_chip_audio_decode_polling_task(audio_decode);
        audio_decode->chip_aduio_decode_polling_timeid = chip_timer_cont->op->AddForFireTimerJob(chip_timer_cont, CHIP_POLLING_TASK_COUNTER, chip_audio_decode_polling_task, audio_decode);
        if(audio_decode->chip_aduio_decode_polling_timeid == ADDJOBERROR){
            ISIL_DBG(ISIL_DBG_INFO, ": add timer hook err\n");
        }
    }
}

/****************************************************************************************************************
 * Function Name :	delete_chip_audio_decode_polling_task
 * Description	:  delete chip audio decode polling task
 * Input 		 : param1 is chip audio decode management struct
 * Output		:	null
 * Return		:	null
 ****************************************************************************************************************/
static void delete_chip_audio_decode_polling_task(chip_audio_decode_t *audio_decode)
{
    if(audio_decode != NULL){
        chip_audio_t    *chip_audio;
        isil_chip_t		*chip;
        isil_chip_timer_controller_t	*chip_timer_cont;

        chip_audio = to_get_chip_audio_with_audio_decode(audio_decode);
        chip = chip_audio->chip;
        chip_timer_cont = &chip->chip_timer_cont;
        chip_timer_cont->op->DeleteForFireTimerJob(chip_timer_cont, audio_decode->chip_aduio_decode_polling_timeid);
        audio_decode->chip_aduio_decode_polling_timeid = INVALIDTIMERID;
    }
}


/****************************************************************************************************************
 * Function Name :   chip_audio_decode_init
 * Description   :   chip audio decode init
 * Input         :   param1 is chip audio decode management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_decode_init(chip_audio_decode_t *audio_decode)
{
    atomic_set(&audio_decode->need_sync_param, 1);
    atomic_set(&audio_decode->opened_decode_chan_number, 0);
    memcpy(&audio_decode->config_param, &default_audio_encode_config_descriptor, sizeof(isil_audio_config_descriptor_t));
    init_audio_decode_property(&audio_decode->running_param, &default_audio_encode_config_descriptor);
    init_audio_decode_control(&audio_decode->audio_control);
    init_audio_decode_ptr_control(&audio_decode->audio_ptr);
    init_audio_decode_data_addr(&audio_decode->audio_data_base);
    audio_decode->chip_aduio_decode_polling_timeid = INVALIDTIMERID;
    audio_decode->add_chip_audio_decode_polling_task = add_chip_audio_decode_polling_task;
    audio_decode->delete_chip_audio_decode_polling_task = delete_chip_audio_decode_polling_task;
}

/****************************************************************************************************************
 * Function Name :   chip_audio_decode_update_enable_audio_decode_chan
 * Description   :   update enable audio decode chan
 * Input         :   param1 is chip audio decode management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_decode_update_enable_audio_decode_chan(chip_audio_decode_t *audio_decode, int chan_id, int enable)
{
    audio_decode_control_t  *audio_control = &audio_decode->audio_control;
    chip_audio_t            *chip_audio = to_get_chip_audio_with_audio_decode(audio_decode);
    audio_enable_control_t  *audio_enable = &chip_audio->audio_enable;
    if(audio_control->op != NULL){
        if(enable == AUDIO_ENABLE){
            audio_enable->op->update_enable_adpcm_decode(audio_enable, AUDIO_ENABLE);
            audio_control->op->update_decode_chan_enable(audio_control, chan_id, AUDIO_ENABLE);
        }else {
            audio_control->op->update_decode_chan_enable(audio_control, chan_id, AUDIO_DISABLE);
            audio_enable->op->update_enable_adpcm_decode(audio_enable, AUDIO_DISABLE);
        }
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_decode_get_audio_decode_chan_state
 * Description   :   get enable audio decode chan's state
 * Input         :   param1 is chip audio decode management struct,
 * Output        :   null
 * Return        :   real state
 ****************************************************************************************************************/
static int  chip_audio_decode_get_audio_decode_chan_state(chip_audio_decode_t *audio_decode, int chan_id)
{
    int	ret = AUDIO_DISABLE;
    audio_decode_control_t  *audio_control = &audio_decode->audio_control;
    if(audio_control->op != NULL){
        ret = audio_control->op->get_decode_chan_enable(audio_control, chan_id);
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   chip_audio_decode_set_enable_audio_decode_chan
 * Description   :   set enable audio decode chan
 * Input         :   param1 is chip audio decode management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_decode_set_enable_audio_decode_chan(chip_audio_decode_t *audio_decode, int chan_id, int enable, isil_chip_t *chip)
{
    audio_decode_control_t  *audio_control = &audio_decode->audio_control;
    chip_audio_t            *chip_audio = to_get_chip_audio_with_audio_decode(audio_decode);
    audio_enable_control_t  *audio_enable = &chip_audio->audio_enable;

    if(audio_control->op != NULL){
        if(enable == AUDIO_ENABLE){
            audio_enable->op->set_enable_adpcm_decode(audio_enable, AUDIO_ENABLE, chip);
            audio_control->op->set_decode_chan_enable(audio_control, chan_id, AUDIO_ENABLE, chip);
        }else if(enable == AUDIO_DISABLE){
            audio_control->op->set_decode_chan_enable(audio_control, chan_id, AUDIO_DISABLE, chip);
            audio_enable->op->set_enable_adpcm_decode(audio_enable, AUDIO_DISABLE, chip);
        }
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_decode_update_mute_audio_decode_chan
 * Description   :   update enable audio decode chan's mute
 * Input         :   param1 is chip audio decode management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_decode_update_mute_audio_decode_chan(chip_audio_decode_t *audio_decode, int chan_id, int mute)
{
    audio_decode_control_t  *audio_control = &audio_decode->audio_control;
    if(audio_control->op != NULL){
        if(mute == AUDIO_MUTE_ENABLE){
            audio_control->op->update_decode_chan_mute(audio_control, chan_id, AUDIO_MUTE_ENABLE);
        } else {
            audio_control->op->update_decode_chan_mute(audio_control, chan_id, AUDIO_MUTE_DISABLE);
        }
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_decode_get_audio_decode_mute_state
 * Description   :   get enable audio decode chan's mute
 * Input         :   param1 is chip audio decode management struct,
 * Output        :   null
 * Return        :   real state
 ****************************************************************************************************************/
static int  chip_audio_decode_get_audio_decode_mute_state(chip_audio_decode_t *audio_decode, int chan_id)
{
    int	ret = AUDIO_MUTE_ENABLE;
    audio_decode_control_t  *audio_control = &audio_decode->audio_control;
    if(audio_control->op != NULL){
        ret = audio_control->op->get_decode_chan_mute(audio_control, chan_id);
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   chip_audio_decode_update_set_mute_audio_decode_chan
 * Description   :   set enable audio decode chan mute
 * Input         :   param1 is chip audio decode management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_decode_update_set_mute_audio_decode_chan(chip_audio_decode_t *audio_decode, int chan_id, int mute, isil_chip_t *chip)
{
    audio_decode_control_t  *audio_control = &audio_decode->audio_control;
    if(audio_control->op != NULL){
        if(mute == AUDIO_MUTE_ENABLE){
            audio_control->op->set_decode_chan_mute(audio_control, chan_id, AUDIO_MUTE_ENABLE, chip);
        }else if(mute == AUDIO_MUTE_DISABLE){
            audio_control->op->set_decode_chan_mute(audio_control, chan_id, AUDIO_MUTE_DISABLE, chip);
        }
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_decode_get_audio_decode_chan_rd_ptr
 * Description   :   get audio decode chan's rd ptr
 * Input         :   param1 is chip audio decode management struct,
 *                   param2 is chan id
 * Output        :   ext_flag and queue_size
 * Return        :   rd ptr
 ****************************************************************************************************************/
static int  chip_audio_decode_get_audio_decode_chan_rd_ptr(chip_audio_decode_t *audio_decode, int chan_id, int *ext_flag, int *queue_size)
{
    int	ret = INVALID_ISIL_AUDIO_SECTION_ID;
    isil_audio_param_t          *running_param;
    audio_decode_ptr_control_t  *audio_ptr;

    running_param = &audio_decode->running_param;
    *ext_flag = 0;
    *queue_size = 0;
    if(running_param->op != NULL){
        switch(running_param->op->get_type(running_param)){
            case ISIL_AUDIO_PCM:
            case ISIL_AUDIO_ADPCM_32K:
                audio_ptr = &audio_decode->audio_ptr;
                if(audio_ptr->op != NULL){
                    ret = audio_ptr->op->get_chan_rd_ptr(audio_ptr, chan_id, ext_flag);
                    *queue_size = audio_ptr->op->get_chan_section_number(audio_ptr);
                }
                break;
            default:
                ISIL_DBG(ISIL_DBG_INFO, ": no surpport %d type\n", running_param->op->get_type(running_param));
                break;
        }
    } 
    return ret;
}

/****************************************************************************************************************
 * Function Name :   chip_audio_decode_get_audio_decode_chan_wr_ptr
 * Description   :   get audio decode chan's wr ptr
 * Input         :   param1 is chip audio decode management struct,
 *                   param2 is chan id
 * Output        :   ext_flag and queue_size
 * Return        :   wr ptr
 ****************************************************************************************************************/
static int  chip_audio_decode_get_audio_decode_chan_wr_ptr(chip_audio_decode_t *audio_decode, int chan_id, int *ext_flag, int *queue_size)
{
    int	ret = INVALID_ISIL_AUDIO_SECTION_ID;
    isil_audio_param_t          *running_param;
    audio_decode_ptr_control_t  *audio_ptr;

    running_param = &audio_decode->running_param;
    *ext_flag = 0;
    *queue_size = 0;
    if(running_param->op != NULL){
        switch(running_param->op->get_type(running_param)){
            case ISIL_AUDIO_PCM:
            case ISIL_AUDIO_ADPCM_32K:
                audio_ptr = &audio_decode->audio_ptr;
                if(audio_ptr->op != NULL){
                    ret = audio_ptr->op->get_chan_wr_ptr(audio_ptr, chan_id, ext_flag);
                    *queue_size = audio_ptr->op->get_chan_section_number(audio_ptr);
                }
                break;
            default:
                ISIL_DBG(ISIL_DBG_INFO, ": no surpport %d type\n", running_param->op->get_type(running_param));
                break;
        }
    } 
    return ret;
}

/****************************************************************************************************************
 * Function Name :   chip_audio_decode_get_audio_decode_param
 * Description   :   get audio decode param
 * Input         :   param1 is chip audio decode management struct,
 *                   param2 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_decode_get_audio_decode_param(chip_audio_decode_t *audio_decode, isil_chip_t *chip)
{
    if(chip != NULL){
        audio_decode->audio_control.op->get_decode_param(&audio_decode->audio_control, chip);
        audio_decode->audio_ptr.op->get_adpcm_wr_rd_ptr(&audio_decode->audio_ptr, chip);
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_decode_set_audio_decode_param
 * Description   :   set audio decode param
 * Input         :   param1 is chip audio decode management struct,
 *                   param2 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_decode_set_audio_decode_param(chip_audio_decode_t *audio_decode, isil_chip_t *chip)
{
    if(chip != NULL){
        audio_decode_control_t	*audio_control = &audio_decode->audio_control;;
        audio_control->op->set_decode_param(audio_control, chip);
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_decode_hardware_resource_open
 * Description   :   open chip decode resource
 * Input         :   param1 is chip audio decode management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void	chip_audio_decode_hardware_resource_open(chip_audio_decode_t *audio_decode)
{
    if(audio_decode != NULL){
        atomic_inc(&audio_decode->opened_decode_chan_number);
        if(atomic_read(&audio_decode->opened_decode_chan_number) == 1){
            if(audio_decode->add_chip_audio_decode_polling_task) {
                audio_decode->add_chip_audio_decode_polling_task(audio_decode);
            }
        }
        ISIL_DBG(ISIL_DBG_INFO, "audio_decode->opened_decode_chan_number = %d\n", atomic_read(&audio_decode->opened_decode_chan_number));
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_decode_hardware_resource_close
 * Description   :   close chip decode resource
 * Input         :   param1 is chip audio decode management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void	chip_audio_decode_hardware_resource_close(chip_audio_decode_t *audio_decode)
{
    if(audio_decode != NULL){
        if(atomic_read(&audio_decode->opened_decode_chan_number) > 0){
            if(atomic_dec_and_test(&audio_decode->opened_decode_chan_number)){
                if(audio_decode->delete_chip_audio_decode_polling_task) {
                    audio_decode->delete_chip_audio_decode_polling_task(audio_decode);
                }
            }
        }
    }
}

/****************************************************************************************************************
 * Veriable Name :   chip_audio_decode_op
 * Description   :   for helping op audio decode, we define general interface
 ****************************************************************************************************************/
static struct chip_audio_decode_operation   chip_audio_decode_op = {
    .init = chip_audio_decode_init,

    .update_enable_audio_decode_chan = chip_audio_decode_update_enable_audio_decode_chan,
    .get_audio_decode_chan_enable_state = chip_audio_decode_get_audio_decode_chan_state,
    .set_enable_audio_decode_chan = chip_audio_decode_set_enable_audio_decode_chan,
    .update_mute_audio_decode_chan = chip_audio_decode_update_mute_audio_decode_chan,
    .get_audio_decode_mute_state = chip_audio_decode_get_audio_decode_mute_state,
    .set_mute_audio_decode_chan = chip_audio_decode_update_set_mute_audio_decode_chan,
    .get_audio_decode_chan_rd_ptr = chip_audio_decode_get_audio_decode_chan_rd_ptr,
    .get_audio_decode_chan_wr_ptr = chip_audio_decode_get_audio_decode_chan_wr_ptr,
    .get_audio_decode_param = chip_audio_decode_get_audio_decode_param,
    .set_audio_decode_param = chip_audio_decode_set_audio_decode_param,

    .audio_decode_open = chip_audio_decode_hardware_resource_open,
    .audio_decode_close = chip_audio_decode_hardware_resource_close,
};

/****************************************************************************************************************
 * Function Name :   init_chip_audio_decode
 * Description   :   init chip audio decode
 * Input         :   param1 is chip audio decode management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void	init_chip_audio_decode(chip_audio_decode_t *audio_decode)
{
    audio_decode->op = &chip_audio_decode_op;
    audio_decode->op->init(audio_decode);
}

/****************************************************************************************************************
 * Function Name :   chip_audio_init
 * Description   :   chip audio init
 * Input         :   param1 is chip audio management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static int chip_audio_init(chip_audio_t *chip_audio)
{
    init_audio_enable_control(&chip_audio->audio_enable);
    init_chip_audio_encode(&chip_audio->audio_encode);
    init_chip_audio_decode(&chip_audio->audio_decode);

    init_register_table_node(&chip_audio->logic_chan_table);
    init_register_table_node(&chip_audio->opened_logic_chan_table);
    return ISIL_OK;
}


/****************************************************************************************************************
 * Function Name :   chip_audio_reset
 * Description   :   chip audio reset
 * Input         :   param1 is chip audio management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static int chip_audio_reset(chip_audio_t *chip_audio)
{
    return ISIL_OK;
}


/****************************************************************************************************************
 * Function Name :   chip_audio_release
 * Description   :   chip audio release
 * Input         :   param1 is chip audio management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static int chip_audio_release(chip_audio_t *chip_audio)
{
    return ISIL_OK;
}

/****************************************************************************************************************
 * Function Name :   chip_audio_set_encode_audio_chan_type
 * Description   :   chip audio set encode audio chan type
 * Input         :   param1 is chip audio management struct,
 *                   param2 is type
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_set_encode_audio_chan_type(chip_audio_t *chip_audio, int type)
{
    chip_audio_encode_t *audio_encode = &chip_audio->audio_encode;
    isil_audio_param_t	*running_param = &audio_encode->running_param;
    if((type!=ISIL_AUDIO_ADPCM_32K) && (type!=ISIL_AUDIO_PCM)){
        ISIL_DBG(ISIL_DBG_INFO, ": type=%d, we only surpport adpcm and pcm mode\n", type);
        type = ISIL_AUDIO_ADPCM_32K;
    }
    if(running_param->op != NULL){
        running_param->op->update_type(running_param, type);
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_update_encode_audio_chan_type
 * Description   :   chip audio update encode audio chan type
 * Input         :   param1 is chip audio management struct,
 *                   param2 is type
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_update_encode_audio_chan_type(chip_audio_t *chip_audio, int type)
{
    chip_audio_encode_t *audio_encode = &chip_audio->audio_encode;
    isil_audio_param_t	*running_param = &audio_encode->running_param;
    if((type!=ISIL_AUDIO_ADPCM_32K) && (type!=ISIL_AUDIO_PCM)){
        ISIL_DBG(ISIL_DBG_ERR, "type=%d, we only surpport adpcm and pcm mode\n", type);
        type = ISIL_AUDIO_ADPCM_32K;
    }
    if(running_param->op != NULL){
        running_param->op->update_type(running_param, type);
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_get_encode_audio_chan_type
 * Description   :   chip audio get encode audio chan type
 * Input         :   param1 is chip audio management struct,
 * Output        :   null
 * Return        :   type
 ****************************************************************************************************************/
static int  chip_audio_get_encode_audio_chan_type(chip_audio_t *chip_audio)
{
    int	ret = ISIL_AUDIO_ADPCM_32K;
    chip_audio_encode_t *audio_encode = &chip_audio->audio_encode;
    isil_audio_param_t	*running_param = &audio_encode->running_param;
    if(running_param->op != NULL) {
        ret = running_param->op->get_type(running_param);
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   chip_audio_set_decode_audio_chan_type
 * Description   :   chip audio set encode audio chan type
 * Input         :   param1 is chip audio management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_set_decode_audio_chan_type(chip_audio_t *chip_audio, int type)
{
    chip_audio_decode_t *audio_decode = &chip_audio->audio_decode;
    isil_audio_param_t	*running_param = &audio_decode->running_param;

    if((type!=ISIL_AUDIO_ADPCM_32K) && (type!=ISIL_AUDIO_PCM)){
        ISIL_DBG(ISIL_DBG_ERR, "type=%d, we only surpport adpcm and pcm mode\n", type);
        type = ISIL_AUDIO_ADPCM_32K;
    }
    if(running_param->op != NULL){
        running_param->op->update_type(running_param, type);
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_update_decode_audio_chan_type
 * Description   :   chip audio update decode audio chan type
 * Input         :   param1 is chip audio management struct,
 *                   param2 is type
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_update_decode_audio_chan_type(chip_audio_t *chip_audio, int type)
{
    chip_audio_decode_t *audio_decode = &chip_audio->audio_decode;
    isil_audio_param_t	*running_param = &audio_decode->running_param;

    if((type!=ISIL_AUDIO_ADPCM_32K) && (type!=ISIL_AUDIO_PCM)){
        ISIL_DBG(ISIL_DBG_ERR, "type=%d, we only surpport adpcm and pcm mode\n", type);
        type = ISIL_AUDIO_ADPCM_32K;
    }
    if(running_param->op != NULL){
        running_param->op->update_type(running_param, type);
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_get_decode_audio_chan_type
 * Description   :   chip audio get decode audio chan type
 * Input         :   param1 is chip audio management struct,
 * Output        :   null
 * Return        :   type
 ****************************************************************************************************************/
static int  chip_audio_get_decode_audio_chan_type(chip_audio_t *chip_audio)
{
    int	ret = ISIL_AUDIO_ADPCM_32K;
    chip_audio_decode_t *audio_decode = &chip_audio->audio_decode;
    isil_audio_param_t  *running_param = &audio_decode->running_param;

    if(running_param->op != NULL){
        ret = running_param->op->get_type(running_param);
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   chip_audio_set_enable_audio_chan
 * Description   :   chip audio set enable audio chan
 * Input         :   param1 is chip audio management struct,
 * Output        :   null
 * Return        :   type
 ****************************************************************************************************************/
static void chip_audio_set_enable_audio_chan(chip_audio_t *chip_audio, int chan_id, int enable)
{
    if(chan_id < ISIL_AUDIO_IN_CHAN_NUMBER){
        chip_audio_encode_t *audio_encode = &chip_audio->audio_encode;
        if(audio_encode->op != NULL){
            audio_encode->op->set_enable_audio_encode_chan(audio_encode, chan_id, enable, chip_audio->chip);
        }
    } else if(chan_id < ISIL_AUDIO_CHAN_NUMBER){
        chip_audio_decode_t *audio_decode = &chip_audio->audio_decode;
        if(audio_decode->op != NULL){
            audio_decode->op->set_enable_audio_decode_chan(audio_decode, chan_id, enable, chip_audio->chip);
        }
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_update_enable_audio_chan
 * Description   :   chip audio update enable audio chan
 * Input         :   param1 is chip audio management struct,
 *                   param2 is chan id,
 *                   param3 is enable
 * Output        :   null
 * Return        :   type
 ****************************************************************************************************************/
static void chip_audio_update_enable_audio_chan(chip_audio_t *chip_audio, int chan_id, int enable)
{
    if(chan_id < ISIL_AUDIO_IN_CHAN_NUMBER){
        chip_audio_encode_t *audio_encode = &chip_audio->audio_encode;
        if(audio_encode->op != NULL){
            audio_encode->op->update_enable_audio_encode_chan(audio_encode, chan_id, enable);
        }
    } else if(chan_id < ISIL_AUDIO_CHAN_NUMBER){
        chip_audio_decode_t *audio_decode = &chip_audio->audio_decode;
        if(audio_decode->op != NULL){
            audio_decode->op->update_enable_audio_decode_chan(audio_decode, chan_id, enable);
        }
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_get_audio_chan_enable_state
 * Description   :   chip audio get audio chan enable state
 * Input         :   param1 is chip audio management struct,
 *                   param2 is chan id,
 *                   param3 is enable
 * Output        :   null
 * Return        :   type
 ****************************************************************************************************************/
static int  chip_audio_get_audio_chan_enable_state(chip_audio_t *chip_audio, int chan_id)
{
    int	ret = AUDIO_DISABLE;
    if(chip_audio != NULL){
        if(chan_id < ISIL_AUDIO_IN_CHAN_NUMBER){
            chip_audio_encode_t *audio_encode = &chip_audio->audio_encode;
            if(audio_encode->op != NULL){
                ret = audio_encode->op->get_audio_encode_chan_enable_state(audio_encode, chan_id);
            }
        } else {
            chip_audio_decode_t *audio_decode = &chip_audio->audio_decode;
            if(audio_decode->op != NULL){
                ret = audio_decode->op->get_audio_decode_chan_enable_state(audio_decode, chan_id);
            }
        }
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   chip_audio_set_mute_audio_chan
 * Description   :   chip audio get audio chan enable state
 * Input         :   param1 is chip audio management struct,
 *                   param2 is chan id,
 *                   param3 is enable
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_set_mute_audio_chan(chip_audio_t *chip_audio, int chan_id, int mute)
{
    if(chan_id < ISIL_AUDIO_IN_CHAN_NUMBER){

    } else if(chan_id < ISIL_AUDIO_CHAN_NUMBER){
        chip_audio_decode_t *audio_decode = &chip_audio->audio_decode;
        if(audio_decode->op != NULL){
            audio_decode->op->set_mute_audio_decode_chan(audio_decode, chan_id, mute, chip_audio->chip);
        }
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_update_mute_audio_chan
 * Description   :   update chip audio decode chan mute flag
 * Input         :   param1 is chip audio management struct,
 *                   param2 is chan id,
 *                   param3 is mute
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_update_mute_audio_chan(chip_audio_t *chip_audio, int chan_id, int mute)
{
    if(chan_id < ISIL_AUDIO_IN_CHAN_NUMBER){

    } else if(chan_id < ISIL_AUDIO_CHAN_NUMBER){
        chip_audio_decode_t *audio_decode = &chip_audio->audio_decode;
        if(audio_decode->op != NULL){
            audio_decode->op->update_mute_audio_decode_chan(audio_decode, chan_id, mute);
        }
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_get_audio_chan_mute_state
 * Description   :   get chip audio decode chan mute state
 * Input         :   param1 is chip audio management struct,
 *                   param2 is chan id,
 * Output        :   null
 * Return        :   mute state
 ****************************************************************************************************************/
static int  chip_audio_get_audio_chan_mute_state(chip_audio_t *chip_audio, int chan_id)
{
    int	ret = AUDIO_MUTE_ENABLE;
    if(chan_id < ISIL_AUDIO_IN_CHAN_NUMBER){

    } else if(chan_id < ISIL_AUDIO_CHAN_NUMBER){
        chip_audio_decode_t *audio_decode = &chip_audio->audio_decode;
        if(audio_decode->op != NULL){
            ret = audio_decode->op->get_audio_decode_mute_state(&chip_audio->audio_decode, chan_id);
        }
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   chip_audio_get_chip_audio_param
 * Description   :   get chip audio all param
 * Input         :   param1 is chip audio management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_get_chip_audio_param(chip_audio_t *chip_audio)
{
    isil_chip_t *chip = chip_audio->chip;
    chip_audio->audio_enable.op->get_enable_control_param(&chip_audio->audio_enable, chip);
    chip_audio->audio_encode.op->get_audio_encode_param(&chip_audio->audio_encode, chip);
    chip_audio->audio_decode.op->get_audio_decode_param(&chip_audio->audio_decode, chip);
}

/****************************************************************************************************************
 * Function Name :   chip_audio_set_chip_audio_param
 * Description   :   set chip audio all param
 * Input         :   param1 is chip audio management struct,
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_set_chip_audio_param(chip_audio_t *chip_audio)
{
    isil_chip_t *chip = chip_audio->chip;
    chip_audio->audio_enable.op->set_enable_control_param(&chip_audio->audio_enable, chip);
    chip_audio->audio_encode.op->set_audio_encode_param(&chip_audio->audio_encode, chip);
    chip_audio->audio_decode.op->set_audio_decode_param(&chip_audio->audio_decode, chip);
}

/****************************************************************************************************************
 * Function Name :   audio_encode_chan_response_ddr_burst_done_isr
 * Description   :   response audio encode chan burst read ddr done isr
 * Input         :   param1 is irq number,
 *                   param2 is context
 * Output        :   null
 * Return        :   always 1
 ****************************************************************************************************************/
static int  audio_encode_chan_response_ddr_burst_done_isr(int irq, void *context)
{
    isil_audio_driver_t *audio_chan_driver = (isil_audio_driver_t*)context;
    int ret = 1;
    if(audio_chan_driver != NULL){
        isil_audio_packet_queue_t   *audio_packet_queue = &audio_chan_driver->audio_packet_queue;
        isil_chip_t                 *chip = audio_chan_driver->chip_audio->chip;
        chip_audio_encode_t         *audio_encode = &audio_chan_driver->chip_audio->audio_encode;
        audio_section_ptr_info_t    *audio_section_param = &audio_chan_driver->audio_section_param;
        chip_ddr_burst_engine_t     *chip_ddr_burst_interface;
        ddr_burst_interface_t       *burst_interface;
        isil_timestamp_t            *timestamp = &audio_chan_driver->timestamp;
        dpram_page_node_t           *dpram_page;
        isil_audio_packet_section_t *audio_section;
        isil_audio_descriptor_t 	*audio_descriptor;
        audio_section_descriptor_t  descriptor;
        u32 dpram_base;
        int finish_id;
        unsigned long   flags;

        spin_lock_irqsave(&audio_chan_driver->fsm_matrix_call_lock, flags);
        chip_ddr_burst_interface = &chip->chip_ddr_burst_interface;
        chip_ddr_burst_interface->op->get_ddr_burst_interface(chip_ddr_burst_interface, &burst_interface);
        burst_interface->op->clear_burst_done(burst_interface);

        audio_packet_queue->op->try_get_curr_producer_from_pool(audio_packet_queue, &audio_chan_driver->audio_buf_pool);
        flush_audio_encode_curr_producer_frame(audio_chan_driver);
        audio_section = audio_packet_queue->curr_producer;
        if(audio_section != NULL){
            dpram_page = audio_chan_driver->dpram_page;
            dpram_base = dpram_page->op->get_page_base(dpram_page);
            dpram_base += AUDIO_SECTION_SIZE;
            dpram_base -= sizeof(audio_section_descriptor_t);
            *((u32*)&descriptor) = chip->io_op->chip_read32(chip, dpram_base);
            timestamp->op->set_timestamp(timestamp, descriptor.timestamp, __FILE__);

            audio_descriptor = &audio_section->descriptor;
            *((u32*)audio_descriptor->descriptor) = *((u32*)&descriptor);
            audio_section->payload_len  = audio_descriptor->op->get_audio_descriptor_valid_len(audio_descriptor);
            audio_section->sample_rate = audio_descriptor->op->get_audio_descriptor_sample_rate(audio_descriptor);
            audio_section->type = audio_descriptor->op->get_audio_descriptor_type(audio_descriptor);
            audio_section->bit_wide = audio_descriptor->op->get_audio_descriptor_bit_wide(audio_descriptor);
            audio_section->timestamp = timestamp->op->get_timestamp(timestamp);
            audio_section->frameSerial = audio_chan_driver->i_frame_serial++;

            audio_section->payload_len = H264_MIN((audio_section->section_size-sizeof(ext_h264_nal_bitstream_t)), audio_section->payload_len);
            burst_interface->op->pio_host_to_sram_read(burst_interface, audio_chan_driver->dpram_page, (u32 *)(audio_section->data + sizeof(ext_h264_nal_bitstream_t)), audio_section->payload_len, 0);
            audio_section->payload_len = gen_ext_nal_audio_header(audio_section->data, audio_section->type, audio_section->payload_len);
            audio_packet_queue->op->put_curr_producer_into_queue(audio_packet_queue);
        } else {
            audio_chan_driver->discard_number++;
        }

        finish_id = audio_section_param->op->get_update_finish_id(audio_section_param);
        audio_encode->op->set_finish_rd_ptr(audio_encode, finish_id, chip);

        //notify data move req queue to update
        chip_free_irq(chip, IRQ_BURST_TYPE_INTR, audio_chan_driver);
        chip_ddr_burst_interface->op->ack_read_audio_req(chip_ddr_burst_interface, audio_chan_driver);
        driver_gen_deliver_event(&audio_chan_driver->opened_logic_chan_ed, 1);	//from transfering to done and continue to idle or other(suspend or close)
        spin_unlock_irqrestore(&audio_chan_driver->fsm_matrix_call_lock, flags);
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   audio_encode_chan_start_read_bitstream
 * Description   :   audio encode chan start read bitstream from ddr
 * Input         :   param1 is dpram request management struct
 *                   param2 is context
 * Output        :   null
 * Return        :   always 1
 ****************************************************************************************************************/
static int audio_encode_chan_start_read_bitstream(dpram_request_t *dpram_req, void *context)
{
    int ret=1;
    if((dpram_req!=NULL) && (context!=NULL)){
        isil_audio_driver_t         *audio_chan_driver = (isil_audio_driver_t*)context;
        isil_chip_t                 *chip = audio_chan_driver->chip_audio->chip;
        audio_section_ptr_info_t    *audio_section_param = &audio_chan_driver->audio_section_param;
        chip_ddr_burst_engine_t     *chip_ddr_burst_interface;
        ddr_burst_interface_t       *burst_interface;
        ed_tcb_t	*opened_logic_chan_ed;
        isil_ed_fsm_t   *ed_fsm;

        chip_ddr_burst_interface = &chip->chip_ddr_burst_interface;
        chip_ddr_burst_interface->op->get_ddr_burst_interface(chip_ddr_burst_interface, &burst_interface);
        opened_logic_chan_ed = &audio_chan_driver->opened_logic_chan_ed;
        driver_gen_deliver_event(opened_logic_chan_ed, 1);	//starndby change to running
        ed_fsm = &opened_logic_chan_ed->ed_fsm;
        if(chip_ddr_burst_interface->op->is_can_submit_move_data_from_ddr_to_dpram_service_req(chip_ddr_burst_interface, &audio_chan_driver->dpram_page)){
            chip_request_irq(chip, IRQ_BURST_TYPE_INTR, audio_encode_chan_response_ddr_burst_done_isr, "BURST", (void*)audio_chan_driver);
            burst_interface->op->start_nonblock_transfer_from_ddr_to_sram(burst_interface, audio_chan_driver->dpram_page, audio_section_param->ddr_end_addr, audio_section_param->page_id, AUDIO_SECTION_SIZE, audio_section_param->chip_a_or_b);
        } else {
            chip_ddr_burst_interface->op->ack_read_audio_req(chip_ddr_burst_interface, audio_chan_driver);
            ed_fsm->op->change_state(ed_fsm, ISIL_ED_TRANSFERING);
        }
        driver_gen_deliver_event(opened_logic_chan_ed, 1);	//case 1 from running to transfering, case 2 from transfering to done and continue to idle or other(suspend or close)
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   audio_encode_chan_delete_start_read_bitstream
 * Description   :   audio encode chan delete read request
 * Input         :   param1 is dpram request management struct
 *                   param2 is context
 * Output        :   null
 * Return        :   always 1
 ****************************************************************************************************************/
static int  audio_encode_chan_delete_start_read_bitstream(dpram_request_t *dpram_req, void *context)
{
    int ret=1;
    if((dpram_req!=NULL) && (context!=NULL)){

    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   init_submit_recv_audio_bitstream_service
 * Description   :   init audio encode chan read request
 * Input         :   param1 is dpram request management struct
 *                   param2 is context
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void init_submit_recv_audio_bitstream_service(dpram_request_t *dpram_req, isil_audio_driver_t *audio_chan_driver)
{
    if((dpram_req!=NULL) || (audio_chan_driver!=NULL)){
        dpram_req->chip = audio_chan_driver->chip_audio->chip;
        dpram_req->type = DPRAM_NONBLOCK_TRANSFER_REQUEST;
        dpram_req->context = (void*)audio_chan_driver;
        dpram_req->req_callback = audio_encode_chan_start_read_bitstream;
        dpram_req->delete_req_notify = audio_encode_chan_delete_start_read_bitstream;
    }
}

/****************************************************************************************************************
 * Function Name :   audio_decode_chan_response_ddr_burst_done_isr
 * Description   :   response audio decode chan burst write ddr done isr
 * Input         :   param1 is irq number,
 *                   param2 is context
 * Output        :   null
 * Return        :   always 1
 ****************************************************************************************************************/
static int  audio_decode_chan_response_ddr_burst_done_isr(int irq, void *context)
{
    isil_audio_driver_t   *audio_chan_driver = (isil_audio_driver_t*)context;
    int ret = 0;
    if(audio_chan_driver != NULL){
        chip_audio_decode_t         *audio_decode = &audio_chan_driver->chip_audio->audio_decode;
        audio_decode_ptr_control_t  *audio_decode_ptr = &audio_decode->audio_ptr;
        audio_section_ptr_info_t    *audio_section_param = &audio_chan_driver->audio_section_param;
        isil_chip_t                  *chip = audio_chan_driver->chip_audio->chip;
        chip_ddr_burst_engine_t     *chip_ddr_burst_interface;
        ddr_burst_interface_t       *burst_interface;
        int finish_id;
        unsigned long   flags;

        spin_lock_irqsave(&audio_chan_driver->fsm_matrix_call_lock, flags);
        chip_ddr_burst_interface = &chip->chip_ddr_burst_interface;
        chip_ddr_burst_interface->op->get_ddr_burst_interface(chip_ddr_burst_interface, &burst_interface);
        burst_interface->op->clear_burst_done(burst_interface);
        finish_id = audio_section_param->op->get_update_finish_id(audio_section_param);
        audio_decode_ptr->op->set_finish_wr_ptr(audio_decode_ptr, finish_id, chip, audio_chan_driver->audio_logic_chan_id);
        audio_chan_driver->i_frame_serial++;

        //notify data move req queue to update
        chip_free_irq(chip, IRQ_BURST_TYPE_INTR, audio_chan_driver);
        chip_ddr_burst_interface->op->ack_write_audio_req(chip_ddr_burst_interface, audio_chan_driver);
        driver_gen_deliver_event(&audio_chan_driver->opened_logic_chan_ed, 1);	//from transfering to done and continue to idle or other(close or suspend)
        spin_unlock_irqrestore(&audio_chan_driver->fsm_matrix_call_lock, flags);
        ret = 1;
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_chan_start_send_bitstream
 * Description   :   audio decode chan start write bitstream to ddr
 * Input         :   param1 is dpram request management struct
 *                   param2 is context
 * Output        :   null
 * Return        :   always 1
 ****************************************************************************************************************/
static int audio_decode_chan_start_send_bitstream(dpram_request_t *dpram_req, void *context)
{
    int ret=1;
    if((dpram_req!=NULL) && (context!=NULL)){
        isil_audio_driver_t         *audio_chan_driver = (isil_audio_driver_t*)context;
        isil_chip_t                 *chip = audio_chan_driver->chip_audio->chip;
        audio_section_ptr_info_t    *audio_section_param = &audio_chan_driver->audio_section_param;
        isil_audio_packet_queue_t   *audio_packet_queue = &audio_chan_driver->audio_packet_queue;
        isil_audio_chan_buf_pool_t  *audio_buf_pool = &audio_chan_driver->audio_buf_pool;
        chip_ddr_burst_engine_t     *chip_ddr_burst_interface;
        ddr_burst_interface_t       *burst_interface;
        ed_tcb_t                    *ed_tcb = &audio_chan_driver->opened_logic_chan_ed;

        chip_ddr_burst_interface = &chip->chip_ddr_burst_interface;
        chip_ddr_burst_interface->op->get_ddr_burst_interface(chip_ddr_burst_interface, &burst_interface);
        driver_gen_deliver_event(ed_tcb, 1);	//from standby to running
        flush_audio_decode_curr_consumer_frame(audio_chan_driver);
        audio_packet_queue->op->try_get_curr_consumer_from_queue(audio_packet_queue);
        if(audio_packet_queue->curr_consumer != NULL){
            isil_audio_packet_section_t   *audio_section = audio_packet_queue->curr_consumer;
            isil_audio_descriptor_t       *audio_decode_descriptor = &audio_section->descriptor;

            if(chip_ddr_burst_interface->op->is_can_submit_move_data_from_host_to_dpram_service_req(chip_ddr_burst_interface, &audio_chan_driver->dpram_page)){
                dpram_page_node_t   *dpram_page;
                u32 dpram_base;

                burst_interface->op->pio_host_to_sram_write(burst_interface, audio_chan_driver->dpram_page, (u32 *)(audio_section->data/* + sizeof(ext_h264_nal_bitstream_t)*/), audio_section->payload_len, 0);

                dpram_page = audio_chan_driver->dpram_page;
                dpram_base = dpram_page->op->get_page_base(dpram_page);
                dpram_base += audio_section->section_size;
                dpram_base -= sizeof(audio_section_descriptor_t);
                audio_decode_descriptor->op->set_audio_descriptor_valid_len(audio_decode_descriptor, audio_section->payload_len);
                audio_decode_descriptor->op->set_audio_descriptor_sample_rate(audio_decode_descriptor, audio_section->sample_rate);
                audio_decode_descriptor->op->set_audio_descriptor_type(audio_decode_descriptor, audio_section->type);
                audio_decode_descriptor->op->set_audio_descriptor_bit_wide(audio_decode_descriptor, audio_section->bit_wide);
                audio_decode_descriptor->op->set_audio_descriptor_timestamp(audio_decode_descriptor, audio_section->timestamp&0xffff);
                chip->io_op->chip_write32(chip, dpram_base, (*((u32*)audio_decode_descriptor->descriptor)));
                chip_request_irq(chip, IRQ_BURST_TYPE_INTR, audio_decode_chan_response_ddr_burst_done_isr, "BURST", (void*)audio_chan_driver);
                burst_interface->op->start_nonblock_transfer_from_sram_to_ddr(burst_interface, audio_chan_driver->dpram_page, audio_section_param->ddr_end_addr, audio_section_param->page_id, AUDIO_SECTION_SIZE, audio_section_param->chip_a_or_b);
                audio_packet_queue->op->release_curr_consumer(audio_packet_queue, audio_buf_pool);
            } else {
                chip_ddr_burst_interface->op->ack_write_audio_req(chip_ddr_burst_interface, audio_chan_driver);
                ed_tcb->ed_fsm.op->change_state(&ed_tcb->ed_fsm, ISIL_ED_TRANSFERING);
            }
        } else {
            chip_ddr_burst_interface->op->ack_write_audio_req(chip_ddr_burst_interface, audio_chan_driver);
            ed_tcb->ed_fsm.op->change_state(&ed_tcb->ed_fsm, ISIL_ED_TRANSFERING);
        }
        driver_gen_deliver_event(ed_tcb, 1);	//case 1 from running to transfering; case2 from transfering to done and continue to idle or other(close or suspend)
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   audio_decode_chan_delete_start_send_bitstream
 * Description   :   audio decode chan delete write bitstream to ddr request
 * Input         :   param1 is dpram request management struct
 *                   param2 is context
 * Output        :   null
 * Return        :   always 1
 ****************************************************************************************************************/
static int  audio_decode_chan_delete_start_send_bitstream(dpram_request_t *dpram_req, void *context)
{
    int ret=1;
    if((dpram_req!=NULL) && (context!=NULL)){

    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   init_submit_send_audio_bitstream_service
 * Description   :   init audio decode chan write request
 * Input         :   param1 is dpram request management struct
 *                   param2 is audio chan management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void init_submit_send_audio_bitstream_service(dpram_request_t *dpram_req, isil_audio_driver_t *audio_chan_driver)
{
    if((dpram_req!=NULL) || (audio_chan_driver!=NULL)){
        dpram_req->chip = audio_chan_driver->chip_audio->chip;
        dpram_req->type = DPRAM_NONBLOCK_TRANSFER_REQUEST;
        dpram_req->context = (void*)audio_chan_driver;
        dpram_req->req_callback = audio_decode_chan_start_send_bitstream;
        dpram_req->delete_req_notify = audio_decode_chan_delete_start_send_bitstream;
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_process_audio_encode
 * Description   :   chip audio process audio encode
 * Input         :   param1 is chip audio management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_process_audio_encode(chip_audio_t *chip_audio)
{
    if(chip_audio != NULL) {
        isil_audio_driver_t *audio_chan_driver;
        isil_audio_chan_buf_pool_t  *audio_buf_pool;
        ed_tcb_t            *opened_logic_chan_ed;
        isil_ed_fsm_t       *ed_fsm;
        unsigned long   flags;
        int i;

        for(i=0; i<ISIL_AUDIO_IN_CHAN_NUMBER; i++){
            if(chip_audio->read_bitstream_mode == READ_BITSTREAM_BY_PCI){
				if(i<ISIL_AUDIO_IN_CHAN16_ID){
					continue;
				}
            }
            chip_audio->op->get_audio_chan_driver(chip_audio, i, &audio_chan_driver);
            if(audio_chan_driver == NULL){
                continue;
            }
            if(atomic_read(&audio_chan_driver->audio_driver_opened_flag) == 0){
                continue;
            }
            opened_logic_chan_ed = &audio_chan_driver->opened_logic_chan_ed;
            ed_fsm = &opened_logic_chan_ed->ed_fsm;
            if(isil_chip_audio_is_in_robust_processing(chip_audio)){
                ed_fsm->op->feed_state_watchdog(ed_fsm);
                ISIL_DBG(ISIL_DBG_INFO, "%d\n", audio_chan_driver->audio_logic_chan_id);
                continue;
            }
            switch(ed_fsm->op->get_curr_state(ed_fsm)){
                case ISIL_ED_IDLE:
					spin_lock_irqsave(&audio_chan_driver->fsm_matrix_call_lock, flags);
                    driver_gen_deliver_event(opened_logic_chan_ed, 1);
					spin_unlock_irqrestore(&audio_chan_driver->fsm_matrix_call_lock, flags);		
                    break;
                case ISIL_ED_SUSPEND:
                case ISIL_ED_UNREGISTER:
                    ed_fsm->op->feed_state_watchdog(ed_fsm);
                default:
                    break;
            }
            audio_buf_pool = &audio_chan_driver->audio_buf_pool;
            if(audio_buf_pool->op->get_audio_section_tcb_pool_entry_number(audio_buf_pool)){
                ed_fsm->op->update_state_timeout_counter(ed_fsm);
            } else {
                ed_fsm->op->feed_state_watchdog(ed_fsm);    
            }
        }
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_process_audio_decode
 * Description   :   chip audio process audio encode
 * Input         :   param1 is chip audio management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_process_audio_decode(chip_audio_t *chip_audio)
{
    if(chip_audio != NULL){
        isil_audio_driver_t *audio_chan_driver;
        isil_audio_chan_buf_pool_t  *audio_buf_pool;
        ed_tcb_t            *opened_logic_chan_ed;
        isil_ed_fsm_t       *ed_fsm;
        unsigned long   flags;
        int i;

        for(i=ISIL_AUDIO_OUT_SPEAKER_ID; i<=ISIL_AUDIO_OUT_PLAYBACK_ID; i++){
            chip_audio->op->get_audio_chan_driver(chip_audio, i, &audio_chan_driver);
            if(audio_chan_driver == NULL){
                continue;
            }
            if(atomic_read(&audio_chan_driver->audio_driver_opened_flag) == 0){
                continue;
            }
            spin_lock_irqsave(&audio_chan_driver->fsm_matrix_call_lock, flags);
            opened_logic_chan_ed = &audio_chan_driver->opened_logic_chan_ed;
            ed_fsm = &opened_logic_chan_ed->ed_fsm;
            if(isil_chip_audio_is_in_robust_processing(chip_audio)){
                ed_fsm->op->feed_state_watchdog(ed_fsm);
                spin_unlock_irqrestore(&audio_chan_driver->fsm_matrix_call_lock, flags);        
                ISIL_DBG(ISIL_DBG_INFO, "%d\n", audio_chan_driver->audio_logic_chan_id);
                continue;
            }
            switch(ed_fsm->op->get_curr_state(ed_fsm)){
                case ISIL_ED_IDLE:
                    driver_gen_deliver_event(opened_logic_chan_ed, 1);
                    break;
                case ISIL_ED_SUSPEND:
                case ISIL_ED_UNREGISTER:
                    ed_fsm->op->feed_state_watchdog(ed_fsm);
                default:
                    break;
            }
            audio_buf_pool = &audio_chan_driver->audio_buf_pool;
            if(audio_buf_pool->op->get_audio_section_tcb_pool_entry_number(audio_buf_pool)){
                ed_fsm->op->update_state_timeout_counter(ed_fsm);
            } else {
                ed_fsm->op->feed_state_watchdog(ed_fsm);    
            }
            spin_unlock_irqrestore(&audio_chan_driver->fsm_matrix_call_lock, flags);        
        }
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_get_audio_chan_driver
 * Description   :   chip audio get audio chan
 * Input         :   param1 is chip audio management struct
 *                   param2 is chan id
 * Output        :   audio chan management struct
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_get_audio_chan_driver(chip_audio_t *chip_audio, int chan_id, isil_audio_driver_t **ptr_audio_driver)
{
    if((chip_audio!=NULL) && (ptr_audio_driver!=NULL)){
        *ptr_audio_driver = NULL;
        if((chan_id>=ISIL_AUDIO_IN_CHAN0_ID) && (chan_id<=ISIL_AUDIO_OUT_PLAYBACK_ID)){
            *ptr_audio_driver = &chip_audio->chip->audio_chan_driver[chan_id];
        }
    }
}

/****************************************************************************************************************
 * Veriable Name :   chip_audio_op
 * Description   :   for helping op chip audio, we define general interface
 ****************************************************************************************************************/
static struct chip_audio_operation	chip_audio_op = {
    .init = chip_audio_init,
    .reset = chip_audio_reset,
    .release = chip_audio_release,

    .set_encode_audio_chan_type = chip_audio_set_encode_audio_chan_type,
    .update_encode_audio_chan_type = chip_audio_update_encode_audio_chan_type,
    .get_encode_audio_chan_type = chip_audio_get_encode_audio_chan_type,

    .set_decode_audio_chan_type = chip_audio_set_decode_audio_chan_type,
    .update_decode_audio_chan_type = chip_audio_update_decode_audio_chan_type,
    .get_decode_audio_chan_type = chip_audio_get_decode_audio_chan_type,

    .set_enable_audio_chan = chip_audio_set_enable_audio_chan,
    .update_enable_audio_chan = chip_audio_update_enable_audio_chan,
    .get_audio_chan_enable_state = chip_audio_get_audio_chan_enable_state,

    .set_mute_audio_chan = chip_audio_set_mute_audio_chan,
    .update_mute_audio_chan = chip_audio_update_mute_audio_chan,
    .get_audio_chan_mute_state = chip_audio_get_audio_chan_mute_state,

    .get_chip_audio_param = chip_audio_get_chip_audio_param,
    .set_chip_audio_param = chip_audio_set_chip_audio_param,

    .process_audio_encode = chip_audio_process_audio_encode,
    .process_audio_decode = chip_audio_process_audio_decode,
    .get_audio_chan_driver = chip_audio_get_audio_chan_driver,
};

static ssize_t audio_proc_read(struct seq_file *seq, void *data)
{
    int i = 0;
    chip_audio_t *chip_audio = (chip_audio_t *)seq->private;
    isil_audio_driver_t           *audio_chan_driver;
    isil_audio_param_t            *running_param;


    running_param = &chip_audio->audio_encode.running_param;
    seq_printf(seq, "%8s %8s %8s\n", "BIT", "RATE", "TYPE");
    switch(running_param->op->get_bit_wide(running_param)) {
        case ISIL_AUDIO_8BIT:
            seq_printf(seq, "%8s ", "8bit");
            break;
        case ISIL_AUDIO_16BIT:
            seq_printf(seq, "%8s ", "16bit");
            break;
        default:
            seq_printf(seq, "%8s ", "N/A");
    }
    switch(running_param->op->get_sample_rate(running_param)) {
        case ISIL_AUDIO_8000:
            seq_printf(seq, "%8s ", "8K");
            break;
        case ISIL_AUDIO_16000:
            seq_printf(seq, "%8s ", "16K");
            break;
        case ISIL_AUDIO_32000:
            seq_printf(seq, "%8s ", "32K");
            break;
        case ISIL_AUDIO_44100:
            seq_printf(seq, "%8s ", "44.1K");
            break;
        case ISIL_AUDIO_48000:
            seq_printf(seq, "%8s ", "48K");
            break;
        default:
            seq_printf(seq, "%8s ", "N/A");
    }
    switch(running_param->op->get_type(running_param)) {
        case ISIL_AUDIO_PCM:
            seq_printf(seq, "%8s ", "PCM");
            break;
        case ISIL_AUDIO_ALAW:
            seq_printf(seq, "%8s ", "ALAW");
            break;
        case ISIL_AUDIO_ULAW:
            seq_printf(seq, "%8s ", "ULAW");
            break;
        case ISIL_AUDIO_ADPCM_32K:
            seq_printf(seq, "%8s ", "ADPCM_32K");
            break;
        case ISIL_AUDIO_ADPCM_16K:
            seq_printf(seq, "%8s ", "ADPCM_16K");
            break;
        case ISIL_AUDIO_ADPCM_48K:
            seq_printf(seq, "%8s ", "ADPCM_48K");
            break;
        default:
            seq_printf(seq, "%8s ", "N/A");
    }
    seq_printf(seq, "\nchannel state:\n");
    seq_printf(seq, "%8s %8s %8s %8s %8s %8s %8s\n", "logic_id", "type", "encode", "discard", "state", "queue", "pool");
    for(i = 0; i <= ISIL_AUDIO_OUT_PLAYBACK_ID; i++) {
        isil_audio_packet_queue_t     *audio_packet_queue;
        isil_audio_chan_buf_pool_t    *audio_buf_pool;
        chip_audio->op->get_audio_chan_driver(chip_audio, i, &audio_chan_driver);

        if(!audio_chan_driver || !atomic_read(&audio_chan_driver->audio_driver_opened_flag)) {
            continue;
        }
        audio_packet_queue = &audio_chan_driver->audio_packet_queue;
        audio_buf_pool     = &audio_chan_driver->audio_buf_pool;
        seq_printf(seq, "%8d %8d %8d %8d %8d %8d %8d\n", audio_chan_driver->audio_logic_chan_id, 
                audio_chan_driver->type, audio_chan_driver->i_frame_serial,
                audio_chan_driver->discard_number,
                audio_chan_driver->opened_logic_chan_ed.ed_fsm.op->get_curr_state(&audio_chan_driver->opened_logic_chan_ed.ed_fsm), 
                audio_packet_queue->op->get_curr_queue_entry_number(audio_packet_queue),
                audio_buf_pool->op->get_audio_section_tcb_pool_entry_number(audio_buf_pool));
    }

    return 0;
}

#if defined(PCI_PLATFORM)
static ssize_t audio_proc_write(struct file *file, const char __user *buffer,
        size_t count, loff_t *offset)
{
    chip_audio_t *chip_audio = (chip_audio_t *)isil_file_get_seq_private(file);
    isil_audio_driver_t         *audio_chan_driver;
    isil_audio_driver_t         *audio_decoder_driver;
    isil_audio_packet_queue_t   *audio_packet_queue;
    isil_audio_packet_section_t *audio_encode_section;
    isil_audio_packet_queue_t   *audio_decoder_packet_queue;
    isil_audio_packet_section_t *audio_decoder_section;
    isil_chip_t                 *chip;
    struct file *audio_file;
    char name[128];
    mm_segment_t fs; 
    char cmdbuf[128];
    char **argv;
    int argc, i;
    u32 loop = 0;

    if(count > 128){
        if(copy_from_user(cmdbuf, buffer, 128) != 0){
            ISIL_DBG(ISIL_DBG_INFO, ": copy_from_user failed\n");                    
        }
    }
    else{
        if(copy_from_user(cmdbuf, buffer, count) != 0){
            ISIL_DBG(ISIL_DBG_INFO, ": copy_from_user failed\n");                    
        }
    }
    argv = argv_split(GFP_KERNEL, cmdbuf, &argc);
    for(i = 0; i < (argc - 1); i++){
        ISIL_DBG(ISIL_DBG_INFO, "cmd %d: %s\n", i, argv[i]);
    }
    if(!chip_audio){
        goto write_release;
    }
    if((argc > 3) && (strcmp(argv[2], "loop") == 0)) {
        loop = 1;
    }

    chip = chip_audio->chip;

    if(argv[0] != NULL) {
        ed_tcb_t *encoder_ed = NULL;
        ed_tcb_t *decoder_ed = NULL;
        memset(name, 0, 128);
        sprintf(name, "%s/audio_chn%d.pcm", argv[1], atoi(argv[0]));
        audio_file = isil_kernel_file_open(name);
        fs=get_fs();
        set_fs(KERNEL_DS);
        chip_audio->op->get_audio_chan_driver(chip_audio, atoi(argv[0]), &audio_chan_driver);
        chip_audio->op->get_audio_chan_driver(chip_audio, ISIL_AUDIO_OUT_SPEAKER_ID, &audio_decoder_driver);
        if(!audio_chan_driver || !audio_decoder_driver) {
            ISIL_DBG(ISIL_DBG_ERR, "no such driver\n");
            goto audio_end;
        }
        encoder_ed = &audio_chan_driver->opened_logic_chan_ed;
        decoder_ed = &audio_decoder_driver->opened_logic_chan_ed;
        if(!encoder_ed->op) {
            init_isil_audio_encode_chan(audio_chan_driver, chip_audio, 1, 0, 0);
        }
        if(!decoder_ed->op) {
            init_isil_audio_decode_chan(audio_decoder_driver, chip_audio, 1, 0, ISIL_AUDIO_OUT_SPEAKER_ID);
        }
        encoder_ed->op->open(encoder_ed);
        driver_gen_resume_event(encoder_ed, 1);
        decoder_ed->op->open(decoder_ed);
        driver_gen_resume_event(decoder_ed, 1);
        audio_packet_queue = &audio_chan_driver->audio_packet_queue;
        audio_decoder_packet_queue = &audio_decoder_driver->audio_packet_queue;
        while(1) {
            schedule();
            if (signal_pending(current)) {
                goto audio_end;
            }
            audio_packet_queue->op->try_get_curr_consumer_from_queue(audio_packet_queue);
            audio_encode_section = audio_packet_queue->curr_consumer;
            //audio_encode_section->op->reference(audio_encode_section, &audio_encode_section);
            if(audio_encode_section){
                audio_file->f_op->write(audio_file, audio_encode_section->data + sizeof(ext_h264_nal_bitstream_t), audio_encode_section->payload_len - sizeof(ext_h264_nal_bitstream_t), &(audio_file->f_pos));
                /*ai to ao*/
                if(loop) {
                    audio_decoder_packet_queue->op->try_get_curr_producer_from_pool(audio_decoder_packet_queue, &audio_decoder_driver->audio_buf_pool);
                    audio_decoder_section = audio_decoder_packet_queue->curr_producer;
                    if(audio_decoder_section) {
                        audio_decoder_section->payload_len = audio_encode_section->payload_len;
                        audio_decoder_section->sample_rate = audio_encode_section->sample_rate;
                        audio_decoder_section->type        = audio_encode_section->type;
                        audio_decoder_section->bit_wide    = audio_encode_section->bit_wide;
                        audio_decoder_section->timestamp   = audio_encode_section->timestamp;

                        memcpy(audio_decoder_section->data, audio_encode_section->data, audio_encode_section->section_size);
                        audio_decoder_packet_queue->op->put_curr_producer_into_queue(audio_decoder_packet_queue);
                    }else{
                        ISIL_DBG(ISIL_DBG_INFO, ": audio decoder discard\n");
                    }

                }
                audio_packet_queue->op->release_curr_consumer(audio_packet_queue, &audio_chan_driver->audio_buf_pool);
            }

        }
audio_end:
        if(encoder_ed) {
            encoder_ed->op->close(encoder_ed);
        }
        if(decoder_ed) {
            decoder_ed->op->close(decoder_ed);
        }
        set_fs(fs);
        isil_kernel_file_close(audio_file);
    }

write_release:
    argv_free(argv);
    return count;
}

/****************************************************************************************************************
 * Function Name :   chip_audio_top_level_irq_handler
 * Description   :   push audio encode bitstream's isr
 * Input         :   param1 is irq number
 *                   param2 is context
 * Output        :   null
 * Return        :   always 0
 ****************************************************************************************************************/
static int chip_audio_top_level_irq_handler(int irq, void *context)
{
    isil_chip_t		*chip;
    chip_audio_t	*chip_audio;
    chip_audio_encode_t *audio_encode;
    audio_encode_control_t  *audio_control;
    chip_audio_push_isr_interface_t *audio_push_isr_interface;
    chip_audio_push_buf_state_t     *audio_push_buf;
    isil_tasklet_entry_t    *audio_encode_tasklet;
    unsigned int    chan_id_flags, ping_pong_flags;
    int chan_id, wr_ptr, clear_chan_id;

    chip = (isil_chip_t*)context;
    chip_audio = &chip->chip_audio;
    audio_push_isr_interface = &chip_audio->audio_push_interface;
    audio_push_buf = &audio_push_isr_interface->audio_push_buf;
    audio_encode = &chip_audio->audio_encode;
    audio_control = &audio_encode->audio_control;

    if(chip->io_op->chip_read32(chip, INTERRUPT_FLAGS_EXT) & (1 << IRQ_EXT_AUDIO_EOF_INTR)){
        chan_id_flags = chip->io_op->chip_read32(chip, PCI_INTR_CHAN_SOURCE);
        chan_id_flags >>= PCI_AUDIO_CHAN_OFFSET_BASE;
        if(chan_id_flags){
            clear_chan_id = 0;
            ping_pong_flags = chip->io_op->chip_read32(chip, PCI_PREVIEW_AUDIO_AB);
            ping_pong_flags >>= PCI_AUDIO_CHAN_OFFSET_BASE;
            for(chan_id=0; chan_id<16; chan_id++){
                wr_ptr = audio_push_buf->op->get_push_buf_wr_ptr(audio_push_buf, chan_id);
                if(chan_id_flags & (1<<chan_id)){
                    clear_chan_id |= (1<<chan_id);
                    audio_push_buf->chan_id_flags[chan_id][wr_ptr] = chan_id_flags;
                    audio_push_buf->ping_pong_flags[chan_id][wr_ptr] = ping_pong_flags;
                }
                if((chan_id==3) && (clear_chan_id&0xf)){
                    audio_push_buf->op->receive_isr_request(audio_push_buf, (clear_chan_id&0xf), 0, chip);
                } else if((chan_id==7) && (clear_chan_id&0xf0)){
                    audio_push_buf->op->receive_isr_request(audio_push_buf, (clear_chan_id&0xf0), 1, chip);
                } else if((chan_id==11) && (clear_chan_id&0xf00)){
                    audio_push_buf->op->receive_isr_request(audio_push_buf, (clear_chan_id&0xf00), 2, chip);
                } else if((chan_id==15) && (clear_chan_id&0xf000)){
                    audio_push_buf->op->receive_isr_request(audio_push_buf, (clear_chan_id&0xf000), 3, chip); 
                }
            }
            chip->io_op->chip_write32(chip, PCI_INTR_CHAN_SOURCE, ((clear_chan_id)<<PCI_AUDIO_CHAN_OFFSET_BASE));
            audio_control->op->toggle_pci_flowcn(audio_control, chip);
            audio_encode_tasklet = &audio_push_isr_interface->audio_encode_tasklet;
            audio_encode_tasklet->op->schedule(audio_encode_tasklet);
        }
        chip->io_op->chip_write32(chip, INTERRUPT_FLAGS_EXT, (1<<IRQ_EXT_AUDIO_EOF_INTR));
    }
    return 0;
}

/****************************************************************************************************************
 * Function Name :   chip_audio_push_buf_state_init
 * Description   :   audio push buf state init
 * Input         :   param1 is audio push buf state management struct
 *                   param2 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_push_buf_state_init(chip_audio_push_buf_state_t *audio_push_buf_state, isil_chip_t *chip)
{
    int i;
    spin_lock_init(&audio_push_buf_state->audio_push_buf_state_lock);
    for(i=0; i<16; i++){
        audio_push_buf_state->chan_push_buf_wr_ptr[i] = 0;
        audio_push_buf_state->chan_push_buf_rd_ptr[i] = 0;
        audio_push_buf_state->cached_chan_push_buf_state[i][0] = 0;
        audio_push_buf_state->cached_chan_push_buf_state[i][1] = 0;
        audio_push_buf_state->cached_chan_push_buf_state[i][2] = 0;
        audio_push_buf_state->cached_chan_push_buf_state[i][3] = 0;
        audio_push_buf_state->chan_id_flags[i][0] = 0;
        audio_push_buf_state->chan_id_flags[i][1] = 0;
        audio_push_buf_state->chan_id_flags[i][2] = 0;
        audio_push_buf_state->chan_id_flags[i][3] = 0;
        audio_push_buf_state->ping_pong_flags[i][0] = 0;
        audio_push_buf_state->ping_pong_flags[i][1] = 0;
        audio_push_buf_state->ping_pong_flags[i][2] = 0;
        audio_push_buf_state->ping_pong_flags[i][3] = 0;
    }
    for(i=0; i<4; i++){
        audio_push_buf_state->push_ping_pong_buf_state[i] = 0;
        audio_push_buf_state->push_ping_buf_addr[i] = get_push_buf(BUF_TYPE_AUDIO, 0, &audio_push_buf_state->push_ping_dma_addr[i], chip);
        audio_push_buf_state->push_pong_buf_addr[i] = get_push_buf(BUF_TYPE_AUDIO, 0, &audio_push_buf_state->push_pong_dma_addr[i], chip);
        audio_push_buf_state->op->set_curr_push(audio_push_buf_state, i, chip);
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_push_buf_state_close
 * Description   :   audio push buf state close
 * Input         :   param1 is audio push buf state management struct
 *                   param2 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_push_buf_state_close(chip_audio_push_buf_state_t *audio_push_buf_state, isil_chip_t *chip)
{
    unsigned long   flags;
    int i;

    spin_lock_irqsave(&audio_push_buf_state->audio_push_buf_state_lock, flags);
    for(i=0; i<16; i++){
        audio_push_buf_state->chan_push_buf_wr_ptr[i] = 0;
        audio_push_buf_state->chan_push_buf_rd_ptr[i] = 0;
        audio_push_buf_state->chan_id_flags[i][0] = 0;
        audio_push_buf_state->chan_id_flags[i][1] = 0;
        audio_push_buf_state->chan_id_flags[i][2] = 0;
        audio_push_buf_state->chan_id_flags[i][3] = 0;
        audio_push_buf_state->ping_pong_flags[i][0] = 0;
        audio_push_buf_state->ping_pong_flags[i][1] = 0;
        audio_push_buf_state->ping_pong_flags[i][2] = 0;
        audio_push_buf_state->ping_pong_flags[i][3] = 0;
    }
    for(i=0; i<4; i++){
        audio_push_buf_state->push_ping_pong_buf_state[i] = 0;
        release_push_buf(BUF_TYPE_AUDIO, 0, audio_push_buf_state->push_ping_buf_addr[i], chip);
        release_push_buf(BUF_TYPE_AUDIO, 0, audio_push_buf_state->push_pong_buf_addr[i], chip);
    }
    spin_unlock_irqrestore(&audio_push_buf_state->audio_push_buf_state_lock, flags);
}

/****************************************************************************************************************
 * Function Name :   chip_audio_push_buf_state_receive_isr_request
 * Description   :   audio push buf state close
 * Input         :   param1 is audio push buf state management struct
 *                   param2 is chan section'a range
 *                   param3 is group id
 *                   param4 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_push_buf_state_receive_isr_request(chip_audio_push_buf_state_t *audio_push_buf_state, int chan_section_range, int group_id, isil_chip_t *chip)
{
    int chan_id;
    unsigned long   flags;

    spin_lock_irqsave(&audio_push_buf_state->audio_push_buf_state_lock, flags);
    for(chan_id=0; chan_id<16; chan_id++){
        if(chan_section_range & (1<<chan_id)){
            if(((audio_push_buf_state->chan_push_buf_wr_ptr[chan_id]+1)&3) == audio_push_buf_state->chan_push_buf_rd_ptr[chan_id]){
                audio_push_buf_state->chan_push_buf_rd_ptr[chan_id]++;
                audio_push_buf_state->chan_push_buf_rd_ptr[chan_id] &= 3;
            }
            audio_push_buf_state->cached_chan_push_buf_state[chan_id][audio_push_buf_state->chan_push_buf_wr_ptr[chan_id]] = audio_push_buf_state->push_ping_pong_buf_state[group_id];
            audio_push_buf_state->chan_push_buf_wr_ptr[chan_id]++;
            audio_push_buf_state->chan_push_buf_wr_ptr[chan_id] &= 3;        
        }
    }
    audio_push_buf_state->push_ping_pong_buf_state[group_id]++;
    audio_push_buf_state->push_ping_pong_buf_state[group_id] &= 3;
    audio_push_buf_state->op->set_curr_push(audio_push_buf_state, group_id, chip);
    spin_unlock_irqrestore(&audio_push_buf_state->audio_push_buf_state_lock, flags);
}

/****************************************************************************************************************
 * Function Name :   chip_audio_push_buf_state_push_buf_is_empty
 * Description   :   audio push buf state is empty
 * Input         :   param1 is audio push buf state management struct
 *                   param2 is chan id
 * Output        :   null
 * Return        :   empty state
 ****************************************************************************************************************/
static int  chip_audio_push_buf_state_push_buf_is_empty(chip_audio_push_buf_state_t *audio_push_buf_state, int chan_id)
{
    unsigned long   flags;
    int ret;

    spin_lock_irqsave(&audio_push_buf_state->audio_push_buf_state_lock, flags);
    ret = (audio_push_buf_state->chan_push_buf_wr_ptr[chan_id] == audio_push_buf_state->chan_push_buf_rd_ptr[chan_id]) ? 1 : 0;
    spin_unlock_irqrestore(&audio_push_buf_state->audio_push_buf_state_lock, flags);
    return ret;
}

/****************************************************************************************************************
 * Function Name :   chip_audio_push_buf_state_get_push_buf_rd_ptr
 * Description   :   get chip audio push buf's rd ptr
 * Input         :   param1 is audio push buf state management struct
 *                   param2 is chan id
 * Output        :   null
 * Return        :   rd_ptr
 ****************************************************************************************************************/
static int  chip_audio_push_buf_state_get_push_buf_rd_ptr(chip_audio_push_buf_state_t *audio_push_buf_state, int chan_id)
{
    return audio_push_buf_state->chan_push_buf_rd_ptr[chan_id];
}

/****************************************************************************************************************
 * Function Name :   chip_audio_push_buf_state_get_push_buf_wr_ptr
 * Description   :   get chip audio push buf's wr ptr
 * Input         :   param1 is audio push buf state management struct
 *                   param2 is chan id
 * Output        :   null
 * Return        :   wr_ptr
 ****************************************************************************************************************/
static int  chip_audio_push_buf_state_get_push_buf_wr_ptr(chip_audio_push_buf_state_t *audio_push_buf_state, int chan_id)
{
    return audio_push_buf_state->chan_push_buf_wr_ptr[chan_id];
}

/****************************************************************************************************************
 * Function Name :   chip_audio_push_buf_state_get_curr_push
 * Description   :   get chip audio curr push buf's postion
 * Input         :   param1 is audio push buf state management struct
 *                   param2 is chan id
 * Output        :   param1 is curr push buf addr
 *                   param2 is curr push buf's dma addr
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_push_buf_state_get_curr_push(chip_audio_push_buf_state_t *audio_push_buf_state, char **curr_push_buf_addr, dma_addr_t *curr_dma_addr, int chan_id)
{
    unsigned long   flags;
    int group_id;
    int cached_chan_push_buf_state, ping_pong_flags, chan_push_buf_rd_ptr;

    spin_lock_irqsave(&audio_push_buf_state->audio_push_buf_state_lock, flags);
    group_id = chan_id>>2;
    chan_push_buf_rd_ptr = audio_push_buf_state->chan_push_buf_rd_ptr[chan_id];
    ping_pong_flags = audio_push_buf_state->ping_pong_flags[chan_id][chan_push_buf_rd_ptr];
    cached_chan_push_buf_state = audio_push_buf_state->cached_chan_push_buf_state[chan_id][chan_push_buf_rd_ptr];
    if(ping_pong_flags & (1<<chan_id)){
        switch(cached_chan_push_buf_state){
            case 0:
            case 3:
                *curr_push_buf_addr = (char *)audio_push_buf_state->push_ping_buf_addr[group_id];
                *curr_dma_addr = audio_push_buf_state->push_ping_dma_addr[group_id];
                break;
            default:
                *curr_push_buf_addr = (char *)(audio_push_buf_state->push_ping_buf_addr[group_id] + AUDIO_PUSH_DATA_OFFSET_OF_PING_PONG);
                *curr_dma_addr = audio_push_buf_state->push_ping_dma_addr[group_id]+AUDIO_PUSH_DATA_OFFSET_OF_PING_PONG;
                break;
        }
    } else {
        switch(cached_chan_push_buf_state){
            case 0:
            case 1:
                *curr_push_buf_addr = (char *)audio_push_buf_state->push_pong_buf_addr[group_id];
                *curr_dma_addr = audio_push_buf_state->push_pong_dma_addr[group_id];
                break;
            default:
                *curr_push_buf_addr = (char *)(audio_push_buf_state->push_pong_buf_addr[group_id] + AUDIO_PUSH_DATA_OFFSET_OF_PING_PONG);
                *curr_dma_addr = audio_push_buf_state->push_pong_dma_addr[group_id]+AUDIO_PUSH_DATA_OFFSET_OF_PING_PONG;
                break;
        }    
    }
    audio_push_buf_state->chan_push_buf_rd_ptr[chan_id]++;
    audio_push_buf_state->chan_push_buf_rd_ptr[chan_id] &= 3;
    spin_unlock_irqrestore(&audio_push_buf_state->audio_push_buf_state_lock, flags);
}

/****************************************************************************************************************
 * Function Name :   chip_audio_push_buf_state_set_curr_push
 * Description   :   set chip audio curr push buf's postion
 * Input         :   param1 is audio push buf state management struct
 *                   param2 is group id
 *                   param3 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_push_buf_state_set_curr_push(chip_audio_push_buf_state_t *audio_push_buf_state, int group_id, isil_chip_t *chip)
{
    int reg_ping_offset, reg_pong_offset;
    unsigned int    ping_dma_addr, pong_dma_addr;

    switch(group_id){
        case 0:
            reg_ping_offset = PCI_AUDIO_CHAN00_CHAN03_PING_BASE_ADDR_REG;
            reg_pong_offset = PCI_AUDIO_CHAN00_CHAN03_PONG_BASE_ADDR_REG;
            break;
        case 1:
            reg_ping_offset = PCI_AUDIO_CHAN04_CHAN07_PING_BASE_ADDR_REG;
            reg_pong_offset = PCI_AUDIO_CHAN04_CHAN07_PONG_BASE_ADDR_REG;
            break;
        case 2:
            reg_ping_offset = PCI_AUDIO_CHAN08_CHAN11_PING_BASE_ADDR_REG;
            reg_pong_offset = PCI_AUDIO_CHAN08_CHAN11_PONG_BASE_ADDR_REG;
            break;
        default:
            reg_ping_offset = PCI_AUDIO_CHAN12_CHAN15_PING_BASE_ADDR_REG;
            reg_pong_offset = PCI_AUDIO_CHAN12_CHAN15_PONG_BASE_ADDR_REG;
            break;
    }
    switch(audio_push_buf_state->push_ping_pong_buf_state[group_id]){
        case 0:
            ping_dma_addr = audio_push_buf_state->push_ping_dma_addr[group_id];
            pong_dma_addr = audio_push_buf_state->push_pong_dma_addr[group_id];
            break;
        case 1:
            ping_dma_addr = audio_push_buf_state->push_ping_dma_addr[group_id]+AUDIO_PUSH_DATA_OFFSET_OF_PING_PONG;
            pong_dma_addr = audio_push_buf_state->push_pong_dma_addr[group_id];
            break;
        case 2:
            ping_dma_addr = audio_push_buf_state->push_ping_dma_addr[group_id]+AUDIO_PUSH_DATA_OFFSET_OF_PING_PONG;
            pong_dma_addr = audio_push_buf_state->push_pong_dma_addr[group_id]+AUDIO_PUSH_DATA_OFFSET_OF_PING_PONG;
            break;
        default:
            ping_dma_addr = audio_push_buf_state->push_ping_dma_addr[group_id];
            pong_dma_addr = audio_push_buf_state->push_pong_dma_addr[group_id]+AUDIO_PUSH_DATA_OFFSET_OF_PING_PONG;
            break;
    }
    chip->io_op->chip_write32(chip, reg_ping_offset, ping_dma_addr);
    chip->io_op->chip_write32(chip, reg_pong_offset, pong_dma_addr);
}

/****************************************************************************************************************
 * Veriable Name :   chip_audio_push_buf_state_op
 * Description   :   for helping op chip audio push buf state, we define general interface
 ****************************************************************************************************************/
static struct chip_audio_push_buf_state_operation   chip_audio_push_buf_state_op = {
    .init = chip_audio_push_buf_state_init,
    .close = chip_audio_push_buf_state_close,
    .receive_isr_request = chip_audio_push_buf_state_receive_isr_request,
    .push_buf_is_empty = chip_audio_push_buf_state_push_buf_is_empty,
    .get_push_buf_rd_ptr = chip_audio_push_buf_state_get_push_buf_rd_ptr,
    .get_push_buf_wr_ptr = chip_audio_push_buf_state_get_push_buf_wr_ptr,
    .get_curr_push = chip_audio_push_buf_state_get_curr_push,
    .set_curr_push = chip_audio_push_buf_state_set_curr_push,
};

/****************************************************************************************************************
 * Function Name :   init_chip_audio_push_buf_state
 * Description   :   init chip audio push buf state
 * Input         :   param1 is audio push buf state management struct
 *                   param2 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void init_chip_audio_push_buf_state(chip_audio_push_buf_state_t *audio_push_buf_state, isil_chip_t *chip)
{
    audio_push_buf_state->op = &chip_audio_push_buf_state_op;
    audio_push_buf_state->op->init(audio_push_buf_state, chip);
}

/****************************************************************************************************************
 * Function Name :   close_chip_audio_push_buf_state
 * Description   :   close chip audio push buf state
 * Input         :   param1 is audio push buf state management struct
 *                   param2 is chip management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void close_chip_audio_push_buf_state(chip_audio_push_buf_state_t *audio_push_buf_state, isil_chip_t *chip)
{
    audio_push_buf_state->op->close(audio_push_buf_state, chip);
}

/****************************************************************************************************************
 * Function Name :   chip_audio_push_isr_interface_init
 * Description   :   chip audio push isr interface
 * Input         :   param1 is audio push buf state management struct
 *                   param2 is irq
 *                   param3 is context
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_push_isr_interface_init(chip_audio_push_isr_interface_t *audio_push_isr_interface, int irq, void *context)
{
    atomic_set(&audio_push_isr_interface->counter, 0);
    audio_push_isr_interface->chip_audio_isr = chip_audio_top_level_irq_handler;
    audio_push_isr_interface->context = context;
    audio_push_isr_interface->irq = irq;
}

/****************************************************************************************************************
 * Function Name :   chip_audio_push_isr_interface_open_chip_audio
 * Description   :   chip audio push isr interface open chip audio
 * Input         :   param1 is audio push buf state management struct
 *                   param2 is logic chan id
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void	chip_audio_push_isr_interface_open_chip_audio(chip_audio_push_isr_interface_t *audio_push_isr_interface)
{
    isil_chip_t     *chip;
	chip_audio_t	*chip_audio;
	chip_audio_encode_t 	*audio_encode;
	audio_encode_control_t	*audio_control;
    unsigned int	value;

    chip = (isil_chip_t *)audio_push_isr_interface->context;
	chip_audio = &chip->chip_audio;
	audio_encode = &chip_audio->audio_encode;
	audio_control = &audio_encode->audio_control;
    if(atomic_read(&audio_push_isr_interface->counter) == 0){
        atomic_inc(&audio_push_isr_interface->counter);
        init_chip_audio_push_buf_state(&audio_push_isr_interface->audio_push_buf, chip);
        init_isil_audio_encode_tasklet(&audio_push_isr_interface->audio_encode_tasklet, (unsigned long)chip);
        if(chip_request_irq(chip, audio_push_isr_interface->irq, audio_push_isr_interface->chip_audio_isr, "audio_push", (void*)chip) != ISIL_OK){
            ISIL_DBG(ISIL_DBG_INFO, ": register audio push top isr err\n\r");
        } else {
            value = chip->io_op->chip_read32(chip, PCI_MASTER_CONTROL);
            value &= 0xff0000ff;
			value |= 0x00ffff00;
            chip->io_op->chip_write32(chip, PCI_MASTER_CONTROL, value);

            value = chip->io_op->chip_read32(chip, PCI_MASTER_INTR_ENABLE_REG);
            value |= (1<<IRQ_EXT_AUDIO_EOF_INTR);
            chip->io_op->chip_write32(chip, PCI_MASTER_INTR_ENABLE_REG, value);

			audio_control->op->toggle_pci_flowcn(audio_control, chip);
        }
    } else {
        atomic_inc(&audio_push_isr_interface->counter);
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_push_isr_interface_close_chip_audio
 * Description   :   chip audio push isr interface close chip audio
 * Input         :   param1 is audio push buf state management struct
 *                   param2 is logic chan id
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void	chip_audio_push_isr_interface_close_chip_audio(chip_audio_push_isr_interface_t *audio_push_isr_interface)
{
    isil_chip_t     *chip;
    unsigned int    value;

    chip = (isil_chip_t *)audio_push_isr_interface->context;
    if(atomic_read(&audio_push_isr_interface->counter)){
        if(atomic_dec_and_test(&audio_push_isr_interface->counter)){
            value = chip->io_op->chip_read32(chip, PCI_MASTER_INTR_ENABLE_REG);
            value &= ~(1<<IRQ_EXT_AUDIO_EOF_INTR);
            chip->io_op->chip_write32(chip, PCI_MASTER_INTR_ENABLE_REG, value);

            value = chip->io_op->chip_read32(chip, PCI_MASTER_CONTROL);
            value &= 0xff0000ff;
            chip->io_op->chip_write32(chip, PCI_MASTER_CONTROL, value);

            chip_free_irq(chip, audio_push_isr_interface->irq, (void*)chip);
            close_isil_audio_encode_tasklet(&audio_push_isr_interface->audio_encode_tasklet);
            close_chip_audio_push_buf_state(&audio_push_isr_interface->audio_push_buf, chip);
        }
    }
}

/****************************************************************************************************************
 * Function Name :   chip_audio_push_isr_interface_disable_chip_audio_chan
 * Description   :   chip audio push isr interface disable chip audio chan
 * Input         :   param1 is audio push buf state management struct
 *                   param2 is logic chan id
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_push_isr_interface_disable_chip_audio_chan(chip_audio_push_isr_interface_t *audio_push_isr_interface, int logic_chan_id)
{
    isil_chip_t     *chip;
    chip_audio_t    *chip_audio;

    chip = (isil_chip_t *)audio_push_isr_interface->context;
    chip_audio = &chip->chip_audio;
    chip_audio->op->set_enable_audio_chan(chip_audio, logic_chan_id, AUDIO_DISABLE);
}

/****************************************************************************************************************
 * Function Name :   chip_audio_push_isr_interface_enable_chip_audio_chan
 * Description   :   chip audio push isr interface enable chip audio chan
 * Input         :   param1 is audio push buf state management struct
 *                   param2 is logic chan id
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_audio_push_isr_interface_enable_chip_audio_chan(chip_audio_push_isr_interface_t *audio_push_isr_interface, int logic_chan_id)
{
    isil_chip_t     *chip;
    chip_audio_t    *chip_audio;

    chip = (isil_chip_t *)audio_push_isr_interface->context;
    chip_audio = &chip->chip_audio;
    chip_audio->op->set_enable_audio_chan(chip_audio, logic_chan_id, AUDIO_ENABLE);
}

/****************************************************************************************************************
 * Function Name :   chip_audio_push_isr_interface_curr_open_chip_audio_chan_number
 * Description   :   get chip audio push isr interface open chip audio chan number
 * Input         :   param1 is audio push buf state management struct
 *                   param2 is logic chan id
 * Output        :   null
 * Return        :   number
 ****************************************************************************************************************/
static int  chip_audio_push_isr_interface_curr_open_chip_audio_chan_number(chip_audio_push_isr_interface_t *audio_push_isr_interface)
{
    return atomic_read(&audio_push_isr_interface->counter);
}

/****************************************************************************************************************
 * Veriable Name :   chip_audio_push_isr_interface_op
 * Description   :   for helping op chip audio push isr interface, we define general interface
 ****************************************************************************************************************/
static struct chip_audio_push_isr_interface_operation    chip_audio_push_isr_interface_op = {
    .init = chip_audio_push_isr_interface_init,
    .open_chip_audio = chip_audio_push_isr_interface_open_chip_audio,
    .close_chip_audio = chip_audio_push_isr_interface_close_chip_audio,
    .disable_chip_audio_chan = chip_audio_push_isr_interface_disable_chip_audio_chan,
    .enable_chip_audio_chan = chip_audio_push_isr_interface_enable_chip_audio_chan,
    .curr_open_chip_audio_chan_number = chip_audio_push_isr_interface_curr_open_chip_audio_chan_number,
};

/****************************************************************************************************************
 * Function Name :   init_chip_audio_push_isr_interface
 * Description   :   init chip audio push isr interface
 * Input         :   param1 is audio push buf state management struct
 *                   param2 is logic chan id
 * Output        :   null
 * Return        :   number
 ****************************************************************************************************************/
static void	init_chip_audio_push_isr_interface(chip_audio_push_isr_interface_t *audio_push_isr_interface, void *context)
{
    audio_push_isr_interface->op = &chip_audio_push_isr_interface_op;
    audio_push_isr_interface->op->init(audio_push_isr_interface, IRQ_AUDIO_EOF_TYPE_INTR, context);
}
#endif

/****************************************************************************************************************
 * Function Name :   start_chip_audio_robust_process
 * Description   :   start chip audio robust process
 * Input         :   param1 is chip audio management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void    start_chip_audio_robust_process(chip_audio_t *chip_audio)
{
    robust_processing_control_t *chip_audio_robust;

    start_chip_robust_process(chip_audio->chip);
    chip_audio_robust = &chip_audio->chip_audio_robust;
    chip_audio_robust->op->start_robust_processing(chip_audio_robust);
}

/****************************************************************************************************************
 * Function Name :   isil_chip_audio_robust_process_done
 * Description   :   notify chip audio robust process over
 * Input         :   param1 is chip audio management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void    isil_chip_audio_robust_process_done(chip_audio_t *chip_audio)
{
    robust_processing_control_t *chip_audio_robust;
    chip_audio_robust = &chip_audio->chip_audio_robust;
    chip_audio_robust->op->robust_processing_done(chip_audio_robust);
}

/****************************************************************************************************************
 * Function Name :   isil_chip_audio_wait_robust_process_done
 * Description   :   wait chip audio robust process over
 * Input         :   param1 is chip audio management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void    isil_chip_audio_wait_robust_process_done(chip_audio_t *chip_audio)
{
    robust_processing_control_t *chip_audio_robust;

    chip_wait_robust_process_done(chip_audio->chip);
    chip_audio_robust = &chip_audio->chip_audio_robust;
    chip_audio_robust->op->wait_robust_processing_done(chip_audio_robust);
}

/****************************************************************************************************************
 * Function Name :   isil_chip_audio_is_in_robust_processing
 * Description   :   whether or not chip audio is robust process
 * Input         :   param1 is chip audio management struct
 * Output        :   null
 * Return        :   state
 ****************************************************************************************************************/
int isil_chip_audio_is_in_robust_processing(chip_audio_t *chip_audio)
{
    robust_processing_control_t *chip_audio_robust;
    int ret = 0;

    chip_audio_robust = &chip_audio->chip_audio_robust;
    if(chip_audio_robust->op->is_in_robust_processing(chip_audio_robust) ||
            chip_is_in_robust_processing(chip_audio->chip)){
        ret = 1;
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   init_chip_audio
 * Description   :   init chip audio
 * Input         :   param1 is chip audio management struct
 *                   param2 is chip management struct
 *                   param3 is bus id
 * Output        :   null
 * Return        :   ISIL_OK is meaning ok, other is meaning err
 ****************************************************************************************************************/
int init_chip_audio(chip_audio_t *chip_audio, isil_chip_t *chip, int bus_id)
{
    int	ret = ISIL_ERR;
    if((chip_audio!=NULL) && (chip!=NULL)){
        isil_proc_register_s *audio_proc;
        chip_audio->chip = chip;
#if defined(PCI_PLATFORM)
		chip_audio->read_bitstream_mode = READ_BITSTREAM_BY_PCI;
#else
        chip_audio->read_bitstream_mode = READ_BITSTREAM_BY_RAM;
#endif
        chip_audio->op = &chip_audio_op;
        ret = chip_audio->op->init(chip_audio);
        if(ret) {
            ISIL_DBG(ISIL_DBG_ERR, "init chip_audio failed %d\n", ret);
            return ret;
        }
        chip_audio->op->set_chip_audio_param(chip_audio);

        init_robust_processing_control(&chip_audio->chip_audio_robust);
#if defined(PCI_PLATFORM)
        init_chip_audio_push_isr_interface(&chip_audio->audio_push_interface, chip);
#endif
        audio_proc = &chip->audio_proc;
        strcpy(audio_proc->name, "audio");
        audio_proc->read  = audio_proc_read;
#if defined(PCI_PLATFORM)
        audio_proc->write = audio_proc_write;
#else
        audio_proc->write = NULL;
#endif
        audio_proc->private = chip_audio;
        isil_module_register(chip, audio_proc);
        ret = ISIL_OK;
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   remove_chip_audio
 * Description   :   remove chip audio
 * Input         :   param1 is chip audio management struct
 * Output        :   null
 * Return        :   ISIL_OK is meaning ok, other is meaning err
 ****************************************************************************************************************/
void	remove_chip_audio(chip_audio_t *chip_audio)
{
    if(chip_audio != NULL){
        isil_module_unregister(chip_audio->chip, &chip_audio->chip->audio_proc);
        chip_audio->audio_enable.op->update_enable_pcm_encode(&chip_audio->audio_enable, AUDIO_DISABLE);
        chip_audio->audio_enable.op->update_enable_adpcm_encode(&chip_audio->audio_enable, AUDIO_DISABLE);
        chip_audio->audio_enable.op->update_enable_adpcm_decode(&chip_audio->audio_enable, AUDIO_DISABLE);
        chip_audio->op->set_chip_audio_param(chip_audio);
    }
}


